@@ -201,7 +201,7 @@ free_digits(DIGIT * dp, ulong count)
   if (!dp) 
     return;
   digits_freed += count;
-  my_free((char *)dp);
+  free((char *)dp);
   dp = NULL;
 }
 #else
@@ -210,10 +210,11 @@ free_digits(DIGIT * dp, ulong count)
 static void
 free_digits(DIGIT * dp, ulong  count)
 {
+  (void) count;
   /* Mar 4 2000 R. Wichmann: check for dp == NULL */
   if (!dp) 
     return;
-  my_free((char *)dp);
+  free((char *)dp);
   dp = NULL;
   count = 0;
 }
@@ -231,7 +232,7 @@ allocate_digits(ulong alloclen)
 #ifdef COUNT_ALLOC
     dgs_alloc += alloclen;
 #endif
-    digit_ptr = (DIGIT *)sl_malloc(alloclen * sizeof(DIGIT));
+    digit_ptr = (DIGIT *) malloc(alloclen * sizeof(DIGIT));
     if (digit_ptr == NULL)
     {
 	big_errno = BIG_MEMERR;
@@ -346,7 +347,7 @@ vect_div_digit(DIGIT *digvect, ulong *len, DIGIT d)
     if (d == 0)
     {
 	big_errno = BIG_DIV_ZERO;
-	return -1;
+	return 0 /* -1 */;
     }
     if (d == 1)
     {
@@ -624,7 +625,7 @@ big_init_pkg()
     big_create(&big_one);
     big_set_long((long)1, &big_one);
     length_last_string = 10;
-    if ((last_string = sl_malloc(length_last_string)) == NULL)
+    if ((last_string = malloc(length_last_string)) == NULL)
     {
 	big_errno = BIG_MEMERR;
     }
@@ -642,7 +643,7 @@ big_release_pkg()
     big_destroy(&tmp_round);
     big_destroy(&big_one);
     if (last_string != NULL)
-      my_free(last_string);
+      free(last_string);
 #ifdef COUNT_ALLOC
     printf("Allocated digits: %lu\n", dgs_alloc);
     printf("Freed digits:     %lu\n", digits_freed);
@@ -953,8 +954,8 @@ int base;
     if (str_length > length_last_string)
     {
 	if (last_string != NULL)
-	  my_free(last_string);
-	if ((last_string = sl_malloc(str_length)) == NULL)
+	  free(last_string);
+	if ((last_string = malloc(str_length)) == NULL)
 	{
 	    big_errno = BIG_MEMERR;
 	    return NULL;
@@ -1234,7 +1235,7 @@ bignum *c;
     DIGIT *dp_x, *dp_xstart, *dp_xend;
     DIGIT *dp_y, *dp_ystart, *dp_yend;
     DIGIT *dp_z, *dp_zstart, *dp_zend, *dp_zsumstart;
-    ulong len_x, len_y, len_z;
+    ulong len_x, len_y /* , len_z */;
     DIGIT2 res;
     DIGIT tmp_res;	     /* Without use of this, gcc (v 1.39) generates */
 			     /* erroneous code on a sun386 machine */
@@ -1293,7 +1294,7 @@ bignum *c;
 	    return big_errno;
 	}
 	dp_zsumstart = tmp_mul.dp;
-	len_z = tmp_mul.dgs_alloc;
+	/* len_z = tmp_mul.dgs_alloc; */
     }	
     else
     {
@@ -1303,7 +1304,7 @@ bignum *c;
 	    return big_errno;
 	}
 	dp_zsumstart = c->dp;
-	len_z = c->dgs_alloc;
+	/* len_z = c->dgs_alloc; */
     }
 
     dp_xend = dp_xstart + len_x;
@@ -1361,7 +1362,10 @@ bignum *r;
     DIGIT norm, qhat, t1, t2, t3, v1, v2, u1, u2, u3;
     DIGIT2 temp, res, carry;
     long a_l, b_l, q_l, r_l;
-    ulong i, j, m, n;
+#if 0
+    ulong i; 
+#endif
+    ulong j, m, n;
     int cmp, q_eq_a, q_eq_b, r_eq_a, r_eq_b;
 
     if (big_errno != BIG_OK)
@@ -1533,7 +1537,9 @@ bignum *r;
 	r_end--;
     }
     r->sign = a->sign;
+#if 0
     i = r->dgs_used;
+#endif
     while ((*r_end == 0) && (r_end > r->dp))
     {
 	r_end--;
@@ -1557,6 +1563,7 @@ bignum *r;
     {
 	q_end--;
     }
+    /* was ifdef'd out already in original */
 #if 0
     i = m - 1;
     while ((i > 0) && (q->dp[i--] == 0))
