@@ -50,6 +50,8 @@ if test "x$1" = "xinstall"; then
    ./install-sh -m 700 -c samhain ${INSTALLDIR}/_INSTALL_NAME_; 
   else :; 
   fi
+  chown root samhainrc.deploy
+  chmod 0600 samhainrc.deploy
   if test -f remote_bootscript; then
     ./remote_bootscript;
   else :; 
@@ -90,9 +92,11 @@ if test "x$1" = "xpack"; then
   ./mkinstalldirs ${PIDDIR}
   ./mkinstalldirs ${INSTALLDIR}
   if test -f samhain.pk; then 
-   ./install-sh -m 700 -c samhain.pk ${INSTALLDIR}/_INSTALL_NAME_; 
+   ./install-sh -m 700 -o root -c samhain.pk ${INSTALLDIR}/_INSTALL_NAME_; 
   else :; 
   fi
+  chown root samhainrc.deploy
+  chmod 0600 samhainrc.deploy
   if test -f remote_bootscript; then
     ./remote_bootscript;
   else :; 
