@@ -25,6 +25,7 @@
 #include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
+
 /* samhainctl */
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -32,6 +33,7 @@
 #include <signal.h>
 #include <errno.h>
 
+
 #if TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
@@ -74,6 +76,7 @@
 #endif
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
 #include "sh_modules.h"
+#include "sh_ignore.h"
 #endif
 
 #undef  FIL__
@@ -96,11 +99,14 @@ extern UINT32  ErrFlag[2];
  **************************************************/
 
 volatile  int      sig_raised;
-volatile  int      sig_debug_on;           /* SIGUSR1 */
-volatile  int      sig_debug_off;          /* SIGUSR2 */
+volatile  int      sig_urgent;
+volatile  int      sig_debug_switch;       /* SIGUSR1 */
+volatile  int      sig_suspend_switch;     /* SIGUSR2 */
+volatile  int      sh_global_suspend_flag;
 volatile  int      sig_fresh_trail;        /* SIGIOT  */
 volatile  int      sig_config_read_again;  /* SIGHUP  */
 volatile  int      sig_terminate;          /* SIGQUIT */
+volatile  int      sig_termfast;           /* SIGTERM */
 long int           eintr__result;
 char               sh_sig_msg[SH_MINIBUF];
 
@@ -118,16 +124,16 @@ char               sh_sig_msg[SH_MINIBUF];
 #endif
 
 #ifndef GLOB_LEN
-#define GLOB_LEN 255
+#define GLOB_LEN 511
 #endif
 
 char * globber(char * str)
 {
-  int i;
-  int j;
+  size_t i;
+  size_t j;
 
-  static   int  items = 0;
-  static   int  count = 0;
+  static   size_t  items = 0;
+  static   size_t  count = 0;
   static   char glob[SH_MAX_GLOBS * (GLOB_LEN+1)];
 
   if (str == NULL)
@@ -186,7 +192,7 @@ void sh_do_encode (char * str, int len)
  **************************************************/
 
 sh_struct   sh;
-sh_key_t  * skey = NULL;
+/*@null@*/ sh_key_t  * skey = NULL;
 
 extern unsigned char TcpFlag[8][PW_LEN+1];
 
@@ -237,12 +243,14 @@ void sh_init (void)
 
   /* Signal handling.
    */
-  sig_raised            = 0;
-  sig_config_read_again = 0;           /* SIGHUP  */
-  sig_debug_on          = 0;           /* SIGUSR1 */
-  sig_debug_off         = 0;           /* SIGUSR2 */
-  sig_fresh_trail       = 0;           /* SIGIOT  */
-  sig_terminate         = 0;           /* SIGQUIT */
+  sig_raised             = 0;
+  sig_config_read_again  = 0;           /* SIGHUP  */
+  sig_debug_switch       = 0;           /* SIGUSR1 */
+  sig_suspend_switch     = 0;           /* SIGUSR2 */
+  sh_global_suspend_flag = 0;           /* SIGUSR2 */
+  sig_fresh_trail        = 0;           /* SIGIOT  */
+  sig_terminate          = 0;           /* SIGQUIT */
+  sig_termfast           = 0;           /* SIGTERM */
   strcpy ( sh_sig_msg, _("None"));
 
 #ifdef MKB_01
@@ -327,14 +335,12 @@ void sh_init (void)
 
   /* The stats.
    */
-  sh.stat.openfiles    = 0;
-  sh.stat.bytes_hashed = 0;
-  sh.stat.mail_success = 0;
-  sh.stat.mail_failed  = 0;
-  sh.stat.log_success  = 0;
-  sh.stat.log_failed   = 0;
-  sh.stat.time_start   = time(NULL);
-  sh.stat.time_check   = (time_t) 0;
+  sh.statistics.bytes_speed  = 0;
+  sh.statistics.bytes_hashed = 0;
+  sh.statistics.mail_success = 0;
+  sh.statistics.mail_failed  = 0;
+  sh.statistics.time_start   = time(NULL);
+  sh.statistics.time_check   = (time_t) 0;
 
 #ifdef MKC_01
   ErrFlag[0] |= (1 << 16);
@@ -364,7 +370,7 @@ void sh_init (void)
 
   /* The local host.
    */
-  sl_strlcpy (sh.host.name,  _("localhost"),  SH_MINIBUF);
+  (void) sl_strlcpy (sh.host.name,  _("localhost"),  SH_MINIBUF);
   sh.host.system[0]     = '\0';
   sh.host.release[0]    = '\0';
   sh.host.machine[0]    = '\0';
@@ -396,9 +402,9 @@ void sh_init (void)
 
   /* The paths.
    */
-  sl_strlcpy (sh.conf.path,  DEFAULT_CONFIGFILE,    SH_PATHBUF);
+  (void) sl_strlcpy (sh.conf.path,  DEFAULT_CONFIGFILE,    SH_PATHBUF);
   sh.conf.hash[0] = '\0';
-  sl_strlcpy (sh.data.path,  DEFAULT_DATA_FILE,     SH_PATHBUF);
+  (void) sl_strlcpy (sh.data.path,  DEFAULT_DATA_FILE,     SH_PATHBUF);
   sh.data.hash[0] = '\0';
   sh.exec.path[0] = '\0';
   sh.exec.hash[0] = '\0';
@@ -433,13 +439,13 @@ void sh_init (void)
 #if defined(SH_WITH_MAIL)
   if (0 == strcmp (DEFAULT_MAILADDRESS, _("NULL")))
     {
-      sl_strncpy(q, DEFAULT_MAILADDRESS, SH_PATHBUF);
+      (void) sl_strncpy(q, DEFAULT_MAILADDRESS, SH_PATHBUF);
       p = strtok (q, ", \t");
       if (p)
 	{
-	  sh_mail_setaddress_int (p);
+	  (void) sh_mail_setaddress_int (p);
 	  while (NULL != (p = strtok (NULL, ", \t")))
-	    sh_mail_setaddress_int (p);
+	    (void) sh_mail_setaddress_int (p);
 	}
     }
 #endif
@@ -525,19 +531,21 @@ void sh_init (void)
 
   sh.looptime     = 60;
 
-
   /* The struct to hold privileged information.
    */
-  skey = (sh_key_t *) sl_malloc (sizeof(sh_key_t));
+  skey = (sh_key_t *) malloc (sizeof(sh_key_t));
   if (skey == NULL) 
     {
       perror(_("sh_init"));
       _exit (EXIT_FAILURE);
     }
 
-  skey->mlock_failed = GOOD;
+  skey->mlock_failed = SL_FALSE;
   skey->rngI         = BAD;
 
+  for (i = 0; i < KEY_BYT; ++i)
+    skey->poolv[i] = '\0';
+
   skey->poolc        = 0;
 
   skey->ErrFlag[0]   = ErrFlag[0];
@@ -548,12 +556,12 @@ void sh_init (void)
   dez = &(TcpFlag[POS_TF-1][0]);
   for (i = 0; i < PW_LEN; ++i)
     { 
-      skey->pw[i] = (*dez); 
+      /*@i@*/ skey->pw[i] = (*dez); 
       (*dez)      = '\0';
       ++dez; 
     }
 
-  sh_unix_mlock();
+  sh_unix_memlock();
   SL_RET0(_("sh_init"));
 }
 
@@ -562,6 +570,10 @@ void sh_init (void)
 #include <sys/mman.h>
 #endif
 
+#if defined(SH_USE_XML)
+extern int    sh_log_file    (char * message, char * inet_peer);
+#endif
+
 /*******************************************************
  * 
  * Exit Handler
@@ -574,9 +586,16 @@ void exit_handler(void)
 #if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
   int modnum;
 #endif
+#if defined(SH_WITH_SERVER)
+  extern int sh_socket_remove ();
+#endif
 
   SL_ENTER(_("exit_handler"));
 
+#if defined(SH_WITH_SERVER)
+  sh_socket_remove ();
+#endif
+
 #if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
   for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
     {
@@ -615,6 +634,10 @@ void exit_handler(void)
   sh_mem_stat();
 #endif
 
+#ifdef MEM_DEBUG
+  sh_unix_count_mlock();
+#endif
+
   /* --- Checksum of executable. ---
    */
   sh_unix_self_check();
@@ -624,6 +647,9 @@ void exit_handler(void)
    */
   sh_error_handle ((-1), FIL__, __LINE__, sh.flag.exit, MSG_EXIT_NORMAL, 
 		   sh.prg_name, sh_sig_msg);
+#ifdef SH_USE_XML
+  sh_log_file (NULL, NULL);
+#endif
 
 
   /* --- Restrict error logging to stderr. ---
@@ -636,7 +662,8 @@ void exit_handler(void)
 
   /* --- Remove lock, delete critical information. ---
    */
-  sh_unix_unlock ();
+  sh_unix_rm_lock_file (sh.srvlog.name);
+  sh_unix_rm_pid_file ();
   if (skey != NULL)
     memset (skey, '\0', sizeof(sh_key_t));
   
@@ -660,15 +687,15 @@ void exit_handler(void)
 
 #if defined(__linux__) || defined(sun) || defined(__sun) || defined(__sun__)
 #include <dirent.h>
-static int * procdirSamhain ()
+static pid_t * procdirSamhain ()
 {
-  int        * pidlist = malloc(sizeof(int) * 65535);
+  pid_t        * pidlist = malloc(sizeof(pid_t) * 65535);
   struct dirent * d;
   DIR *        dp;
   long         ino;
   struct stat  buf;
-  int          pid, i = 0;
-  pid_t        mypid = getpid();
+  int          i = 0;
+  pid_t        pid, mypid = getpid();
   char       * tail;
   char         exef[128];
 
@@ -687,25 +714,26 @@ static int * procdirSamhain ()
       if (0 != strcmp(d->d_name, ".") && 0 != strcmp(d->d_name, ".."))
 	{
 	  errno = 0;
-	  pid = strtol (d->d_name, &tail, 0);
+	  pid = (pid_t) strtol (d->d_name, &tail, 0);
 	  if (*tail != '\0' || errno != 0)
 	    continue;
 	  if (pid == mypid)
 	    continue;
 #if defined(__linux__) 
-	  sprintf(exef, _("/proc/%d/exe"), pid);         /* known to fit  */
+          sprintf(exef, _("/proc/%d/exe"), (int) pid); /* known to fit  */
 #else
-	  sprintf(exef, _("/proc/%d/object/a.out"), pid);/* known to fit  */
+          sprintf(exef, _("/proc/%d/object/a.out"),    /* known to fit  */
+		  (int) pid);
 #endif
 	  if (0 == stat(exef, &buf) && ino == (long) buf.st_ino)
-	    { pidlist[i] = pid; ++i; }
+	    { pidlist[i] = (pid_t) pid; ++i; }
 	}
     }
   closedir(dp);
   return pidlist;
 }
 #else
-static int * procdirSamhain ()
+static pid_t * procdirSamhain ()
 {
   return NULL;
 }
@@ -716,13 +744,13 @@ static int killprocSamhain (pid_t pid)
   /* fprintf(stderr, "Killing %d\n", pid); */
   if (pid > 0 && 0 == kill (pid, SIGTERM))
     {
-      sleep(1);
+      retry_msleep(1, 0);
       if (0 != kill (pid, 0) && errno == ESRCH)
 	return (0);
-      sleep(2);
+      retry_msleep(2, 0);
       if (0 != kill (pid, 0) && errno == ESRCH)
 	return (0);
-      sleep(2);
+      retry_msleep(2, 0);
       if (0 != kill (pid, 0) && errno == ESRCH)
 	return (0);
       
@@ -741,13 +769,14 @@ static int killprocSamhain (pid_t pid)
     return (7);
 }
 
-static int pidofSamhain (int flag)
+static pid_t pidofSamhain (int flag)
 {
   FILE      * fp;
   char        line[256];
   char      * tail;
   char      * p;
   pid_t       pid;
+  long        inpid;
   struct stat buf;
  
   fp = fopen (DEFAULT_ERRLOCK, "r");
@@ -762,15 +791,19 @@ static int pidofSamhain (int flag)
 	 *p == '\r' || *p == '\t' || *p == '\v')
     ++p;
   errno = 0;
-  pid = strtol (p, &tail, 0);
+  inpid = strtol (p, &tail, 0);
   if (p == tail || errno != 0)
     { perror(_("strtol")); return 0; }
 
+  pid = (pid_t) inpid;
+  if (inpid != (long) pid)
+    { perror(_("strtol")); return 0; }
+
   /* remove stale pid file
    */
   if (flag == 1 && pid > 0 && 0 != kill(pid, 0) && errno == ESRCH)
     {
-      lstat (DEFAULT_ERRLOCK, &buf);
+      /*@i@*/ lstat (DEFAULT_ERRLOCK, &buf);
       if (S_ISREG(buf.st_mode)) unlink(DEFAULT_ERRLOCK);
 
       pid = 0;
@@ -783,14 +816,17 @@ static int pidofSamhain (int flag)
 static int samhainctl(int ctl, int * argc, char * argv[])
 {
   char * fullpath;
-  int    pid;
+  pid_t  pid;
   int    status;
   int    res;
+  pid_t  respid;
   int    times;
   char * argp[32];
   char * envp[2];
-  int       * pidlist;
+  pid_t       * pidlist;
   int         i;
+  char * ctz;
+  int    ltz;
 
 
   fullpath = malloc(strlen(SH_INSTALL_PATH)+1);
@@ -816,17 +852,21 @@ static int samhainctl(int ctl, int * argc, char * argv[])
       if (argp[times-1] == NULL)
 	{ perror(_("malloc")); exit (1); }
       else
-	strcpy(argp[times-1], argv[times]);            /* known to fit  */
+	strcpy(argp[times-1], argv[times]);              /* known to fit  */
     }
 
-  if (getenv("TZ") != NULL)
+  if ((ctz = getenv("TZ")) != NULL)
     {
-      envp[0] = 
-	malloc(strlen(getenv("TZ"))+1);
+      ltz = strlen(ctz) + 4;
+      envp[0] = malloc(ltz);
       if (envp[0] == NULL)
 	{ perror(_("malloc")); exit (1); }
       else
-	strcpy(envp[0], getenv("TZ"));                   /* known to fit  */
+	{
+	  strncpy(envp[0], _("TZ="), ltz);               /* known to fit  */
+	  strncat(envp[0], getenv("TZ"), ltz);           /* known to fit  */
+	  envp[0][ltz-1] = '\0';
+	}
     }
   else
     {
@@ -843,7 +883,7 @@ static int samhainctl(int ctl, int * argc, char * argv[])
 
       pid = fork();
       switch (pid) {
-      case -1:
+      case ((pid_t) -1):
 	perror(_("fork"));
 	exit (1);
       case  0:
@@ -857,13 +897,13 @@ static int samhainctl(int ctl, int * argc, char * argv[])
       default:
 	times = 0;
 	while (times < 10) {
-	  res = waitpid(pid, &status, WNOHANG|WUNTRACED);
-	  if (-1 == res)
+	  respid = waitpid(pid, &status, WNOHANG|WUNTRACED);
+	  if ((pid_t)-1 == respid)
 	    {
 	      perror(_("waitpid"));
 	      exit (1);
 	    }
-	  else if (pid == res)
+	  else if (pid == respid)
 	    {
 	      if (0 != WIFEXITED(status))
 		{
@@ -874,7 +914,7 @@ static int samhainctl(int ctl, int * argc, char * argv[])
 		exit (1);
 	    }
 	  ++times;
-	  sleep(1);
+	  retry_msleep(1, 0);
 	}
 	exit (1);
       }
@@ -939,6 +979,7 @@ static int samhainctl(int ctl, int * argc, char * argv[])
     }
 
   exit (1); /* no exit handler installed yet */
+  return (0); /* notreached */
 }
 
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
@@ -946,31 +987,61 @@ static int samhainctl(int ctl, int * argc, char * argv[])
 static sh_schedule_t * FileSchedOne = NULL;
 static sh_schedule_t * FileSchedTwo = NULL;
 
-sh_schedule_t * sh_set_schedule_int (char * str, sh_schedule_t * FileSched, 
+/* free a linked list of schedules
+ */
+sh_schedule_t *  free_sched (sh_schedule_t * isched)
+{
+  sh_schedule_t * current = isched;
+  sh_schedule_t * next    = NULL;
+
+  while (current != NULL)
+    {
+      next = current->next;
+      SH_FREE(current);
+      current = next;
+    }
+  return NULL;
+}
+
+/* Add a new schedule to the linked list of schedules
+ */
+sh_schedule_t * sh_set_schedule_int (char * str, sh_schedule_t * FileSchedIn, 
 				     int * status)
 {
+  sh_schedule_t * FileSched;
+
   SL_ENTER(_("sh_set_schedule_int"));
-  if (FileSched != NULL)
+
+  if (0 == sl_strncmp(str, _("NULL"), 4))
     {
-      SH_FREE(FileSched);
-      FileSched = NULL;
+      free_sched(FileSchedIn);
+      FileSchedIn = NULL;
+      return 0;
     }
+
   FileSched = SH_ALLOC(sizeof(sh_schedule_t));
   *status = create_sched(str, FileSched);
   if (*status != 0)
     {
       SH_FREE(FileSched);
       FileSched = NULL;
+      SL_RETURN(FileSchedIn , _("sh_set_schedule_int"));
     }
+  FileSched->next = FileSchedIn;
   SL_RETURN(FileSched , _("sh_set_schedule_int"));
 }
 
+/* Add a new schedule to the linked list FileSchedOne
+ */
 int sh_set_schedule_one (char * str)
 {
   int status;
   FileSchedOne = sh_set_schedule_int (str, FileSchedOne, &status);
   return status;
 }
+
+/* Add a new schedule to the linked list FileSchedTwo
+ */
 int sh_set_schedule_two (char * str)
 {
   int status;
@@ -990,9 +1061,8 @@ int main(int argc, char * argv[])
 #if defined(INET_SYSLOG)
   extern int    create_syslog_socket (int flag);
 #endif
-
-#if defined(SH_USE_XML)
-  extern int    sh_log_file    (char * message);
+#if defined(SH_WITH_SERVER)
+  extern int    sh_create_tcp_socket();
 #endif
 
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
@@ -1099,56 +1169,47 @@ int main(int argc, char * argv[])
   ErrFlag[0] |= (1 << 7);
 #endif
 
-  BREAKEXIT(sh_derr);
-  (void) sh_derr();
+#if defined(SCREW_IT_UP)
+  BREAKEXIT(sh_sigtrap_prepare);
+  (void) sh_sigtrap_prepare();
+#endif
 
   /* Save the timezone.
    */
   if (getenv("TZ") != NULL)
     {
       tzlen       = strlen(getenv("TZ"));
-      sh.timezone = sl_malloc (tzlen + 1);
+      sh.timezone = malloc (tzlen + 1);
       if (sh.timezone != NULL)
 	sl_strlcpy (sh.timezone, getenv("TZ"), tzlen + 1);
     }
   else
     sh.timezone = NULL;
-  
 
 
   /* --------  INIT  --------    
    */
 
-  /* drop privileges eventually
-   */
-#if defined(INET_SYSLOG) && defined(SH_WITH_SERVER)
-  create_syslog_socket (S_TRUE);
-  SL_REQUIRE(sl_policy_get_real(DEFAULT_IDENT) == SL_ENONE, 
-	     _("sl_policy_get_real(DEFAULT_IDENT) == SL_ENONE"));
-#else
-  SL_REQUIRE(sl_policy_get_user(DEFAULT_IDENT) == SL_ENONE, 
-	     _("sl_policy_get_user(DEFAULT_IDENT) == SL_ENONE"));
-#endif
-
   /* Restrict error logging to stderr.
    */
   sh_error_only_stderr (S_TRUE);
 
+  BREAKEXIT(sh_derr);
+  (void) sh_derr();
+
   /* Check that first three descriptors are open.
    */
   if ( retry_fcntl(FIL__, __LINE__, 0, F_GETFL, 0) == (-1))
-    aud_open(FIL__, __LINE__, SL_YESPRIV, _("/dev/null"), O_RDWR, 0);
+    (void) aud_open(FIL__, __LINE__, SL_NOPRIV, _("/dev/null"), O_RDWR, 0);
   if ( retry_fcntl(FIL__, __LINE__, 1, F_GETFL, 0) == (-1))
-    aud_open(FIL__, __LINE__, SL_YESPRIV, _("/dev/null"), O_RDWR, 1);
+    (void) aud_open(FIL__, __LINE__, SL_NOPRIV, _("/dev/null"), O_RDWR, 1);
   if ( retry_fcntl(FIL__, __LINE__, 2, F_GETFL, 0) == (-1))
-    aud_open(FIL__, __LINE__, SL_YESPRIV, _("/dev/null"), O_RDWR, 2);
+    (void) aud_open(FIL__, __LINE__, SL_NOPRIV, _("/dev/null"), O_RDWR, 2);
 
   /* --- Set default values. ---
    */
   BREAKEXIT(sh_init);
-  sh_init ();
-
-
+  sh_init ();    /* we are still privileged here, so we can mlock skey */
 #if (defined (SH_WITH_SERVER) && !defined (SH_WITH_CLIENT))
   sh.flag.isserver = S_TRUE;
 #endif
@@ -1173,7 +1234,7 @@ int main(int argc, char * argv[])
 	{
 	  my_argv[0] = argv[0]; ++my_argc;  
 	  command_line[0] = '\0';
-	  fgets (command_line, 255, stdin);
+	  (void*) fgets (command_line, 255, stdin);
 	  command_line[255] = '\0';
 	  do {
 	    my_argv[my_argc] = 
@@ -1197,9 +1258,14 @@ int main(int argc, char * argv[])
   sh.flag.opts = S_FALSE;
   
 
+  /* close all other files
+   */
+  sh_unix_closeall(3);
+
+
   /* --- Get user info. ---
    */
-  TPT((0, FIL__, __LINE__, _("msg=<Get user hostname.>\n")))
+  TPT((0, FIL__, __LINE__, _("msg=<Get user name.>\n")))
   if (0 != sh_unix_getUser ())
     {
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
@@ -1222,26 +1288,21 @@ int main(int argc, char * argv[])
   if (sh.flag.checkSum == SH_CHECK_NONE)
     {
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN,
-		       _("No action specified"), _("main"));
+		       _("No action specified: init, update, or check"), 
+		       _("main"));
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
 		       sh.prg_name);
       aud_exit (FIL__, __LINE__, EXIT_FAILURE);
     }
 #endif
 
-#if defined(INET_SYSLOG) && defined(SH_WITH_SERVER)
-  /* now check whether we really wanted it; if not, close
-   */
-  create_syslog_socket (S_FALSE);
-#endif
-
   /* do not append to database if run SUID
    */
   if (sh.flag.checkSum == SH_CHECK_INIT && sl_is_suid()) 
     {
       dlog(1, FIL__, __LINE__, 
 	   _("Cannot initialize database when running with SUID credentials.\nYou need to run this with the user ID %d.\nYour current user ID is %d."), 
-	   (int) sl_ret_euid(), (int) sh.real.uid);
+	   (int) geteuid(), (int) sh.real.uid);
       sh_error_handle ((-1), FIL__, __LINE__, EACCES, MSG_ACCESS,
 		       (long) sh.real.uid, sh.data.path);
       aud_exit(FIL__, __LINE__, EXIT_FAILURE);
@@ -1250,7 +1311,10 @@ int main(int argc, char * argv[])
   /* avoid daemon mode for initialization 
    */
   if (sh.flag.checkSum == SH_CHECK_INIT)
-    sh.flag.isdaemon = S_FALSE;
+    {
+      sh.flag.isdaemon = S_FALSE;
+      sh.flag.loop     = S_FALSE;
+    }
 
   /* initialize signal handling etc
    */
@@ -1264,6 +1328,36 @@ int main(int argc, char * argv[])
       aud_exit(FIL__, __LINE__, EXIT_FAILURE);
     }
 
+  /* --- drop privileges eventually ---
+   */
+#if defined(SH_WITH_SERVER)
+  sh_create_tcp_socket ();
+#if defined(INET_SYSLOG)
+  create_syslog_socket (S_TRUE);
+#endif
+  SL_REQUIRE(sl_policy_get_real(DEFAULT_IDENT) == SL_ENONE, 
+	     _("sl_policy_get_real(DEFAULT_IDENT) == SL_ENONE"));
+#else
+  SL_REQUIRE(sl_policy_get_user(DEFAULT_IDENT) == SL_ENONE, 
+	     _("sl_policy_get_user(DEFAULT_IDENT) == SL_ENONE"));
+#endif
+
+  /* --- Get user info (again). ---
+   */
+  TPT((0, FIL__, __LINE__, _("msg=<Get user name.>\n")))
+  if (0 != sh_unix_getUser ())
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
+		       sh.prg_name);
+      aud_exit(FIL__, __LINE__, EXIT_FAILURE);
+    }
+
+  /* --- now check whether we really wanted it; if not, close ---
+   */
+#if defined(INET_SYSLOG) && defined(SH_WITH_SERVER)
+  create_syslog_socket (S_FALSE);
+#endif
+
 
   /* checksum of database
    */
@@ -1355,14 +1449,14 @@ int main(int argc, char * argv[])
 #endif
 
  
-  if (skey->mlock_failed == BAD)
+  if (skey->mlock_failed == SL_TRUE)
     sh_error_handle ((-1), FIL__, __LINE__, EPERM, MSG_MLOCK);
 
   /* timer
    */
   tcurrent                   = (unsigned long) time (NULL);
   told                       = tcurrent;
-  sh.mailTime.alarm_last     = tcurrent;
+  sh.mailTime.alarm_last     = told;
 
 
   /****************************************************
@@ -1404,9 +1498,16 @@ int main(int argc, char * argv[])
     {
       if (0 != (status = modList[modnum].mod_init()))
 	{
-	  sh_error_handle ((-1), FIL__, __LINE__, status, MSG_MOD_FAIL,
-			   _(modList[modnum].name),
-			   status);
+	  if (status == (-1)) {
+	    sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, status, 
+			     MSG_MOD_FAIL,
+			     _(modList[modnum].name),
+			     status);
+	  } else {
+	    sh_error_handle ((-1), FIL__, __LINE__, status, MSG_MOD_FAIL,
+			     _(modList[modnum].name),
+			     status);
+	  }
 	  modList[modnum].initval = S_FALSE;
 	}
       else
@@ -1427,17 +1528,18 @@ int main(int argc, char * argv[])
    */
   if (0 != sh.flag.nice)
     {
-      sl_set_suid();
 #ifdef HAVE_SETPRIORITY
-      setpriority(PRIO_PROCESS, 0, sh.flag.nice);
+      /*@i@*/ setpriority(PRIO_PROCESS, 0, sh.flag.nice);
 #else
       nice(sh.flag.nice);
 #endif
-      sl_unset_suid();
     }
 
   /*  --------  MAIN LOOP  ---------
    */
+  sh.statistics.bytes_speed  = 0;
+  sh.statistics.bytes_hashed = 0;
+
   while (1) 
     {
       ++cct;
@@ -1448,11 +1550,26 @@ int main(int argc, char * argv[])
 
       tcurrent                   = (unsigned long) time (NULL);
 
-      if (sig_raised != 0)
+#if 0
+      fprintf(stderr, 
+	      "sig_raised: %d sig_urgent: %d sig_suspend_switch %d sh_global_suspend_flag %d\n",
+	      sig_raised, sig_urgent, sig_suspend_switch, 
+	      sh_global_suspend_flag);
+#endif
+
+      if (sig_raised > 0) 
 	{
 
 	  TPT((0, FIL__, __LINE__, _("msg=<Process a signal.>\n")))
 
+	  if (sig_termfast == 1)  /* SIGTERM */
+	    {
+	      TPT((0, FIL__, __LINE__, _("msg=<Terminate.>\n")));
+	      strncpy (sh_sig_msg, _("SIGTERM"), 20);
+	      --sig_raised; --sig_urgent;
+	      aud_exit (FIL__, __LINE__, EXIT_SUCCESS);
+	    }
+	  
 	  if (sig_config_read_again == 1) /* SIGHUP */
 	    {
 	      TPT((0, FIL__, __LINE__, _("msg=<Re-read configuration.>\n")))
@@ -1468,11 +1585,25 @@ int main(int argc, char * argv[])
 	       */
 	      sh_files_deldirstack ();
 	      sh_files_delfilestack ();
+	      sh_ignore_clean ();
 	      hash_full_tree ();
 
 #if defined(SH_WITH_CLIENT)
 	      reset_count_dev_server();
 #endif
+#if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
+
+
+	      FileSchedOne = free_sched(FileSchedOne);
+	      FileSchedTwo = free_sched(FileSchedTwo);
+
+	      for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
+		{
+		  if (modList[modnum].initval == GOOD)
+		    modList[modnum].mod_reconf();
+		}
+#endif
+
 #if defined(SH_WITH_MAIL)
 	      reset_count_dev_mail();
 #endif
@@ -1487,22 +1618,61 @@ int main(int argc, char * argv[])
 #ifdef RELOAD_DATABASE
 	      sh_hash_hashdelete();
 #endif
-	      sl_trust_add_user(-1);
+	      sl_trust_purge_user();
+	      sh_files_hle_reg (NULL);
 
+	      /* --------------------------
+	       * --- READ CONFIGURATION ---
+	       * --------------------------
+	       */
 	      (void) sh_readconf_read ();
 	      sig_config_read_again = 0;
 	      sh_files_setrec();
 	      sh_files_test_setup();
 	      if (0 != sh.flag.nice)
 		{
-		  sl_set_suid();
 #ifdef HAVE_SETPRIORITY
 		  setpriority(PRIO_PROCESS, 0, sh.flag.nice);
 #else
 		  nice(sh.flag.nice);
 #endif
-		  sl_unset_suid();
 		}
+
+	      if (sh.flag.checkSum == SH_CHECK_INIT)
+		{
+		  sh.flag.isdaemon = S_FALSE;
+		  sh.flag.loop     = S_FALSE;
+		}
+
+	      /* --- Initialize modules. ---
+	       */
+	      TPT((0, FIL__, __LINE__, _("msg=<Initialize modules.>\n")));
+	      for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
+		{
+		  if (0 != (status = modList[modnum].mod_init()))
+		    {
+		      if (status == (-1)) {
+			sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, 
+					 status, MSG_MOD_FAIL,
+					 _(modList[modnum].name),
+					 status);
+		      } else {
+			sh_error_handle ((-1), FIL__, __LINE__, 
+					 status, MSG_MOD_FAIL,
+					 _(modList[modnum].name),
+					 status);
+		      }
+		      modList[modnum].initval = S_FALSE;
+		    }
+		  else
+		    {
+		      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_MOD_OK,
+				       _(modList[modnum].name));
+		      modList[modnum].initval = S_TRUE;
+		    }
+		}
+	      
+	      --sig_raised;
 	    }
 	  
 	  if (sig_fresh_trail == 1) /* SIGIOT */
@@ -1510,41 +1680,56 @@ int main(int argc, char * argv[])
 	      /* Logfile access 
 	       */
 #ifdef SH_USE_XML
-	      sh_log_file (NULL);
+	      sh_log_file (NULL, NULL);
 #endif
-	      TPT((0, FIL__, __LINE__, _("msg=<Logfile stop/restart.>\n")))
+	      TPT((0, FIL__, __LINE__, _("msg=<Logfile stop/restart.>\n")));
 	      sh_error_only_stderr (S_TRUE);
-	      sh_unix_unlock();
-	      sleep(3);
+	      sh_unix_rm_lock_file(sh.srvlog.name);
+	      retry_msleep(3, 0);
 	      sh.flag.log_start = S_TRUE;
 	      sh_error_only_stderr (S_FALSE);
 	      sig_fresh_trail       = 0;
+	      --sig_raised;
 	    }
 	  
 	  if (sig_terminate == 1)  /* SIGQUIT */
 	    {
-	      TPT((0, FIL__, __LINE__, _("msg=<Terminate.>\n")))
+	      TPT((0, FIL__, __LINE__, _("msg=<Terminate.>\n")));
 	      strncpy (sh_sig_msg, _("SIGQUIT"), 20);
+	      --sig_raised; --sig_urgent;
 	      aud_exit (FIL__, __LINE__, EXIT_SUCCESS);
 	    }
 	  
-	  if (sig_debug_on == 1)  /* SIGUSR1 */
+	  if (sig_debug_switch == 1)  /* SIGUSR1 */
 	    {
-	      TPT((0, FIL__, __LINE__, _("msg=<Debug on.>\n")))
-	      sh_error_dbg_on();
-	      sig_debug_on = 0;
+	      TPT((0, FIL__, __LINE__, _("msg=<Debug switch.>\n")));
+	      sh_error_dbg_switch();
+	      sig_debug_switch = 0;
 	    }
 	  
-	  if (sig_debug_off == 1)  /* SIGUSR2 */
+	  if (sig_suspend_switch == 1)  /* SIGUSR2 */
 	    {
-	      TPT((0, FIL__, __LINE__, _("msg=<Debug off.>\n")))
-	      sh_error_dbg_off();
-	      sig_debug_off = 0;
+	      TPT((0, FIL__, __LINE__, _("msg=<Suspend switch.>\n")));
+	      if (sh_global_suspend_flag == 1) {
+		sh_global_suspend_flag = 0;
+	      } else {
+		sh_error_handle((-1), FIL__, __LINE__, 0, MSG_SUSPEND, 
+				sh.prg_name);
+		sh_global_suspend_flag = 1;
+	      }
+	      sig_suspend_switch = 0;
+	      --sig_raised; --sig_urgent;
 	    }
-	  sig_raised = 0;
-	  TPT((0, FIL__, __LINE__, _("msg=<End signal processing.>\n")))
+	  sig_raised = (sig_raised < 0) ? 0 : sig_raised;
+	  sig_urgent = (sig_urgent < 0) ? 0 : sig_urgent;
+	  TPT((0, FIL__, __LINE__, _("msg=<End signal processing.>\n")));
 	}
       
+      if (sh_global_suspend_flag == 1)
+	{
+	  (void) retry_msleep (1, 0);
+	  continue;
+	}
       
       /* see whether its time to check files
        */
@@ -1554,7 +1739,9 @@ int main(int argc, char * argv[])
 	  if (FileSchedTwo != NULL) 
 	    flag_check_2 = 1;
 	}
-      else if (sh.flag.checkSum == SH_CHECK_CHECK)
+      else if (sh.flag.checkSum == SH_CHECK_CHECK || 
+	       (sh.flag.update == S_TRUE && 
+		(sh.flag.isdaemon == S_TRUE || sh.flag.loop == S_TRUE)))
 	{
 	  if (FileSchedOne == NULL)
 	    {
@@ -1581,21 +1768,33 @@ int main(int argc, char * argv[])
 	   * check directories and files
 	   * ORDER IS IMPORTANT -- DIRZ FIRST
 	   */
-	  sh.stat.bytes_hashed  = 0;
-	  sh.stat.time_start    = time (NULL);
+	  sh.statistics.bytes_hashed   = 0;
+	  sh.statistics.time_start     = time (NULL);
+	  sh.statistics.dirs_checked   = 0;
+	  sh.statistics.files_checked  = 0;
 
 	  TPT((0, FIL__, __LINE__, _("msg=<Check directories.>\n")))
 	  BREAKEXIT(sh_dirs_chk);
 	  if (flag_check_1 == 1)
-	    sh.stat.dirs_checked  = sh_dirs_chk  (1); 
+	    {
+	      sh_dirs_chk  (1);
+#ifndef SH_PROFILE
+	      chdir ("/");
+#endif
+	    }
 	  if (flag_check_2 == 1)
-	    sh.stat.dirs_checked  = sh_dirs_chk  (2); 
+	    {
+	      sh_dirs_chk  (2); 
+#ifndef SH_PROFILE
+	      chdir ("/");
+#endif
+	    }
 	  TPT((0, FIL__, __LINE__, _("msg=<Check files.>\n")))
 	  BREAKEXIT(sh_files_chk);
 	  if (flag_check_1 == 1)
-	    sh.stat.files_checked = sh_files_chk ();
+	    sh_files_chk ();
 
-	  if (sig_terminate == 1)
+	  if (sig_urgent == 1)
 	    continue;
 
 	  /*
@@ -1607,38 +1806,34 @@ int main(int argc, char * argv[])
 	      sh_hash_unvisited (ShDFLevel[SH_ERR_T_FILE]);
 	    }
 
-	  if (sig_terminate == 1)
+	  if (sig_urgent == 1)
 	    continue;
 
-	  /* 27.05.2002 avoid empty database
-	   */ 
-	  if      (sh.flag.checkSum == SH_CHECK_INIT)
-	    sh_hash_pushdata (NULL, NULL);
-
 	  /* reset
 	   */
 	  TPT((0, FIL__, __LINE__, _("msg=<Reset status.>\n")))
 	  sh_dirs_reset  ();
-	  if (sig_terminate == 1)
+	  if (sig_urgent == 1)
 	    continue;
 
 	  sh_files_reset ();
 	  flag_check_1 = 0;
 	  flag_check_2 = 0;
 
-	  if (sig_terminate == 1)
+	  if (sig_urgent == 1)
 	    continue;
 
-	  runtim = time(NULL) - sh.stat.time_start;
-
+	  runtim = time(NULL) - sh.statistics.time_start;
+	  sh.statistics.time_check = runtim;
 	
-	  if (sh.stat.dirs_checked == 0 && sh.stat.files_checked == 0)
+	  if ((sh.statistics.dirs_checked == 0) && 
+	      (sh.statistics.files_checked == 0))
 	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_CHECK_0);
 
 	  else
 	    {
-	      st_1 = sh.stat.bytes_hashed;
-	      st_2 = runtim;
+	      st_1 = (float) sh.statistics.bytes_hashed;
+	      st_2 = (float) runtim;
 
 
 	      if (st_1 > 0.0 && st_2 > 0.0) 
@@ -1647,15 +1842,16 @@ int main(int argc, char * argv[])
 		st_1 = st_1 * 1.0;
 	      else
 		st_1 = 0.0;
-	      st_1 = 0.001 * st_1;
-	       
+
+	      sh.statistics.bytes_speed = (unsigned long) st_1;
+
 	      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_CHECK_1,
 			       (long) runtim, 
-			       st_1);
+			       0.001 * st_1);
 	    }
 	  sh.fileCheck.alarm_last = (unsigned long) time (NULL);
 
-	  if (sig_terminate == 1)
+	  if (sig_urgent == 1)
 	    continue;
 
 	  /*
@@ -1667,7 +1863,7 @@ int main(int argc, char * argv[])
 #endif
 	}
       
-      if (sig_terminate == 1)
+      if (sig_urgent == 1)
 	continue;
       
       /* execute modules
@@ -1682,11 +1878,17 @@ int main(int argc, char * argv[])
 			       _(modList[modnum].name), (long) status);
 	}
       
+      /* 27.05.2002 avoid empty database
+       * 22.10.2002 moved here b/o suid check initialization
+       */ 
+      if      (sh.flag.checkSum == SH_CHECK_INIT)
+	sh_hash_pushdata (NULL, NULL);
+
       /* no loop if not daemon
        */
       if (sh.flag.isdaemon != S_TRUE && sh.flag.loop == S_FALSE)
 	break; 
-      if (sig_terminate == 1)
+      if (sig_urgent == 1)
 	continue;
       
       /* see whether its time to send mail
@@ -1699,7 +1901,7 @@ int main(int argc, char * argv[])
 	  sh.mailTime.alarm_last = (unsigned long) time (NULL);
 	}
 #endif
-      if (sig_terminate == 1)
+      if (sig_urgent == 1)
 	continue;
             
       /* log the timestamp
@@ -1710,6 +1912,7 @@ int main(int argc, char * argv[])
 	  told = tcurrent;
 #ifdef MEM_DEBUG
 	  sh_mem_check();
+	  sh_unix_count_mlock();
 #else
 	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_STAMP);
 #endif
@@ -1719,12 +1922,12 @@ int main(int argc, char * argv[])
        */
       (void) taus_seed();
       
-      if (sig_terminate == 1)
+      if (sig_urgent == 1)
 	continue;
       
       /* go to sleep
        */
-      (void) sleep (1);
+      (void) retry_msleep (1, 0);
 
       BREAKEXIT(sh_derr);
       (void) sh_derr();
