@@ -62,7 +62,7 @@
 
 /* Buffer for privileged content.
  */
-#define PRIV_MAX  512    
+#define PRIV_MAX  4096   
 
 #undef  GOOD
 #define GOOD  1
@@ -174,16 +174,14 @@ typedef struct {
 } sh_sh_remote;
 
 typedef struct {
-  unsigned long   openfiles;
-  unsigned long   bytes_hashed;
-  unsigned long   mail_success;
-  unsigned long   mail_failed;
-  unsigned long   log_success;
-  unsigned long   log_failed;
-  time_t          time_start;
-  time_t          time_check;
-  unsigned long   dirs_checked;
-  unsigned long   files_checked;
+  unsigned long   bytes_hashed;  /* bytes     last check */
+  unsigned long   bytes_speed;   /* bytes/sec last check */
+  unsigned long   mail_success;  /* mails sent           */ 
+  unsigned long   mail_failed;   /* mails not sent       */
+  time_t          time_start;    /* start     last check */
+  time_t          time_check;    /* time      last check */
+  unsigned long   dirs_checked;  /* #dirs     last check */
+  unsigned long   files_checked; /* #files    last check */
 } sh_sh_stat;
 
 typedef struct {
@@ -227,7 +225,7 @@ typedef struct {
   sh_sh_remote srvcons;
   sh_sh_remote srvlog;
 
-  sh_sh_stat   stat;
+  sh_sh_stat   statistics;
   sh_sh_flag   flag;
 
 #ifdef SH_STEALTH
@@ -244,11 +242,14 @@ typedef struct {
 
 
 extern volatile  int      sig_raised;
-extern volatile  int      sig_debug_on;           /* SIGUSR1 */
-extern volatile  int      sig_debug_off;          /* SIGUSR2 */
+extern volatile  int      sig_urgent;
+extern volatile  int      sig_debug_switch;       /* SIGUSR1 */
+extern volatile  int      sig_suspend_switch;     /* SIGUSR2 */
+extern volatile  int      sh_global_suspend_flag; /* SIGUSR2 */
 extern volatile  int      sig_fresh_trail;        /* SIGIOT  */
 extern volatile  int      sig_config_read_again;  /* SIGHUP  */
 extern volatile  int      sig_terminate;          /* SIGQUIT */
+extern volatile  int      sig_termfast;           /* SIGTERM */
 
 extern long int eintr__result;
 
@@ -256,6 +257,7 @@ extern long int eintr__result;
 
 
 typedef struct {
+  char   sh_sockpass[32];
   char   sigkey_old[KEY_LEN+1];
   char   sigkey_new[KEY_LEN+1];
   char   mailkey_old[KEY_LEN+1];
@@ -284,14 +286,14 @@ typedef struct {
 } sh_key_t;
 
 extern sh_struct sh; 
-extern sh_key_t  *skey; 
+/*@null@*/ extern sh_key_t  *skey; 
 
 
 #if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
 #define MLOCK(a, b) \
-      if (skey && skey->mlock_failed == GOOD){ \
+      if (skey && skey->mlock_failed == SL_FALSE){ \
         sl_set_suid(); \
-	mlock( a, b); \
+	if (sh_unix_mlock( a, b) < 0) skey->mlock_failed = SL_TRUE; \
         sl_unset_suid(); } 
 #else
 #define MLOCK(a, b) \
@@ -300,9 +302,9 @@ extern sh_key_t  *skey;
 
 #if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
 #define MUNLOCK(a, b) \
-      if (skey && skey->mlock_failed == GOOD){ \
+      if (skey && skey->mlock_failed == SL_FALSE){ \
         sl_set_suid(); \
-	munlock( a, b );\
+	sh_unix_munlock( a, b );\
         sl_unset_suid(); } 
 #else
 #define MUNLOCK(a, b) \
@@ -314,7 +316,14 @@ void sh_do_encode (char * str, int len);
 #define sh_do_decode sh_do_encode
 #endif
 
-#if defined(SCREW_IT_UP) && defined(__linux__) && defined(__i386__)
+/* #if defined(SCREW_IT_UP)
+ * extern volatile int sh_not_traced;
+ * inline int  sh_sigtrap_prepare();
+ * inline int  sh_derr();
+ * #endif
+ */
+
+#if defined(SCREW_IT_UP) && (defined(__FreeBSD__) || defined(__linux__)) && defined(__i386__)
 #define BREAKEXIT(expr) \
   do { \
     int ixi; \
