@@ -18,14 +18,22 @@
  *   "someString" (in quotes !).
  */
 
+#define _(string) string
 #include "config.h" 
+
+/* define if this is a 2.6 kernel                 */
+/* #define LINUX26                                */
+
 #define MAGIC_HIDE SH_MAGIC_HIDE
 
-/*  #define MAGIC_HIDE "someString"              */
+/*  #define MAGIC_HIDE "someString"               */
 
-/* define this if you have a modversioned kernel */
+/* define this if you have a modversioned kernel  */
 /*  #define MODVERSIONS                           */
 
+/* the address of the sys_call_table (not exported in 2.5 kernels) */
+#define MAGIC_ADDRESS SH_SYSCALLTABLE
+
 /*
  * Install:
  * -------
@@ -50,7 +58,7 @@
  * 
  * Tested on:
  * ---------
- *   Linux 2.2, 2.4
+ *   Linux 2.2, 2.4, 2.6
  *
  * Copyright:
  * ---------
@@ -84,8 +92,10 @@
 /* This is a Linux Loadable Kernel Module.
  */
 
+#ifndef LINUX26
 #define __KERNEL__
 #define MODULE
+#endif
 #define LINUX
 
 /* Define for debugging.   
@@ -95,6 +105,7 @@
 /* #define READ_DEBUG  */   /* read         */
 /* #define PROC_DEBUG  */   /* procfs       */
 
+
 /*****************************************************
  *
  *  The include files:
@@ -106,9 +117,16 @@
  */
 #include <linux/config.h>
 
+#ifndef LINUX26
 #ifdef CONFIG_MODVERSIONS
 #include <linux/modversions.h>
 #endif
+#endif
+
+
+#ifdef LINUX26
+#include <linux/init.h>
+#endif
 
 #include <linux/module.h>
 
@@ -121,7 +139,13 @@
 
 /* Include the SYS_syscall defines.
  */
+#ifndef LINUX26
 #include <sys/syscall.h>
+#else
+#define SYS_getdents 141
+#define SYS_getdents64 220
+#endif
+
 
 /* Includes for 'getdents' per the manpage.
  */
@@ -147,9 +171,16 @@
  *
  *****************************************************/
 
-/* The kernel syscall table.
+/* The kernel syscall table. Not exported anymore in 2.5 ff., and also
+ * not in the RedHat 2.4 kernel.
  */
-extern void *sys_call_table[];
+
+#if 0
+extern void * sys_call_table[];
+#define sh_sys_call_table sys_call_table
+#endif
+
+unsigned long * sh_sys_call_table = (unsigned long *) MAGIC_ADDRESS;
 
 /* The old address of the sys_getdents syscall.
  */
@@ -170,10 +201,33 @@ char hidden[] = MAGIC_HIDE;
 
 MODULE_AUTHOR("Rainer Wichmann");
 MODULE_DESCRIPTION("Hide files/processes/modules with MAGIC_HIDE in name.");
+#if defined(MODULE_LICENSE) || defined(LINUX26)
+MODULE_LICENSE("GPL");  
+#endif
+
+#ifdef LINUX26
+/* Default is to hide ourselves.
+ */
+static int removeme = 1;
+
+MODULE_PARM (removeme, "i");
+#endif
 
 
+/* 
+ *  struct task_struct is defined in linux/sched.h
+ *
+ *  as of 2.4.20, the vanilla kernel holds (among others):
+ *        struct task_struct *next_task, *prev_task;
+ *
+ *  Redhat kernel seems to have a different scheduler.
+ *  use:
+ *        struct task_struct * find_task_by_pid (int pid);
+ */
 
-/* Fetch the task struct for a given PID.
+#if defined(SH_VANILLA_KERNEL) && !defined(LINUX26)
+/*
+ * Fetch the task struct for a given PID.
  */
 struct task_struct * fetch_task_struct (int pid)
 {
@@ -200,6 +254,17 @@ struct task_struct * fetch_task_struct (int pid)
   return (NULL);
 }
 
+#else
+/*
+ *  RedHat 2.4.20 kernel
+ */
+struct task_struct * fetch_task_struct (int pid)
+{
+  struct task_struct * task_ptr = NULL;
+  task_ptr = find_task_by_pid (pid);
+  return (task_ptr);
+}
+#endif
 
 /* Convert a string to an int. 
  * Does not recognize integers with a sign (+/-) in front.
@@ -250,9 +315,11 @@ int new_getdents (unsigned int fd, struct dirent *dirp, unsigned int count)
   int                  status = 0;    /* Return value from original getdents */
   struct inode       * dir_inode;
   int                  dir_is_proc = 0;
+
   struct dirent      * dirp_prev;
   struct dirent      * dirp_new;
   struct dirent      * dirp_current;
+
   int                  dir_table_bytes;
   int                  forward_bytes;
   struct task_struct * task_ptr;
@@ -286,15 +353,19 @@ int new_getdents (unsigned int fd, struct dirent *dirp, unsigned int count)
 
   /* Check for the /proc directory
    */
-  if (dir_inode->i_ino == PROC_ROOT_INO && 
-      !MAJOR(dir_inode->i_dev) && 
-      MINOR(dir_inode->i_dev) == 1)
+  if (dir_inode->i_ino == PROC_ROOT_INO 
+#ifndef LINUX26
+      && !MAJOR(dir_inode->i_dev) && 
+      MINOR(dir_inode->i_dev) == 1
+#endif
+      )
     dir_is_proc = 1;
 
   /* Allocate space for new dirent table. Can't use GFP_KERNEL 
    * (kernel oops)
    */
   dirp_new = (struct dirent *) kmalloc (status, GFP_ATOMIC);
+
   if (dirp_new == NULL)
     {
       unlock_kernel();
@@ -415,6 +486,7 @@ int new_getdents (unsigned int fd, struct dirent *dirp, unsigned int count)
   return (status);
 }
 
+
 /* For 2.4 kernel
  */
 #ifdef __NR_getdents64
@@ -424,9 +496,11 @@ long new_getdents64 (unsigned int fd, struct dirent64 *dirp,
   long                 status = 0;    /* Return value from original getdents */
   struct inode       * dir_inode;
   int                  dir_is_proc = 0;
+
   struct dirent64    * dirp_prev;
   struct dirent64    * dirp_new;
   struct dirent64    * dirp_current;
+
   int                  dir_table_bytes;
   int                  forward_bytes;
   struct task_struct * task_ptr;
@@ -464,10 +538,13 @@ long new_getdents64 (unsigned int fd, struct dirent64 *dirp,
 
   /* Check for the /proc directory
    */
-  if (dir_inode->i_ino == PROC_ROOT_INO  && 
-      !MAJOR(dir_inode->i_dev) /*  && 
+  if (dir_inode->i_ino == PROC_ROOT_INO
+#ifndef LINUX26  
+      && !MAJOR(dir_inode->i_dev) /*  && 
       MINOR(dir_inode->i_dev) == 1 */
-      /* MINOR commented out because of problems with 2.4.17 */)
+      /* MINOR commented out because of problems with 2.4.17 */
+#endif
+      )
     {
       dir_is_proc = 1;
 
@@ -608,10 +685,28 @@ long new_getdents64 (unsigned int fd, struct dirent64 *dirp,
 }
 #endif
 
+#ifdef LINUX26
+static struct module *find_module(const char *name)
+{
+        struct module *mod;
+	struct list_head * modules = (struct list_head *) SH_LIST_MODULES;
+
+        list_for_each_entry(mod, modules, list) {
+                if (strcmp(mod->name, name) == 0)
+                        return mod;
+        }
+        return NULL;
+}
+#endif
+
 /* The initialisation function. Automatically called when module is inserted
  * via the 'insmod' command.
  */
+#ifdef LINUX26
+static int __init samhain_hide_init(void)
+#else
 int init_module(void)
+#endif
 {
 
   lock_kernel();
@@ -619,17 +714,34 @@ int init_module(void)
   /* Unfortunately this does not fully prevent the module from appearing
    * in /proc/ksyms. 
    */
-
+#ifndef LINUX26
   EXPORT_NO_SYMBOLS;
+#endif
 
   /* Replace the 'sys_getdents' syscall with the new version.
    */
-  old_getdents                     = sys_call_table[SYS_getdents];
-  sys_call_table[SYS_getdents]     = new_getdents;
+  old_getdents                        = (void*) sh_sys_call_table[SYS_getdents];
+  sh_sys_call_table[SYS_getdents]     = (unsigned long) new_getdents;
   
 #ifdef __NR_getdents64
-  old_getdents64                   = sys_call_table[SYS_getdents64];
-  sys_call_table[SYS_getdents64]   = new_getdents64;
+  old_getdents64                      = (void*) sh_sys_call_table[SYS_getdents64];
+  sh_sys_call_table[SYS_getdents64]   = (unsigned long) new_getdents64;
+#endif
+
+#ifdef LINUX26
+  {
+    spinlock_t * modlist_lock = (spinlock_t * ) SH_MODLIST_LOCK;
+    struct module *mod = find_module(SH_INSTALL_NAME"_hide");
+    if (mod) {
+      /* Delete from various lists */
+      spin_lock_irq(modlist_lock);
+      if (removeme == 1)
+	{
+	  list_del(&mod->list);
+	}
+      spin_unlock_irq(modlist_lock);
+    }
+  }
 #endif
 
   unlock_kernel();
@@ -639,20 +751,27 @@ int init_module(void)
 /* The cleanup function. Automatically called when module is removed
  * via the 'rmmod' command.
  */
+#ifdef LINUX26
+static void __exit samhain_hide_cleanup(void)
+#else
 void cleanup_module(void)
+#endif
 {
   lock_kernel();
 
   /* Restore the new syscalls to the original version.
    */
-  sys_call_table[SYS_getdents]     = old_getdents;
+  sh_sys_call_table[SYS_getdents]     = (unsigned long) old_getdents;
 #ifdef __NR_getdents64
-  sys_call_table[SYS_getdents64]   = old_getdents64;
+  sh_sys_call_table[SYS_getdents64]   = (unsigned long) old_getdents64;
 #endif
 
   unlock_kernel();
 }
 
-
+#ifdef LINUX26
+module_init(samhain_hide_init);
+module_exit(samhain_hide_cleanup);
+#endif
 
 
