@@ -169,7 +169,7 @@ int sh_entropy(int getbytes, char * nbuf)
 
         memset( &addr, 0, sizeof(addr) );
         addr.sun_family = AF_UNIX;
-        strcpy( addr.sun_path, name );
+        strcpy( addr.sun_path, name );              /* known to fit  */
         addr_len = offsetof( struct sockaddr_un, sun_path )
                    + strlen( addr.sun_path );
 
@@ -277,19 +277,12 @@ int sh_entropy(int getbytes, char * nbuf)
 
 #include <setjmp.h>
 
-static jmp_buf entropy_timeout;
+static jmp_buf          entropy_timeout;
 
 static void sh_entropy_alarmhandle (int mysignal)
 {
-  struct  sigaction  def_act;
-  def_act.sa_handler = SIG_IGN;
-
-  if (mysignal == SIGALRM)   /* use 'signal' to avoid compiler warning */
-    {  
-      alarm(0);
-      sigaction (SIGALRM, &def_act, NULL);
-      longjmp(entropy_timeout, 1);
-    }
+  (void) mysignal; /* avoid compiler warning */
+  longjmp (entropy_timeout, 1);
 }
 
 
@@ -297,11 +290,12 @@ int read_mbytes(int timeout_val, char * path, char * nbuf, int nbytes)
 {
   int count, m_count;
   int fd2;
-  struct  sigaction  def_act;
+
   struct  sigaction  new_act;
   sigset_t           unblock;
 
-  def_act.sa_handler = SIG_IGN;
+  struct sigaction      old_act;
+  volatile unsigned int old_alarm = 0;
 
   new_act.sa_handler = sh_entropy_alarmhandle;
   sigemptyset( &new_act.sa_mask );       /* set an empty mask       */
@@ -325,7 +319,8 @@ int read_mbytes(int timeout_val, char * path, char * nbuf, int nbytes)
 	  if (setjmp(entropy_timeout) != 0)
 	    {
 	      alarm(0);
-	      sigaction (SIGALRM, &def_act, NULL);
+	      sigaction (SIGALRM, &old_act, NULL);
+	      alarm(old_alarm);
 	      sigprocmask(SIG_UNBLOCK, &unblock, NULL);
 	      TPT((0,FIL__,__LINE__, _("msg=<read_mbytes: timeout>\n"))); 
 	      close (fd2);
@@ -334,8 +329,8 @@ int read_mbytes(int timeout_val, char * path, char * nbuf, int nbytes)
 
 	  /* timeout after 30 seconds
 	   */
-	  alarm(0);
-	  sigaction (SIGALRM, &new_act, &def_act);
+	  old_alarm = alarm(0);
+	  sigaction (SIGALRM, &new_act, &old_act);
 	  alarm(timeout_val);
 
 	  m_count = 0;
@@ -364,6 +359,11 @@ int read_mbytes(int timeout_val, char * path, char * nbuf, int nbytes)
                 }
 	    }
 	  close (fd2);
+
+	  alarm(0);
+	  sigaction (SIGALRM, &old_act, NULL);
+	  alarm(old_alarm);
+	  sigprocmask(SIG_UNBLOCK, &unblock, NULL);
 	}
       else
 	m_count = 0;
@@ -371,11 +371,7 @@ int read_mbytes(int timeout_val, char * path, char * nbuf, int nbytes)
   else
     m_count = 0;
 
-
-  alarm(0);
   TPT((0, FIL__, __LINE__, _("msg=<read_mbytes: OK>\n"))); 
-  sigaction (SIGALRM, &def_act, NULL); 
-  
   SL_RETURN(m_count, _("read_mbytes"));
 }
 
@@ -405,9 +401,15 @@ int sh_entropy(int nbytes, char * nbuf)
 #endif
 
   if (m_count == 0)
-    sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_NODEV, 
-		     (long) sh.real.uid, NAME_OF_DEV_RANDOM);
-
+    {
+#ifdef NAME_OF_DEV_URANDOM
+      sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, EIO, MSG_NODEV, 
+		       (long) sh.real.uid, NAME_OF_DEV_RANDOM);
+#else
+      sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_NODEV, 
+		       (long) sh.real.uid, NAME_OF_DEV_RANDOM);
+#endif
+    }
 
 #ifdef NAME_OF_DEV_URANDOM
   if (m_count < nbytes)
@@ -432,7 +434,7 @@ int sh_entropy(int nbytes, char * nbuf)
       for (i = 0; i < KEY_BYT; ++i)
 	addbuf[i+KEY_BYT] = skey->poolv[i];
       keybuf = (char *) sh_tiger_hash_uint32 (addbuf, 
-					      TIGER_DATA, sizeof(addbuf));
+					      TIGER_DATA, 2 * KEY_BYT);
       memset(addbuf, '\0', sizeof(addbuf));
       
       /* -- Give out nbytes bytes from the new pool. --
@@ -471,6 +473,8 @@ int sh_entropy(int nbytes, char * nbuf)
 #define FD_ZERO(p)      memset((char *)(p), '\0', sizeof(*(p)))
 #endif
 
+#include "sh_static.h"
+
 static
 char   * com_path[] = {
   N_("/usr/ucb/"),
@@ -563,7 +567,6 @@ sourcetable_t source[] = {
 };
 
 
-
 static FILE * sh_popen (sourcetable_t  *source, char * command)
 {
   int i;
@@ -582,7 +585,7 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
 
   if (sh.timezone != NULL)
     {
-      envp[0] = sl_malloc (sl_strlen(sh.timezone) + 4);
+      envp[0] = malloc (sl_strlen(sh.timezone) + 4);     /* free() ok     */
       if (envp[0] != NULL)
 	sprintf (envp[0], "TZ=%s", sh.timezone);         /* known to fit  */
       else
@@ -598,7 +601,7 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
   /* Create the pipe 
    */
   if (aud_pipe(FIL__, __LINE__, pipedes) < 0) {
-    if (envp[0] != NULL) my_free(envp[0]);
+    if (envp[0] != NULL) free(envp[0]);
     SL_RETURN(NULL, _("sh_popen"));
   }
   
@@ -609,7 +612,7 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
   if (source->pid == (pid_t) - 1) {
     close(pipedes[0]);
     close(pipedes[1]);
-    if (envp[0] != NULL) my_free(envp[0]);
+    if (envp[0] != NULL) free(envp[0]);
     SL_RETURN(NULL, _("sh_popen"));
   }
 
@@ -639,9 +642,11 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
        */
       i = 0; 
       if (0 == geteuid()) {  
-	tempres = getpwnam(DEFAULT_IDENT);
+	tempres = sh_getpwnam(DEFAULT_IDENT);
 	if (NULL != tempres) {
-	  i = aud_setgid(FIL__, __LINE__, tempres->pw_gid);  
+	  i = aud_setgid(FIL__, __LINE__, tempres->pw_gid); 
+	  if (i == 0)
+	    i = sh_unix_initgroups(DEFAULT_IDENT ,tempres->pw_gid);
 	  if (i == 0) 
 	    i = aud_setuid(FIL__, __LINE__, tempres->pw_uid);
 	  /* make sure we cannot get root again
@@ -672,12 +677,12 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
     /* parent
      */
     if (envp[0] != NULL) 
-      my_free(envp[0]);
+      free(envp[0]);
 
     close (pipedes[STDOUT_FILENO]);
     retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
 
-    MBLK( outf = fdopen (pipedes[STDIN_FILENO], "r"); )
+    outf = fdopen (pipedes[STDIN_FILENO], "r");
 
     if (outf == NULL) 
       {
@@ -698,7 +703,7 @@ static int sh_pclose (sourcetable_t *source)
 
     SL_ENTER(_("sh_pclose"));
 
-    MBLK( status = fclose(source->pipe) )
+    status = fclose(source->pipe);
     if (status)
       {
 	SL_RETURN((-1), _("sh_pclose"));
@@ -712,7 +717,7 @@ static int sh_pclose (sourcetable_t *source)
     SL_RETURN(status, _("sh_pclose"));
 }
 
-#define BUF_ENT 8190
+#define BUF_ENT 32766
 
 /* Poll the system for randomness, mix results with 
  * previous reads using a hash function, and give out
@@ -848,10 +853,14 @@ int sh_entropy(int nbytes, char * nbuf)
 		       source[i].pipe );
 	if (count == 0) 
 	  {
+	    if (0 != feof(source[i].pipe))
+	      sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_ENCLOS,
+			       (long) source[i].pipeFD);
+	    else
+	      sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_ENCLOS1,
+			       (long) source[i].pipeFD);
 	    source[i].isset = 0;
 	    sh_pclose ( &source[i] );
-	    sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_ENCLOS,
-			     (long) source[i].pipeFD);
 	    --icount;
 	  }
 	else
