@@ -21,6 +21,8 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <sys/types.h>
+#include <unistd.h>
 
 #include "samhain.h"
 #include "sh_error.h"
@@ -108,10 +110,9 @@ int get_key_from_file(char * path, char * keyid, char * key)
 	}
     }
 	  
-  /* not found
+  /* notreached
    */
-  sl_close(fd);
-  return -1;
+  /* return -1; */
 }
 
 static int just_list = S_FALSE;
@@ -137,6 +138,7 @@ int sh_error_logverify (char * s)
 #ifdef SH_USE_XML
   char * ptr;
   int fixed_xml = S_TRUE;
+  char c_start;
 #endif
   char signature[64];
   char key[KEY_LEN+1];
@@ -144,7 +146,6 @@ int sh_error_logverify (char * s)
   char timestamp[64];
   char c_cont;
   int  chk_mode = CHK_KEY;
-  char c_start;
 
   sh_error_logoff();
 
@@ -156,6 +157,12 @@ int sh_error_logverify (char * s)
 
   /* Open the file, then check it. 
    */
+  if (sl_is_suid())
+    {
+      fprintf(stderr, _("Cannot open file %s in suid mode\n"), s);
+      _exit (EXIT_FAILURE);
+    }
+
   if ( SL_ISERROR(fd = sl_open_read (s, SL_NOPRIV)) )
     {
       fprintf(stderr, 
@@ -170,9 +177,11 @@ int sh_error_logverify (char * s)
   c_cont ^= XOR_CODE;
 #endif
 
+#ifdef SH_USE_XML
   c_start  = '<';
 #ifdef SH_STEALTH
   c_start ^= XOR_CODE;
+#endif
 #endif
 
   buf  = (char *) SH_ALLOC( 2*SH_BUFSIZE+1 );
@@ -493,106 +502,240 @@ int sh_error_logverify (char * s)
  *
  ********************************************************************/
 
-/*
- *   --- Log error message to log file. ---
- */
-int  sh_log_file (char *errmsg)
+int sh_log_open (char * inet_peer, 
+                 char * logfile, int * service_failure, SL_TICKET * fildesc)
 {
-  int                  store1;
-  int                  store2;
   SL_TICKET            fd = -1;
   long int             status;
   char               * tmp = NULL;
-  struct _sh_log_buf   log_msg;
   uid_t                uid;
-  static int           service_failure = 0;
-
-
-  SL_ENTER(_("sh_log_file"));
-
-  if (errFlags.HaveLog == BAD)  /* paranoia */ 
-    SL_RETURN((-1), _("sh_log_file"));
+  int                  i;
+  char               * lockfile = NULL;
 
   /* open/create the file, then check it 
    */
-  tmp  = sh_util_safe_name (sh.srvlog.name);
 
-  if (  0 !=  (status = tf_trust_check (sh.srvlog.name, SL_YESPRIV))
-	&& service_failure == 0)
-    sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_TRUST,
-		      (long) sh.effective.uid, tmp);
+  if (  0 !=  (status = tf_trust_check (logfile, SL_YESPRIV))
+        && (*service_failure) == 0)
+    {
+      tmp  = sh_util_safe_name (logfile);
+      sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_TRUST,
+                      (long) sh.effective.uid, tmp);
+    }
 
   if (status == 0)
     {
-      fd = sl_open_write (sh.srvlog.name, SL_YESPRIV);
+      fd = sl_open_write (logfile, SL_YESPRIV);
       if (SL_ISERROR(fd))
-	{
+        {
+	  tmp  = sh_util_safe_name (logfile);
 	  sl_get_euid(&uid);
-	  if (service_failure == 0)
-	    sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_ACCESS,
-			     (long) uid, tmp);
-	  status = -1;
-	}
+          if ((*service_failure) == 0)
+            sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_ACCESS,
+                             (long) uid, tmp);
+          status = -1;
+        }
     }
 
 
-  if (status == 0)
+  if (status == 0 && inet_peer == NULL )
     {
-      status = sh_unix_testlock(sh.srvlog.name);
+      status = sh_unix_write_lock_file(logfile);
       if (status < 0)
-	{
-	  sl_get_euid(&uid);
-	  if (service_failure == 0)
-	    sh_error_handle ((-1), FIL__, __LINE__, status, MSG_LOCKED,
-			     (long) uid, tmp, sh.srvlog.alt);
-	  status = -1;
-	  sl_close(fd);
-	}
+        {
+	  tmp  = sh_util_safe_name (logfile);
+	  i        = 6 + sl_strlen(tmp);
+	  lockfile = SH_ALLOC(i);
+	  sl_strlcpy(lockfile,        tmp, i);
+	  sl_strlcat(lockfile, _(".lock"), i);
+          sl_get_euid(&uid);
+          if ((*service_failure) == 0)
+            sh_error_handle ((-1), FIL__, __LINE__, status, MSG_LOCKED,
+                             (long) uid, tmp, lockfile);
+          status = -1;
+	  SH_FREE(lockfile);
+          sl_close(fd);
+        }
     }
 
   if (status == 0)
     {
       status = SL_ISERROR(sl_forward(fd) ); 
       if (status < 0)
-	{
-	  sl_get_euid(&uid);
-	  if (service_failure == 0)
-	    sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_ACCESS,
-			     (long) uid, tmp);
-	  status = -1;
-	  sl_close(fd);
-	}
+        {
+	  tmp  = sh_util_safe_name (logfile);
+          sl_get_euid(&uid);
+          if ((*service_failure) == 0)
+            sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_ACCESS,
+                             (long) uid, tmp);
+          status = -1;
+          sl_close(fd);
+        }
     }
   
   if (status < 0)
     {
-      if (service_failure == 0) {
-	sh_error_handle ((-1), FIL__, __LINE__, status, MSG_SRV_FAIL,
-			 _("logfile"), tmp);
-	service_failure = 1;
+      if ((*service_failure) == 0) {
+        sh_error_handle ((-1), FIL__, __LINE__, status, MSG_SRV_FAIL,
+                         _("logfile"), tmp);
+        (*service_failure) = 1;
       }
       SH_FREE(tmp);
-      SL_RETURN ((-1), _("sh_log_file"));
+      return -1;
     }
 
-  service_failure = 0;
+  *fildesc         = fd;
+  *service_failure = 0;
+  return 0;
+}
+
+typedef struct lfstc {
+  char          * logfile;
+  int             service_failure;
+  int             log_start;
+  char            sigkey_old[KEY_LEN+1];
+  char            sigkey_new[KEY_LEN+1];
+  char            crypt[KEY_LEN+1]; 
+  struct  lfstc * next;
+} open_logfile;
+
+static open_logfile * logfile_list = NULL;
+
+static int flag_sep_log = S_FALSE;
+
+#ifdef SH_WITH_SERVER
+int set_flag_sep_log (char * str)
+{
+  return sh_util_flagval(str, &flag_sep_log);
+}
+#endif
+
+/*
+ *   --- Log error message to log file. ---
+ */
+int  sh_log_file (char *errmsg, char * inet_peer)
+{
+  int                  store1;
+  int                  store2;
+  int                  store3;
+  int                  store4;
+  int                  store5;
+  int                  store6;
+  int                  store7;
+  int                  store8;
+
+  SL_TICKET            fd = -1;
+  long int             status;
+  struct _sh_log_buf   log_msg;
+
+  char                 logfile[SH_PATHBUF+SH_MINIBUF+2];
+  open_logfile       * current = logfile_list;  
+  open_logfile       * next    = NULL;
+  char               * sigkey_new;
+  char               * sigkey_old;
+  char               * crypt;
+
+  SL_ENTER(_("sh_log_file"));
+
+  if (errFlags.HaveLog == BAD)  /* paranoia */ 
+    SL_RETURN((-1), _("sh_log_file"));
 
 #ifdef SH_USE_XML
   if (NULL == errmsg)
     {
+      while (current != NULL)
+        {
+	  /* don't write second EOF mark
+	   */
+	  if (current->log_start != S_TRUE)
+	    {
+	      /* Don't use inet_peer == NULL, userwise a lock file will
+	       * be created.
+	       */
+	      sh_log_open ("\0", 
+			   current->logfile, &(current->service_failure), &fd);
+          
 #ifdef SH_STEALTH
-      sl_write_line (fd, N_("</trail>"), 7);
-      sl_write (fd, "\n", 1);
-      sl_sync(fd);
+	      sl_write_line (fd, N_("</trail>"), 7);
+	      sl_write (fd, "\n", 1);
+	      sl_sync(fd);
 #else
-      sl_write_line (fd, _("</trail>\n"),  8);
-      sl_sync(fd);
+	      sl_write_line (fd, _("</trail>\n"),  8);
+	      sl_sync(fd);
 #endif
-      sl_close(fd);
+	      sl_close(fd);
+	      /* sh_unix_rm_lock_file (current->logfile); */
+	    }
+	  next    = current->next;
+	  SH_FREE(current->logfile);
+	  SH_FREE(current);
+	  current = next;
+	}
+      logfile_list = NULL;
+      SL_RETURN( 0, _("sh_log_file"));
+    }
+#else
+  if (NULL == errmsg)
+    {
+      while (current != NULL)
+        {
+	  /* sh_unix_rm_lock_file (current->logfile); */
+	  next    = current->next;
+          SH_FREE(current->logfile);
+          SH_FREE(current);
+          current = next;
+        }
+      logfile_list = NULL;
       SL_RETURN( 0, _("sh_log_file"));
     }
 #endif
 
+  sl_strlcpy (logfile, sh.srvlog.name, sizeof(logfile));
+  if (inet_peer != NULL && flag_sep_log == S_TRUE)
+    {
+      sl_strlcat (logfile, ".",       sizeof(logfile));
+      sl_strlcat (logfile, inet_peer, sizeof(logfile));
+    }
+
+  if (sh.flag.log_start == S_TRUE)
+    {
+      while (current != NULL)
+        {
+          current->log_start = S_TRUE;
+          current = current->next;
+        }
+      sh.flag.log_start    = S_FALSE;
+      current = logfile_list;
+    }
+
+  while (current != NULL)
+    {
+      if (strcmp(logfile, current->logfile) == 0)
+        break;
+      current = current->next;
+    }
+
+  if (current == NULL)
+    {
+      current                  = SH_ALLOC(sizeof(open_logfile));
+      current->logfile         = SH_ALLOC(strlen(logfile) + 1);
+      sl_strlcpy(current->logfile, logfile, strlen(logfile) + 1);
+      current->service_failure = 0;
+      current->log_start       = S_TRUE;
+      memset(current->sigkey_old, '\0', KEY_LEN+1);
+      memset(current->sigkey_new, '\0', KEY_LEN+1);
+      memset(current->crypt,      '\0', KEY_LEN+1);
+      current->next            = logfile_list;
+      logfile_list             = current;
+    }
+
+  if (0 != sh_log_open (inet_peer, current->logfile, 
+                        &(current->service_failure), &fd))
+    {
+      SL_RETURN ((-1), _("sh_log_file"));
+    }
+
+
   /* --- Allocate storage and mlock it. ---
    */
 
@@ -600,30 +743,30 @@ int  sh_log_file (char *errmsg)
   log_msg.msg = (char *) SH_ALLOC ( 2*KEY_LEN + status + 32); 
 
 #if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
-  if (skey->mlock_failed == GOOD) 
+  if (skey->mlock_failed == SL_FALSE) 
     {
-      sl_set_suid ();
-      if ( (-1) == mlock( log_msg.msg, 2*KEY_LEN + status + 32 ) ) 
+      if ( (-1) == sh_unix_mlock( log_msg.msg, 2*KEY_LEN + status + 32 ) ) 
 	{
-	  sl_unset_suid ();
-	  skey->mlock_failed = BAD;
-	  sh_error_handle ((-1), FIL__, __LINE__, EPERM, MSG_MLOCK);
+	  skey->mlock_failed = SL_TRUE;
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
+	  sh_error_handle ((-1), FIL__, __LINE__, EPERM, MSG_MLOCK); 
+#endif
 	}
-      else
-	sl_unset_suid ();
     }
 #else
-  if (skey->mlock_failed == GOOD) 
+  if (skey->mlock_failed == SL_FALSE) 
     {
-      skey->mlock_failed = BAD;
+      skey->mlock_failed = SL_TRUE;
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
       sh_error_handle ((-1), FIL__, __LINE__, EPERM, MSG_MLOCK);
+#endif
     }
 #endif
 
   /* --- Write the start marker. --- 
    */
 
-  if (sh.flag.log_start == S_TRUE) 
+  if (current->log_start == S_TRUE) 
     {
 #ifdef SH_USE_XML
 #ifdef SH_STEALTH
@@ -683,68 +826,116 @@ int  sh_log_file (char *errmsg)
   sh_do_encode (log_msg.msg, status);
 #endif
 
+  if (flag_sep_log == S_TRUE && inet_peer != NULL)
+    {
+      sigkey_old = current->sigkey_old;
+      sigkey_new = current->sigkey_new;
+      crypt      = current->crypt;
+    }
+  else
+    {
+      sigkey_old = skey->sigkey_old;
+      sigkey_new = skey->sigkey_new;
+      crypt      = skey->crypt;
+    }
+
   /* write the signature 
    */
-  if (sh.flag.log_start == S_TRUE) 
+  if (current->log_start == S_TRUE) 
     {
-      store1               = errFlags.loglevel;
-      store2               = errFlags.sysloglevel;
-      errFlags.loglevel    = SH_ERR_NOT;
-      errFlags.sysloglevel = SH_ERR_NOT;
-
       if (sh.real.user[0] == '\0') 
 	(void) sh_unix_getUser();
 
       /* Initialize the key.
        */
-      sh_util_keyinit(skey->sigkey_old, KEY_LEN+1);
+      sh_util_keyinit(sigkey_old, KEY_LEN+1);
 
       /* Hash the key to make sure it has the correct format.
        */
-      sl_strlcpy(skey->sigkey_new, 
-		 sh_tiger_hash (skey->sigkey_old, TIGER_DATA, KEY_LEN), 
+      sl_strlcpy(sigkey_new, 
+		 sh_tiger_hash (sigkey_old, TIGER_DATA, KEY_LEN), 
 		 KEY_LEN+1);
 
       /* Copy it to 'crypt' for encryption.
        */
-      sl_strlcpy(skey->crypt, skey->sigkey_new, KEY_LEN+1);
+      sl_strlcpy(crypt, sigkey_new, KEY_LEN+1);
 
       /* Use message and compiled-in key to encrypt.
        */
       BREAKEXIT(sh_util_encode);
-      sh_util_encode(skey->crypt, log_msg.msg, 0, 'B');
+      sh_util_encode(crypt, log_msg.msg, 0, 'B');
 
       /* Send out the key.
        */
       sl_strlcpy(log_msg.timestamp, sh_unix_time(0), KEY_LEN+1); 
+
+      store1               = errFlags.loglevel;
+      store2               = errFlags.sysloglevel;
+      store3               = errFlags.printlevel;
+      store4               = errFlags.exportlevel;
+      store5               = errFlags.maillevel;
+      store6               = errFlags.externallevel;
+      store7               = errFlags.databaselevel;
+      store8               = errFlags.preludelevel;
+
+      /* mail the key
+       */
+      errFlags.loglevel       = SH_ERR_NOT;
+      errFlags.sysloglevel    = SH_ERR_NOT;
+      errFlags.printlevel     = SH_ERR_NOT;
+      errFlags.exportlevel    = SH_ERR_NOT;
+      errFlags.externallevel  = SH_ERR_NOT;
+      errFlags.databaselevel  = SH_ERR_NOT;
+      errFlags.preludelevel   = SH_ERR_NOT;
+
+      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_KEY_MAIL,
+		       sh.prg_name, crypt, 
+		       crypt, log_msg.timestamp);
+
+      /* send to other allowed channels
+       */
+      errFlags.maillevel      = SH_ERR_NOT;
+      /* errFlags.printlevel     = store3; */
+      errFlags.exportlevel    = store4;
+      errFlags.externallevel  = store6;
+      errFlags.databaselevel  = store7;
+      errFlags.preludelevel   = store8;
+
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_KEY,
-		       sh.prg_name, skey->crypt, 
-		       skey->crypt, log_msg.timestamp);
+		       sh.prg_name, crypt);
 
       /* Cleanup.
        */
-      memset (skey->crypt, '\0', KEY_LEN);
-      errFlags.loglevel    = store1;
-      errFlags.sysloglevel = store2;
-      sh.flag.log_start   = S_FALSE;  
+      errFlags.loglevel       = store1;
+      errFlags.sysloglevel    = store2;
+      errFlags.printlevel     = store3;
+      errFlags.exportlevel    = store4;
+      errFlags.maillevel      = store5;
+      errFlags.externallevel  = store6;
+      errFlags.databaselevel  = store7;
+
+
+      memset (crypt, '\0', KEY_LEN);
+      sh.flag.log_start    = S_FALSE;  
+      current->log_start   = S_FALSE;
     } 
   else 
     {
       log_msg.timestamp[0] = '\0';
-      sl_strlcpy (skey->sigkey_new, 
-		  sh_tiger_hash (skey->sigkey_old, TIGER_DATA, KEY_LEN),
+      sl_strlcpy (sigkey_new, 
+		  sh_tiger_hash (sigkey_old, TIGER_DATA, KEY_LEN),
 		  KEY_LEN+1);
     }
 
   /* --- Sign the message with the signature key. ---
    */
 
-  sl_strlcat (log_msg.msg, skey->sigkey_new, status + KEY_LEN + 2);
+  sl_strlcat (log_msg.msg, sigkey_new, status + KEY_LEN + 2);
   
   sl_strlcpy (log_msg.signature,
 	      sh_tiger_hash (log_msg.msg, TIGER_DATA, status + KEY_LEN), 
 	      KEY_LEN+1);
-  sl_strlcpy (skey->sigkey_old, skey->sigkey_new, KEY_LEN+1); 
+  sl_strlcpy (sigkey_old, sigkey_new, KEY_LEN+1); 
 
 #ifdef SH_USE_XML
   if (log_msg.timestamp[0] != '\0')
@@ -789,7 +980,9 @@ int  sh_log_file (char *errmsg)
     {
       sl_strlcat (log_msg.msg,   _("</trail>"), status + 2*KEY_LEN + 32); 
 #endif
+      
       sl_strlcat (log_msg.msg,   _("\n"), status + 2*KEY_LEN + 32); 
+      current->log_start = S_TRUE;
     }
 #else
   log_msg.msg[status] = '\0';
@@ -808,13 +1001,11 @@ int  sh_log_file (char *errmsg)
 
   /* --- Clean up and free record. ---
    */
-  memset (log_msg.msg,       '\0', status + 2*KEY_LEN + 3);
+  memset (log_msg.msg,       '\0', status + 2*KEY_LEN + 32);
   memset (log_msg.signature, '\0', KEY_LEN);
-  MUNLOCK(log_msg.msg,  status + 2*KEY_LEN + 3);
+  sh_unix_munlock (log_msg.msg,  status + 2*KEY_LEN + 32);
   SH_FREE(log_msg.msg);
 
-  if (tmp != NULL)
-    SH_FREE(tmp);
   SL_RETURN (0, _("sh_log_file"));
 }
 
