@@ -51,6 +51,7 @@
 #include "sh_tiger.h"
 #include "sh_mail.h"
 #include "sh_forward.h"
+#include "sh_prelude.h"
 
 #if defined(WITH_DATABASE)
 #include "sh_tools.h"
@@ -65,13 +66,16 @@
 
 extern int clt_class;
 
+int flag_err_debug = SL_FALSE;
+int flag_err_info  = SL_FALSE;
 
 int  ShDFLevel[SH_ERR_T_END];
 
 typedef struct _log_t {
   char file[SH_PATHBUF];
   char format[SH_PATHBUF];
-  char msg[SH_BUFSIZE];
+  char * msg;
+  int  msg_len;
   int  severity;
   int  class;
   int  pid;
@@ -92,8 +96,7 @@ static int sh_error_string (struct _log_t * lmsg, va_list vl);
 
 extern int  sh_log_console (char *message);
 extern int  sh_log_syslog  (int  severity, char *message);
-extern int  sh_log_file    (char *message);
-
+extern int  sh_log_file    (char *message, char * inet_peer);
 
 static int  IsInitialized = BAD;
 
@@ -141,30 +144,73 @@ void sh_error_only_stderr (int flag)
 static int dbg_store = 0;
 static int dbg_flag  = 0;
 
-void sh_error_dbg_on()
+void compute_flag_err_debug()
 {
-  if (dbg_flag == 1)
-    return;
-
-  dbg_store           = errFlags.printlevel;
-  errFlags.printlevel = (SH_ERR_ALL    | SH_ERR_INFO  | SH_ERR_NOTICE | 
-			 SH_ERR_WARN   | SH_ERR_STAMP | SH_ERR_ERR    | 
-			 SH_ERR_SEVERE | SH_ERR_FATAL);
-  dbg_flag  = 1;
+  if ((errFlags.loglevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.printlevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.maillevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.exportlevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.sysloglevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.externallevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.databaselevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.preludelevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else
+    flag_err_debug = SL_FALSE;
   return;
 }
 
-void sh_error_dbg_off()
+void compute_flag_err_info()
+{
+  if ((errFlags.loglevel & SH_ERR_INFO) != 0)
+    flag_err_info = SL_TRUE;
+  else if ((errFlags.printlevel & SH_ERR_INFO) != 0)
+    flag_err_info = SL_TRUE;
+  else if ((errFlags.maillevel & SH_ERR_INFO) != 0)
+    flag_err_info = SL_TRUE;
+  else if ((errFlags.exportlevel & SH_ERR_INFO) != 0)
+    flag_err_info = SL_TRUE;
+  else if ((errFlags.sysloglevel & SH_ERR_INFO) != 0)
+    flag_err_info = SL_TRUE;
+  else if ((errFlags.externallevel & SH_ERR_INFO) != 0)
+    flag_err_info = SL_TRUE;
+  else if ((errFlags.databaselevel & SH_ERR_INFO) != 0)
+    flag_err_info = SL_TRUE;
+  else if ((errFlags.preludelevel & SH_ERR_INFO) != 0)
+    flag_err_info = SL_TRUE;
+  else
+    flag_err_info = SL_FALSE;
+  return;
+}
+
+void sh_error_dbg_switch()
 {
   if (dbg_flag == 0)
-    return;
-
-  errFlags.printlevel = dbg_store;
-  dbg_store = 0;
-  dbg_flag  = 0;
+    {
+      dbg_store           = errFlags.printlevel;
+      errFlags.printlevel = (SH_ERR_ALL    | SH_ERR_INFO  | SH_ERR_NOTICE | 
+			     SH_ERR_WARN   | SH_ERR_STAMP | SH_ERR_ERR    | 
+			     SH_ERR_SEVERE | SH_ERR_FATAL);
+      dbg_flag  = 1;
+      flag_err_debug = SL_TRUE;
+    }
+  else {
+    errFlags.printlevel = dbg_store;
+    dbg_store = 0;
+    dbg_flag  = 0;
+    compute_flag_err_debug();
+  }
   return;
 }
 
+
 static int sh_error_set_classmask (char * c, int * facility_mask)
 {
   char * p;
@@ -202,8 +248,19 @@ static int sh_error_set_classmask (char * c, int * facility_mask)
 
     for (i = 0; i < SH_CLA_MAX; ++i)
       {
-	if (0 == strcmp(p, _(class_cat[i])))
-	  *facility_mask |= (1 << i);
+	if (i < SH_CLA_RAW_MAX) {
+	  if (0 == strcmp(p, _(class_cat[i])))
+	    *facility_mask |= (1 << i);
+	} else {
+	  if (0 == strcmp(p, _(class_cat[SH_CLA_RAW_MAX + 0])))
+	    *facility_mask |= OTHER_CLA;
+	  if (0 == strcmp(p, _(class_cat[SH_CLA_RAW_MAX + 1])))
+	    *facility_mask |= RUN_NEW;
+	  if (0 == strcmp(p, _(class_cat[SH_CLA_RAW_MAX + 2])))
+	    *facility_mask |= FIL_NEW;
+	  if (0 == strcmp(p, _(class_cat[SH_CLA_RAW_MAX + 3])))
+	    *facility_mask |= ERROR_CLA;
+	}	  
       }
 
   } while (p);
@@ -239,6 +296,10 @@ int sh_error_database_mask (char * c)
 {
   return (sh_error_set_classmask(c, &(errFlags.database_class)));
 }
+int sh_error_prelude_mask (char * c)
+{
+  return (sh_error_set_classmask(c, &(errFlags.prelude_class)));
+}
   
 
 
@@ -503,6 +564,8 @@ int sh_error_set_level(char * str_s, int * facility)
 		       _("priority"), str_s);
       SL_RETURN (-1, _("sh_error_set_level"));
     }
+  compute_flag_err_debug();
+  compute_flag_err_info();
   SL_RETURN (0, _("sh_error_set_level"));
 }
 
@@ -608,6 +671,22 @@ int sh_error_set_database (char * str_s)
 }
 #endif
 
+#if defined(HAVE_LIBPRELUDE)
+/* set severity for prelude
+ */
+int sh_error_set_prelude (char * str_s)
+{
+  static int reject = 0;
+  if (reject == 1)
+    return (0);
+
+  if (sh.flag.opts == S_TRUE)  
+    reject = 1;
+
+  return (sh_error_set_level(str_s, &errFlags.preludelevel));
+}
+#endif
+
 /* set severity for mailing
  */
 int sh_error_setseverity (char * str_s)
@@ -622,6 +701,17 @@ int sh_error_setseverity (char * str_s)
   return (sh_error_set_level(str_s, &errFlags.maillevel));
 }
 
+#ifdef SH_WITH_SERVER
+static char inet_peer[SH_MINIBUF] = { '\0' };
+
+void sh_error_set_peer(const char * str)
+{
+  if (str == NULL)
+    memset(inet_peer, '\0', SH_MINIBUF);
+  else
+    sl_strlcpy(inet_peer, str, SH_MINIBUF);
+}
+#endif
   
 
 /**********************************************************
@@ -647,6 +737,10 @@ void sh_error_handle (int sev, char * file, long line,
   int    flag_inet  = S_FALSE;
   int    class_inet = clt_class;      /* initialize from global */
 
+#ifdef SH_WITH_SERVER
+  char   local_inet_peer[SH_MINIBUF];
+#endif
+
 #ifdef SH_WITH_CLIENT
   char   * ex_msg;
 #endif
@@ -675,9 +769,24 @@ void sh_error_handle (int sev, char * file, long line,
 #if defined(WITH_DATABASE)
   static int database_block = 0;
 #endif
+#ifdef HAVE_LIBPRELUDE
+  static int prelude_block = 0;
+#endif
 
   SL_ENTER(_("sh_error_handle"));
 
+#ifdef SH_WITH_SERVER
+  /* copy the global string into a local array
+   */
+  if ((msg_id == MSG_TCP_MSG) && (inet_peer[0] != '\0'))
+    {
+      sl_strlcpy(local_inet_peer, inet_peer, SH_MINIBUF);
+      sh_error_set_peer(NULL);
+    }
+  else
+    local_inet_peer[0] = '\0';
+#endif
+
   clt_class = (-1);      /* reset global */
 
   if (own_block == 1)
@@ -736,6 +845,10 @@ void sh_error_handle (int sev, char * file, long line,
        ( (errFlags.externallevel  & severity    ) == 0 ||
 	 (errFlags.external_class & (1 << class)) == 0 )     &&
 #endif
+#ifdef HAVE_LIBPRELUDE
+       ( (errFlags.preludelevel   & severity    ) == 0 ||
+	 (errFlags.prelude_class  & (1 << class)) == 0 )     &&
+#endif
 #ifdef WITH_DATABASE
        ( (errFlags.databaselevel  & severity    ) == 0 ||
 	 (errFlags.database_class & (1 << class)) == 0 )     &&
@@ -755,12 +868,13 @@ void sh_error_handle (int sev, char * file, long line,
       SL_RET0(_("sh_error_handle"));
     }
 
+
   /* Allocate space for the message.
    */
   own_block = 1;
   lmsg = (struct _log_t *) SH_ALLOC(sizeof(struct _log_t));
   MLOCK( (char *) lmsg, sizeof(struct _log_t));
-  lmsg->msg[0] = '\0';
+  lmsg->msg = NULL;
 
   sl_strlcpy(lmsg->format, fmt, SH_PATHBUF);
   sl_strlcpy(lmsg->file, file, SH_PATHBUF);
@@ -772,6 +886,7 @@ void sh_error_handle (int sev, char * file, long line,
 
 
   /* Format the log message with timestamp etc.
+   * Allocate lmsg->msg
    */
   own_block = 1;
   va_start (vl, msg_id);
@@ -792,7 +907,13 @@ void sh_error_handle (int sev, char * file, long line,
     { 
       if (print_block == 0 && (errFlags.printlevel & SH_ERR_NOT) == 0) 
 	{
+	  /* no truncation
+	   */
 	  print_block = 1;
+	  TPT(( 0, FIL__, __LINE__, lmsg->msg)); 
+	  /*
+	   *  Reports first error after failure. Always tries.
+	   */
 	  sh_log_console (lmsg->msg);
 	  print_block = 0;
 	}
@@ -813,9 +934,14 @@ void sh_error_handle (int sev, char * file, long line,
 #endif
 	   (errFlags.sysloglevel  & SH_ERR_NOT)    == 0 ) 
 	{
+	  /* will truncate to 1023 bytes 
+	   */
 	  if (syslog_block == 0)
 	    {
 	      syslog_block = 1;
+	      /*
+	       * Ignores errors. Always tries.
+	       */
 	      sh_log_syslog (lmsg->severity, lmsg->msg);
 	      syslog_block = 0;
 	    }
@@ -832,7 +958,12 @@ void sh_error_handle (int sev, char * file, long line,
 	{
 	  if (external_block == 0)
 	    {
+	      /* no truncation
+	       */
 	      external_block = 1;
+	      /*
+	       *  Reports first error after failure. Always tries.
+	       */
 	      sh_ext_execute ('l', 'o', 'g', lmsg->msg, 0);
 	      external_block = 0;
 	    }
@@ -850,6 +981,8 @@ void sh_error_handle (int sev, char * file, long line,
 	{
 	  if (database_block == 0)
 	    {
+	      /* truncates; query_max is 16k
+	       */
 	      database_block = 1;
 #ifndef SH_STANDALONE
 	      if (msg_id == MSG_TCP_MSG 
@@ -860,12 +993,18 @@ void sh_error_handle (int sev, char * file, long line,
 		{
 		  /* do not escape twice
 		   */
+		  /*
+		   *  Reports failure every 60 min. Always tries.
+		   */
 		  sh_database_insert (lmsg->msg);
 		}
 	      else
 #endif
 		{
 		  escape_msg = sh_tools_safe_name(lmsg->msg, 0);
+		  /*
+		   *  Reports failure every 60 min. Always tries.
+		   */
 		  sh_database_insert (escape_msg);
 		  SH_FREE(escape_msg);
 		}
@@ -890,6 +1029,8 @@ void sh_error_handle (int sev, char * file, long line,
         {
           if (export_block == 0)
             {
+	      /* will truncate to 65280 bytes 
+	       */
               export_block = 1;
 	      ex_msg = SH_ALLOC (64 + sl_strlen(lmsg->msg) + 1);
 	      sprintf(ex_msg, _("%d?%d?%s"),             /* known to fit  */
@@ -913,6 +1054,8 @@ void sh_error_handle (int sev, char * file, long line,
 	{
 	  if (mail_block == 0)
 	    {
+	      /* will truncate to 998 bytes 
+	       */
 	      mail_block = 1;
 
 	      BREAKEXIT(sh_mail_msg);
@@ -926,6 +1069,33 @@ void sh_error_handle (int sev, char * file, long line,
 	}
 #endif
 
+#ifdef HAVE_LIBPRELUDE
+      if (((errFlags.preludelevel  & severity  )   != 0  &&
+	   (errFlags.prelude_class & (1 << class)) != 0  &&
+	   (errFlags.preludelevel  & SH_ERR_NOT)   == 0  &&
+	   (class != AUD)) 
+#ifdef SH_WITH_SERVER
+	     || (flag_inet == S_TRUE)
+#endif
+	  )
+	{
+	  if (prelude_block == 0)
+	    {
+	      /* will truncate to 998 bytes 
+	       */
+	      prelude_block = 1;
+
+	      BREAKEXIT(sh_prelude_alert);
+	      /*
+	       *  Reports first error after failure. Always tries.
+	       */
+	      sh_prelude_alert (severity, class, lmsg->msg);
+
+	      prelude_block = 0;
+	    }
+	}
+#endif
+
       /* Log to logfile
        */
 
@@ -943,9 +1113,25 @@ void sh_error_handle (int sev, char * file, long line,
 	{
 	  if (log_block == 0)
 	    {
+	      /* no truncation
+	       */
 	      log_block = 1;
 	      BREAKEXIT(sh_log_file);
-	      sh_log_file (lmsg->msg);
+#ifdef SH_WITH_SERVER
+	      if (0 != sl_ret_euid())
+		{
+		  /*
+		   *  Reports first error after failure. Always tries.
+		   */
+		  if (local_inet_peer[0] == '\0')
+		    sh_log_file (lmsg->msg, NULL);
+		  else
+		    sh_log_file (lmsg->msg, local_inet_peer);
+		}
+#else
+              sh_log_file (lmsg->msg, NULL);
+#endif
+	      /* sh_log_file (lmsg->msg); */
 	      log_block = 0;
 	    }
 	}
@@ -955,6 +1141,10 @@ void sh_error_handle (int sev, char * file, long line,
   /* Cleanup.
    */
   own_block = 1;
+
+  if (lmsg->msg)
+    SH_FREE( lmsg->msg );
+
   memset ( lmsg, '\0', sizeof(struct _log_t) );
   MUNLOCK( (char *) lmsg,       sizeof(struct _log_t) );
   SH_FREE( lmsg );
@@ -1029,17 +1219,29 @@ int sh_error_ehead (char * str_s)
   SL_RETURN( 0, _("sh_error_ehead"));
 }
 
+#if !defined (VA_COPY)
+#if defined (__GNUC__) && defined (__PPC__) && (defined (_CALL_SYSV) || defined (_WIN32))
+#define VA_COPY(ap1, ap2)     (*(ap1) = *(ap2))
+#elif defined (VA_COPY_AS_ARRAY)
+#define VA_COPY(ap1, ap2)     memmove ((ap1), (ap2), sizeof (va_list))
+#else /* va_list is a pointer */
+#define VA_COPY(ap1, ap2)     ((ap1) = (ap2))
+#endif
+#endif 
+
 
 /* print an error  into string
  */
 static int sh_error_string (struct _log_t * lmsg, va_list vl)
 {
   int len;
+  int required;
   unsigned long line;
   char sev[16];
   char cla[16];
   char tst[64];
   char *p;
+  va_list       vl2;
 
   st_format rep_ehead_tab[] = {
     { 'S', S_FMT_STRING,  0, 0, NULL},  /* severity  */
@@ -1130,10 +1332,15 @@ static int sh_error_string (struct _log_t * lmsg, va_list vl)
   rep_ehead_tab[2].data_str   = lmsg->file;
   rep_ehead_tab[3].data_ulong = line;
   rep_ehead_tab[4].data_str   = cla;
-  rep_ehead_tab[5].data_long = lmsg->status;
+  rep_ehead_tab[5].data_long  = lmsg->status;
   
   p = sh_util_formatted(ehead_format, rep_ehead_tab);
 
+  /* ---  copy the header to lmsg->msg  ---
+   */
+  lmsg->msg     = SH_ALLOC(SH_BUFSIZE);
+  lmsg->msg_len = SH_BUFSIZE;
+
   if (p)
     {
       sl_strlcpy (lmsg->msg, p, SH_BUFSIZE);
@@ -1144,24 +1351,38 @@ static int sh_error_string (struct _log_t * lmsg, va_list vl)
       lmsg->msg[0] = '\0';
     }
 
+
+  /* --- copy message to lmsg->msg ---
+   */
   if ( NULL == strchr(lmsg->format, '%') ) 
     {
-      sl_strlcat (lmsg->msg, lmsg->format, SH_BUFSIZE);
+      sl_strlcat (lmsg->msg, lmsg->format, lmsg->msg_len);
     }
-
-  /* use VA_COPY 
-   */
   else 
     {
-      len = sl_strlen(lmsg->msg);
-      sl_vsnprintf(&(lmsg->msg[len]), (SH_BUFSIZE-len), lmsg->format, vl);
+      /* use VA_COPY */
+      VA_COPY(vl2, vl);
+      len      = sl_strlen(lmsg->msg);
+      required = sl_vsnprintf(&(lmsg->msg[len]), 
+			      (lmsg->msg_len - len), lmsg->format, vl);
+      if ( (required + len) > (lmsg->msg_len - 4) )
+	{
+	  p = SH_ALLOC(required + len + 4);
+	  sl_strlcpy (p, lmsg->msg, required + len + 1);
+	  SH_FREE(lmsg->msg);
+	  lmsg->msg = p;
+	  lmsg->msg_len = required + len + 4;
+	  sl_vsnprintf(&(lmsg->msg[len]), 
+			      (required + len + 1), lmsg->format, vl2);
+	}
+      va_end(vl2);
     }
 
 #ifdef SH_USE_XML
   /* closing tag
    */
   if (lmsg->msg[sl_strlen(lmsg->msg)-1] != '>')
-    sl_strlcat (lmsg->msg, _(" />"), SH_BUFSIZE);
+    sl_strlcat (lmsg->msg, _(" />"), lmsg->msg_len);
 #endif
 
   SL_RETURN(0, _("sh_error_string"));
@@ -1181,9 +1402,15 @@ static int  sh_error_init ()
   errFlags.debug          = 0;
   errFlags.HaveLog        = GOOD;
   errFlags.sysloglevel    = SH_ERR_NOT;
+#if defined(SH_STEALTH)
   errFlags.loglevel       = SH_ERR_NOT;
+#else
+  errFlags.loglevel       = (SH_ERR_STAMP | SH_ERR_ERR    | SH_ERR_SEVERE |
+			     SH_ERR_FATAL);
+#endif
   errFlags.externallevel  = SH_ERR_NOT;
   errFlags.databaselevel  = SH_ERR_NOT;
+  errFlags.preludelevel   = SH_ERR_NOT;
   errFlags.maillevel      = SH_ERR_FATAL;
 #if defined(SH_STEALTH)
   errFlags.printlevel     = SH_ERR_NOT;
@@ -1202,12 +1429,12 @@ static int  sh_error_init ()
   errFlags.syslog_class   = 0xFFFF;
   errFlags.external_class = 0xFFFF;
   errFlags.database_class = 0xFFFF;
+  errFlags.prelude_class  = 0xFFFF;
 
 
   for (j = 0; j < SH_ERR_T_END; ++j) 
-    ShDFLevel[j] = SH_ERR_NOT;
+    ShDFLevel[j] = SH_ERR_SEVERE;
 
   IsInitialized = GOOD;
   SL_RETURN (0, _("sh_error_init"));
 }
-
