@@ -97,6 +97,8 @@ FILE * pdbgc;
 #include "sh_tiger.h"
 #include "sh_extern.h"
 #include "sh_calls.h"
+#define SH_NEED_PWD_GRP 1
+#include "sh_static.h"
 
 #ifndef FD_SET
 #define NFDBITS         32
@@ -113,7 +115,10 @@ FILE * pdbgc;
 
 
 #ifndef SH_UID
-#define SH_UID long
+#define SH_UID uid_t
+#endif
+#ifndef SH_GID
+#define SH_GID gid_t
 #endif
 
 #undef  FIL__
@@ -143,7 +148,7 @@ typedef struct _sh_com_t
 
   SH_UID   trusted_users[32];
   SH_UID   run_user_uid;
-  SH_UID   run_user_gid;
+  SH_GID   run_user_gid;
   int      privileged;
 
   int      pipeFD;
@@ -210,15 +215,15 @@ sh_com_t * command_init(void)
       ext_com->fnot_v[i]        = NULL;
       ext_com->argv[i]          = NULL;
       ext_com->envv[i]          = NULL;
-      ext_com->trusted_users[i] = -1;
+      ext_com->trusted_users[i] = (uid_t) -1;
     }
 
-  ext_com->run_user_uid     = (SH_UID) getuid();
-  ext_com->run_user_gid     = (SH_UID) getgid();
+  ext_com->run_user_uid     = (uid_t) getuid();
+  ext_com->run_user_gid     = (gid_t) getgid();
 
   sl_get_euid(&ff_euid);
-  ext_com->trusted_users[0] = 0;
-  ext_com->trusted_users[1] = (SH_UID) (ff_euid);
+  ext_com->trusted_users[0] = (uid_t) 0;
+  ext_com->trusted_users[1] = (uid_t) (ff_euid);
 
 
   ext_com->next             = NULL;
@@ -319,7 +324,6 @@ int sh_ext_init(char * command)
   SL_RETURN (0, _("sh_ext_init"));
 }
 
-
 static
 int sh_ext_uid (char * user, uid_t * uid, gid_t * gid)
 {
@@ -331,7 +335,8 @@ int sh_ext_uid (char * user, uid_t * uid, gid_t * gid)
     {
       SL_RETURN (-1, _("sh_ext_uid"));
     }
-  tempres = getpwnam(user);
+  tempres = sh_getpwnam(user);
+
   if (NULL != tempres) 
     {
       *uid = tempres->pw_uid;  
@@ -491,6 +496,11 @@ int sh_ext_add_default (char * dummy)
   int    i;
 
   SL_ENTER(_("sh_ext_add_default"));
+  if (dummy[0] == 'n' ||  dummy[0] == 'N' ||
+      dummy[0] == 'f' ||  dummy[0] == 'F' || dummy[0] == '0')
+    {
+      SL_RETURN(0, _("sh_ext_add_default"));
+    }
   p = sh_unix_getUIDdir (SH_ERR_ERR, (uid_t) ext_coms->run_user_uid);
   if (p)
     sh_ext_add_envv (_("HOME"), p);
@@ -603,8 +613,8 @@ int sh_ext_priv (char * c)
   SL_ENTER(_("sh_ext_priv"));
   if (0 == sh_ext_uid (c, &m_uid, &m_gid))
     {
-      ext_coms->run_user_uid = (SH_UID) m_uid;
-      ext_coms->run_user_gid = (SH_UID) m_gid;
+      ext_coms->run_user_uid = (uid_t) m_uid;
+      ext_coms->run_user_gid = (gid_t) m_gid;
       if (m_uid != (uid_t) 0)
 	ext_coms->privileged   = 0;
       SL_RETURN((0), _("sh_ext_priv"));
@@ -631,6 +641,8 @@ int sh_ext_popen (sh_com_t * task)
   char * envp[1];
   char * argp[1];
 
+  static int some_error = 0;
+
 #if defined (__linux__)
   SL_TICKET   fd;
   char        pname[128];
@@ -657,20 +669,24 @@ int sh_ext_popen (sh_com_t * task)
   if ( SL_ENONE != status)
     { 
       PDBG_S("SL_ENONE != status");
-      tmp  = sh_util_safe_name (task->command);
-      if (sl_trust_errfile != '\0')
+      if (some_error == 0)
 	{
-	  tmp2  = sh_util_safe_name (sl_trust_errfile());
-	  sh_error_handle((-1), FIL__, __LINE__, status, MSG_E_TRUST2,
-			  sl_error_string(status), tmp, tmp2);
-	  SH_FREE(tmp2);  
+	  tmp  = sh_util_safe_name (task->command);
+	  if (sl_trust_errfile != '\0')
+	    {
+	      tmp2  = sh_util_safe_name (sl_trust_errfile());
+	      sh_error_handle((-1), FIL__, __LINE__, status, MSG_E_TRUST2,
+			      sl_error_string(status), tmp, tmp2);
+	      SH_FREE(tmp2);  
+	    }
+	  else
+	    {
+	      sh_error_handle((-1), FIL__, __LINE__, status, MSG_E_TRUST1,
+			      sl_error_string(status), tmp);
+	    }
+	  SH_FREE(tmp);
 	}
-      else
-	{
-	  sh_error_handle((-1), FIL__, __LINE__, status, MSG_E_TRUST1,
-			  sl_error_string(status), tmp);
-	}
-      SH_FREE(tmp);
+      some_error = 1;
       SL_RETURN ((-1), _("sh_ext_popen"));
     }
 
@@ -687,13 +703,19 @@ int sh_ext_popen (sh_com_t * task)
 	  )
 	{
 	  PDBG_S("checksum mismatch");
-	  tmp  = sh_util_safe_name (task->command);
-	  sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_HASH, tmp);
-	  SH_FREE(tmp);
+	  if (some_error == 0)
+	    {
+	      tmp  = sh_util_safe_name (task->command);
+	      sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_HASH, tmp);
+	      SH_FREE(tmp);
+	    }
+	  some_error = 1;
 	  SL_RETURN ((-1), _("sh_ext_popen"));
  	}
     }
 
+  some_error = 0;
+
   PDBG(2);
 
   /* 
@@ -768,7 +790,6 @@ int sh_ext_popen (sh_com_t * task)
 
 	  /* drop root privileges, if possible && requested
 	   */
-	  sl_set_suid();
 	  if (task->privileged == 0 && 0 == getuid())
 	    {
 	      PDBGC_S("privileged");
@@ -899,14 +920,14 @@ int sh_ext_popen (sh_com_t * task)
       PDBG_S("is w");
       close (pipedes[STDIN_FILENO]);
       retry_fcntl (FIL__, __LINE__, pipedes[STDOUT_FILENO], F_SETFD, FD_CLOEXEC);
-      MBLK( outf = fdopen (pipedes[STDOUT_FILENO], "w"); )
+      outf = fdopen (pipedes[STDOUT_FILENO], "w");
     }
   else
     {
       PDBG_S("is r");
       close (pipedes[STDOUT_FILENO]);
       retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
-      MBLK( outf = fdopen (pipedes[STDIN_FILENO], "r"); )
+      outf = fdopen (pipedes[STDIN_FILENO], "r");
     }
 
   
@@ -918,10 +939,10 @@ int sh_ext_popen (sh_com_t * task)
       
       if (task->privileged == 0 && 0 == getuid())
 	sh_error_handle((-1), FIL__, __LINE__, errnum, MSG_NOEXEC,
-			(long) task->run_user_uid, tmp);
+			(UID_CAST) task->run_user_uid, tmp);
       else
 	sh_error_handle((-1), FIL__, __LINE__, errnum, MSG_NOEXEC,
-			(long) getuid(), tmp);
+			(UID_CAST) getuid(), tmp);
 
       SH_FREE(tmp);
 
@@ -962,7 +983,7 @@ int sh_ext_pclose (sh_com_t * task)
 
   PDBG_S(" -> pclose");
   fflush(task->pipe);
-  MBLK( fclose(task->pipe); )
+  fclose(task->pipe);
 
   task->pipe     = NULL;
   task->pid      = 0;
@@ -1055,11 +1076,11 @@ int sh_ext_filter (char * message, sh_com_t * task)
 int sh_ext_execute (char t1, char t2, char t3, char * message, int msg_siz)
 {
   sh_com_t * listval = ext_coms;
-
   int        status = 0;
-
   char     * tmp;
 
+  static  int some_error = 0;
+
   struct  sigaction  new_act;
   struct  sigaction  old_act;
 
@@ -1126,8 +1147,13 @@ int sh_ext_execute (char t1, char t2, char t3, char * message, int msg_siz)
 		  if (status >= 0)
 		    status = write (listval->pipeFD, "\n", 1);
 		  PDBG_D(status);
-		  if (status < 0)
+		  if (status >= 0)
 		    {
+		      some_error = 0;
+		    }
+		  if ((status < 0) && (some_error == 0))
+		    {
+		      some_error = 1;
 		      PDBG_S("some error");
 		      PDBG_D(status);
 		      tmp  = sh_util_safe_name (listval->command);
@@ -1136,16 +1162,19 @@ int sh_ext_execute (char t1, char t2, char t3, char * message, int msg_siz)
 			{
 			  if (listval->privileged == 0 && 
 			      (0 == getuid() || sl_is_suid()) )
-			    sh_error_handle((-1), FIL__, __LINE__, 0, MSG_NOEXEC,
-					    (long) listval->run_user_uid, tmp);
+			    sh_error_handle((-1), FIL__, __LINE__, 0, 
+					    MSG_NOEXEC,
+					    (UID_CAST) listval->run_user_uid, 
+					    tmp);
 			  else
-			    sh_error_handle((-1), FIL__, __LINE__, 0, MSG_NOEXEC,
-					    (long) getuid(), tmp);
+			    sh_error_handle((-1), FIL__, __LINE__, 0, 
+					    MSG_NOEXEC,
+					    (UID_CAST) getuid(), tmp);
 			  
 			  SH_FREE(tmp);
 			}
 
-		    }
+		    } 
 		  PDBG(-7);
 		  fflush(listval->pipe);
 		}
