@@ -65,11 +65,15 @@
 #include "sh_files.h"
 #include "sh_tiger.h"
 #include "sh_hash.h"
+#include "sh_ignore.h"
 
 #undef  FIL__
 #define FIL__  _("sh_files.c")
 
 
+extern int flag_err_debug;
+extern int flag_err_info;
+
 int sh_files_reportonce(char * c)
 {
   int i;
@@ -99,7 +103,7 @@ typedef struct dir_struct {
   long    NumBDev;
   long    NumAll;
   long    TotalBytes;
-  char    DirPath[PATH_MAX+2];
+  char    DirPath[PATH_MAX];
 } dir_type;
 
 typedef struct dirstack_entry {
@@ -125,9 +129,11 @@ static int        sh_files_fullpath  (char * testdir, char * d_name,
 				      char * statpath);
 static int        sh_files_pushdir   (int class, char * str_s);
 static int        sh_files_pushfile  (int class, char * str_s);
-static int        sh_files_checkdir  (int class, int rdepth, char * dirName);
+static int        sh_files_checkdir  (int class, int rdepth, char * dirName,
+				      char * relativeName);
 static ShFileType sh_files_filecheck (int class, char * dirName, 
-				      char * fileName, int * reported);
+				      char * fileName, int * reported,
+				      int rsrcflag);
 
 static long MaxRecursionLevel = 0;
 
@@ -160,7 +166,7 @@ int sh_files_setrecursion (char * flag_s)
 
 unsigned long sh_files_chk ()
 {
-  char       * tmp;
+  char       * tmp = NULL;
   ShFileType   status;
 
   dirstack_t * ptr;
@@ -175,21 +181,32 @@ unsigned long sh_files_chk ()
 
   while (ptr) 
     {
-      if (sig_terminate == 1)
+
+      if (sig_urgent == 1) {
 	SL_RETURN(fcount, _("sh_files_chk"));
+      }
 
       if (ptr->checked == S_FALSE)
 	{
 	  base = sh_util_basename (ptr->name);
 	  file = sh_util_filename (ptr->name);
+#if defined(WITH_TPT)
 	  tmp = sh_util_safe_name (ptr->name);
+#endif
+
 	  
-	  sh_error_handle ((-1),  FIL__, __LINE__, 0, MSG_FI_CHK,
-			   tmp);
+	  if (flag_err_info == SL_TRUE)
+	    {
+#if !defined(WITH_TPT)
+	      tmp = sh_util_safe_name (ptr->name);
+#endif
+	      sh_error_handle ((-1),  FIL__, __LINE__, 0, MSG_FI_CHK,
+			       tmp);
+	    }
 
 	  BREAKEXIT(sh_files_filecheck);
 	  status = sh_files_filecheck (ptr->class, base, file, 
-				       (int *) &(ptr->reported));
+				       (int *) &(ptr->reported), 0);
 	  
 	  TPT(( 0, FIL__, __LINE__, 
 		_("msg=<filecheck complete: %s> status=<%d> reported=<%d>\n"), 
@@ -202,17 +219,27 @@ unsigned long sh_files_chk ()
 
 	      if ( sh.flag.checkSum == SH_CHECK_INIT  || 
 		  sh_hash_have_it (ptr->name) >= 0)
-		  sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
-				   ShDFLevel[ptr->class] : 
-				   ShDFLevel[SH_ERR_T_FILE],
-				   FIL__, __LINE__, 0, MSG_FI_MISS,
-				   tmp);
-
+		{
+		  if (S_FALSE == sh_ignore_chk_del(ptr->name))
+		    {
+		      if (tmp == NULL) 
+			tmp = sh_util_safe_name (ptr->name);
+		      sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
+				       ShDFLevel[ptr->class] : 
+				       ShDFLevel[SH_ERR_T_FILE],
+				       FIL__, __LINE__, 0, MSG_FI_MISS,
+				       tmp);
+		    }
+		}
 	      else /* not there at init, and still missing */
+		{
+		  if (tmp == NULL) 
+		    tmp = sh_util_safe_name (ptr->name);
 		  sh_error_handle (SH_ERR_NOTICE,
 				   FIL__, __LINE__, 0,
 				   MSG_FI_FAIL,
 				   tmp);
+		}
 #ifndef REPLACE_OLD
 	      /* this will tell that we have seen the file, and thus prevent
 	       * deletion from the database, resulting in an incomplete
@@ -245,11 +272,16 @@ unsigned long sh_files_chk ()
 		   */
 		  if (sh_hash_have_it (ptr->name) >= 0)
 		    {
-		      sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
-				       ShDFLevel[ptr->class] : 
-				       ShDFLevel[SH_ERR_T_FILE],
-				       FIL__, __LINE__, 0, MSG_FI_MISS,
-				       tmp);
+		      if (S_FALSE == sh_ignore_chk_del(ptr->name))
+			{
+			  if (tmp == NULL) 
+			    tmp = sh_util_safe_name (ptr->name);
+			  sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE)? 
+					   ShDFLevel[ptr->class] : 
+					   ShDFLevel[SH_ERR_T_FILE],
+					   FIL__, __LINE__, 0, MSG_FI_MISS,
+					   tmp);
+			}
 #ifndef REPLACE_OLD
 		      if (sh.flag.checkSum != SH_CHECK_INIT) 
 			sh_hash_set_visited_true(ptr->name);
@@ -262,6 +294,8 @@ unsigned long sh_files_chk ()
 		    }
 		  else
 		    {
+		      if (tmp == NULL) 
+			tmp = sh_util_safe_name (ptr->name);
 		      sh_error_handle (SH_ERR_INFO, FIL__, __LINE__, 0,
 				       MSG_FI_FAIL,
 				       tmp);
@@ -272,7 +306,11 @@ unsigned long sh_files_chk ()
 	      ++fcount;
 	    }
 	  
-	  SH_FREE(tmp);
+	  if (tmp != NULL)
+	    {
+	      SH_FREE(tmp);
+	      tmp = NULL;
+	    }
 	  SH_FREE(file);
 	  SH_FREE(base);
 
@@ -427,7 +465,7 @@ static void sh_files_set_mask (unsigned long * mask,
   else 
     (*mask) &= ~val;
 
-  SL_RET0(_("sh_files_set_mask"));;
+  SL_RET0(_("sh_files_set_mask"));
 }
 
 /* set mask(class)
@@ -449,7 +487,7 @@ static int sh_files_parse_mask (unsigned long * mask, char * str)
   while (i < l) {
     if (str[i] == '\0')
       break;
-    if (str[i] == ' ' || str[i] == '\t')
+    if (str[i] == ' ' || str[i] == '\t' || str[i] == ',')
       {
 	++i;
 	continue;
@@ -468,7 +506,7 @@ static int sh_files_parse_mask (unsigned long * mask, char * str)
     else /* a word */
       {
 	k = 0;
-	while (k < 63 && str[i] != ' ' && str[i] != '\t'
+	while (k < 63 && str[i] != ' ' && str[i] != '\t' && str[i] != ','
 	       && str[i] != '+' && str[i] != '-' && str[i] != '\0') {
 	  myword[k] = str[i]; 
 	  ++i; ++k;
@@ -589,26 +627,17 @@ int sh_files_has_metachar (const char * str)
     SL_RETURN(0, _("sh_files_has_metachar"));
 }
 
-extern int myblock_flag;
 
 int sh_files_globerr (const char * epath, int errnum)
 {
   char * p;
-  int    mb_flag = 0;
 
   SL_ENTER(_("sh_files_globerr"));
 
-  if (myblock_flag == 1) {
-    mb_flag      = 1;
-    myblock_flag = 0;
-  }
   p = sh_util_safe_name (epath);
   sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, errnum, MSG_FI_GLOB,
 		   sh_error_message (errnum), p);
   SH_FREE(p);
-  if (mb_flag == 1) {
-    myblock_flag = 1;
-  }
 
   SL_RETURN(0, _("sh_files_globerr"));
 }
@@ -643,7 +672,7 @@ int sh_files_push_file_int (int class, const char * str_s, int len)
 }
 
 
-int sh_files_pushfile (int class, char * str_s)
+static int sh_files_pushfile (int class, char * str_s)
 {
   char  * tmp;
   int     len;
@@ -675,8 +704,10 @@ int sh_files_pushfile (int class, char * str_s)
 
   len = sl_strlen(str_s);
 
-  if (len > PATH_MAX) 
+  if (len >= PATH_MAX) 
     {
+      /* Name too long
+       */
       tmp = sh_util_safe_name (str_s);
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_2LONG,
 		       tmp);
@@ -685,10 +716,14 @@ int sh_files_pushfile (int class, char * str_s)
     } 
   else if (len < 1) 
     {
+      /* Should not happen (str_s == NULL caught further above)
+       */
       SL_RETURN((-1),_("sh_files_pushfile"));
     } 
   else if (str_s[0] != '/') 
     {
+      /* Not an absolute path
+       */
       tmp = sh_util_safe_name (str_s);
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_NOPATH,
 		       tmp);
@@ -697,7 +732,9 @@ int sh_files_pushfile (int class, char * str_s)
     } 
   else 
     {
-      
+      /* remove a terminating '/', take care of the 
+       * special case of the root directory.
+       */
       if (str_s[len-1] == '/' && len > 1)
 	{
 	  str_s[len-1] = '\0';
@@ -714,7 +751,7 @@ int sh_files_pushfile (int class, char * str_s)
   else
     {
       pglob.gl_offs = 0;
-      MBLK( globstatus    = glob (str_s, 0, sh_files_globerr, &pglob); )
+      globstatus    = glob (str_s, 0, sh_files_globerr, &pglob);
 
       if (globstatus == 0 && pglob.gl_pathc > 0)
 	{
@@ -755,7 +792,7 @@ int sh_files_pushfile (int class, char * str_s)
 
 	}
 
-      MBLK( globfree(&pglob); )
+      globfree(&pglob);
     }
 
 #else
@@ -812,8 +849,9 @@ unsigned long sh_dirs_chk (int which)
 
   while (ptr) 
     {
-      if (sig_terminate == 1)
+      if (sig_urgent == 1) {
 	SL_RETURN(dcount, _("sh_dirs_chk"));
+      }
 
       if (ptr->checked == S_FALSE)
 	{
@@ -829,7 +867,7 @@ unsigned long sh_dirs_chk (int which)
 		    {
 		      BREAKEXIT(sh_files_filecheck);
 		      sh_files_filecheck (dst_ptr->class,  ptr->name,  
-					  NULL,  &status);
+					  NULL,  &status, 0);
 		      dst_ptr->checked = S_TRUE;
 		      status           = S_TRUE;
 		      break;
@@ -843,21 +881,25 @@ unsigned long sh_dirs_chk (int which)
 	      dst_ptr = dst_ptr->next;
 	    }
 	  if (status == S_FALSE)
-	    sh_files_filecheck (ptr->class,  ptr->name,  NULL,  &status);
+	    sh_files_filecheck (ptr->class,  ptr->name,  NULL,  &status, 0);
 
 	  BREAKEXIT(sh_files_checkdir);
-	  status = sh_files_checkdir (ptr->class, ptr->rdepth, ptr->name);
+	  status = sh_files_checkdir (ptr->class, ptr->rdepth, ptr->name, 
+				      ptr->name);
 
 	  if (status < 0 && ptr->reported == S_FALSE) 
 	    {
 	      /* directory is missing
 	       */
-	      tmp = sh_util_safe_name (ptr->name);
-	      sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
-			       ShDFLevel[ptr->class] : 
-			       ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__,
-			       0, MSG_FI_MISS, tmp);
-	      SH_FREE(tmp);
+	      if (S_FALSE == sh_ignore_chk_del(ptr->name))
+		{
+		  tmp = sh_util_safe_name (ptr->name);
+		  sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
+				   ShDFLevel[ptr->class] : 
+				   ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__,
+				   0, MSG_FI_MISS, tmp);
+		  SH_FREE(tmp);
+		}
 	      if (sh.flag.reportonce == S_TRUE)
 		ptr->reported = S_TRUE;
 	    } 
@@ -899,8 +941,9 @@ unsigned long sh_dirs_chk (int which)
 	  ptr->checked = S_TRUE;
 	}
 
-      if (sig_terminate == 1)
+      if (sig_urgent == 1) {
 	SL_RETURN(dcount, _("sh_dirs_chk"));
+      }
 
       ptr = ptr->next;
     }
@@ -991,7 +1034,7 @@ int sh_files_push_dir_int (int class, char * tail, int len, int rdepth)
   SL_RETURN(0, _("sh_files_push_dir_int"));
 }
 
-int sh_files_pushdir (int class, char * str_s)
+static int sh_files_pushdir (int class, char * str_s)
 {
   char  * tmp;
   int     len;
@@ -1030,7 +1073,7 @@ int sh_files_pushdir (int class, char * str_s)
 
   len = sl_strlen(tail);
 
-  if (len > PATH_MAX) 
+  if (len >= PATH_MAX) 
     {
       tmp = sh_util_safe_name (tail);
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_2LONG,
@@ -1069,7 +1112,7 @@ int sh_files_pushdir (int class, char * str_s)
   else
     {
       pglob.gl_offs = 0;
-      MBLK( globstatus    = glob (tail, 0, sh_files_globerr, &pglob); )
+      globstatus    = glob (tail, 0, sh_files_globerr, &pglob);
 
       if (globstatus == 0 && pglob.gl_pathc > 0)
 	{
@@ -1110,7 +1153,7 @@ int sh_files_pushdir (int class, char * str_s)
 	  SH_FREE(tmp);
 	}
 
-      MBLK( globfree(&pglob); )
+      globfree(&pglob);
     }
 #else  
   sh_files_push_dir_int (class, tail, len, rdepth);
@@ -1120,6 +1163,7 @@ int sh_files_pushdir (int class, char * str_s)
 }  
 
 struct sh_dirent {
+  /* char               sh_d_name[NAME_MAX + 2]; */
   char             * sh_d_name;
   struct sh_dirent * next;
 };
@@ -1148,12 +1192,12 @@ static struct sh_dirent * addto_sh_dirlist (struct dirent * thisEntry,
 
   if (thisEntry == NULL)
     return dirlist;
-
+  
   i = sl_strlen(thisEntry->d_name);
   if (i == 0)
     return dirlist;
   ++i;
-
+  
   this = SH_ALLOC(sizeof(struct sh_dirent));
   if (!this)
     return dirlist;
@@ -1165,9 +1209,94 @@ static struct sh_dirent * addto_sh_dirlist (struct dirent * thisEntry,
   return this;
 }
 
+static int sh_check_hardlinks = S_TRUE;
+
+/* Simply sets our boolean as to whether this check is active 
+ */
+int sh_files_check_hardlinks (char * opt)
+{
+  int i;
+  SL_ENTER(_("sh_files_check_hardlinks"));
+  i = sh_util_flagval(opt, &sh_check_hardlinks);
+  SL_RETURN(i, _("sh_files_check_hardlinks"));
+}
+
+struct sh_hle_struct {
+  long   offset;
+  char * path;
+  struct sh_hle_struct * next;
+};
+
+static struct sh_hle_struct * sh_hl_exc = NULL;
+
+int sh_files_hle_reg (char * str)
+{
+  long   offset;
+  size_t len;
+  char * path;
+  
+  struct sh_hle_struct * tmp = sh_hl_exc;
+
+  SL_ENTER(_("sh_files_hle_reg"));
+
+  /* Free the linked list if called with NULL argument
+   */
+  if (str == NULL)
+    {
+      while (tmp)
+	{
+	  sh_hl_exc = tmp->next;
+	  SH_FREE(tmp->path);
+	  SH_FREE(tmp);
+	  tmp = sh_hl_exc;
+	}
+      sh_hl_exc = NULL;
+      SL_RETURN(0, _("sh_files_hle_reg"));
+    }
+
+  /* We expect 'offset:/path'
+   */
+  offset = strtol(str, &path, 0);
+  if ((path == NULL) || (*path == '\0') || (*path != ':') || (path[1] != '/'))
+    {
+      SL_RETURN(-1, _("sh_files_hle_reg"));
+    }
+  ++path;
+  len = 1 + sl_strlen(path);
+
+  tmp         = SH_ALLOC(sizeof(struct sh_hle_struct));
+  tmp->path   = SH_ALLOC(len);
+  sl_strlcpy (tmp->path, path, len);
+  tmp->offset = offset;
+  tmp->next   = sh_hl_exc;
+  sh_hl_exc   = tmp;
+
+  SL_RETURN(0, _("sh_files_hle_reg"));
+}
+
+#if !defined(HOST_IS_DARWIN)
+static int sh_files_hle_test (int offset, char * path)
+{
+  struct sh_hle_struct * tmp = sh_hl_exc;
+
+  SL_ENTER(_("sh_files_hle_reg"));
+
+  while(tmp)
+    {
+      if ((offset == tmp->offset) && (0 == strcmp(path, tmp->path)))
+	{
+	  SL_RETURN(0, _("sh_files_hle_test"));
+	}
+      tmp = tmp->next;
+    }
+  SL_RETURN(-1, _("sh_files_hle_test"));
+}
+#endif
+
 /* -- check a single directory and its content
  */
-int sh_files_checkdir (int iclass, int idepth, char * iname)
+static int sh_files_checkdir (int iclass, int idepth, char * iname, 
+			      char * relativeName)
 {
   struct sh_dirent * dirlist = NULL;
   struct sh_dirent * dirlist_orig = NULL;
@@ -1195,10 +1324,14 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 
   dirstack_t *    dst_ptr;
 
+  int             hardlink_num = 0;
+
+
   SL_ENTER(_("sh_files_checkdir"));
 
-  if (sig_terminate == 1)
+  if (sig_urgent == 1) {
     SL_RETURN((0), _("sh_files_checkdir"));
+  }
 
   if (iname == NULL || idepth < (-1))
     SL_RETURN((-1), _("sh_files_checkdir"));
@@ -1216,14 +1349,17 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 
   /* ---- check for obscure name ----
    */
-  sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], iname, S_TRUE);
-  
+  if (iclass != SH_LEVEL_ALLIGNORE)
+    {
+      sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], iname, S_TRUE);
+    }
+
   sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_CHK,
 		   tmpname);
   
   /* ---- check input ----
    */
-  if ( sl_strlen(iname) > PATH_MAX) 
+  if ( sl_strlen(iname) >= PATH_MAX) 
     {
       sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__, 0, 
 		       MSG_FI_2LONG,
@@ -1245,10 +1381,12 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
     
   /* ---- stat the directory ----
    */
-  sl_strlcpy (theFile.fullpath, iname, PATH_MAX+1);
+  sl_strlcpy (theFile.fullpath, iname, PATH_MAX);
+
+  (void) relativeName;
   status = sh_unix_getinfo (ShDFLevel[SH_ERR_T_DIR], 
 			    iname,
-			    &theFile, NULL);
+			    &theFile, NULL, iclass);
 
   if (status == -1)
     {
@@ -1265,14 +1403,15 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
       SL_RETURN((-1), _("sh_files_checkdir"));
     }
 
+  hardlink_num = theFile.hardlinks;
+
+
   /* ---- open directory for reading ---- 
    *
    * opendir() will fail with ENOTDIR if the path has been changed
    * to a non-directory in between lstat() and opendir().
    */
-  sl_set_suid();
   thisDir = opendir (iname);
-  sl_unset_suid();
 
   if (thisDir == NULL) 
     {
@@ -1281,6 +1420,7 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 		       MSG_E_OPENDIR,
 		       sh_error_message (status), tmpname);
       SH_FREE(tmpname); 
+
       SL_RETURN((-1), _("sh_files_checkdir"));
     }
 
@@ -1293,7 +1433,7 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
   theDir.NumBDev     = 0;
   theDir.NumAll      = 0;
   theDir.TotalBytes  = 0;
-  sl_strlcpy (theDir.DirPath, iname, PATH_MAX+1); 
+  sl_strlcpy (theDir.DirPath, iname, PATH_MAX); 
 
 
   /* ---- read ----
@@ -1319,6 +1459,8 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 
   closedir (thisDir);
 
+  ++sh.statistics.dirs_checked;
+
   dirlist_orig = dirlist;
 
   do {
@@ -1328,17 +1470,22 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
     if (!dirlist)
       break;
 
+    if (sig_termfast == 1) 
+      {
+	SL_RETURN((0), _("sh_files_checkdir"));
+      }
+
     BREAKEXIT(sh_derr);
     if (0 == (rand() % 5))
       (void) sh_derr();
     
     /* ---- Check the file. ---- 
      */
-    tmpcat = SH_ALLOC(PATH_MAX+2);
-    sl_strlcpy(tmpcat, iname,                   PATH_MAX+2);
+    tmpcat = SH_ALLOC(PATH_MAX);
+    sl_strlcpy(tmpcat, iname,                   PATH_MAX);
     if (sl_strlen(tmpcat) > 1 || tmpcat[0] != '/')
-      sl_strlcat(tmpcat, "/",                   PATH_MAX+2);
-    sl_strlcat(tmpcat, dirlist->sh_d_name,      PATH_MAX+2);
+      sl_strlcat(tmpcat, "/",                   PATH_MAX);
+    sl_strlcat(tmpcat, dirlist->sh_d_name,      PATH_MAX);
     
     dst_ptr         = fileList;
     rdepth_next     = rdepth - 1;
@@ -1367,7 +1514,16 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
     /* ---- Has been checked already. ----
      */
     if (checked_flag == S_TRUE && cchecked_flag == S_TRUE)
-      continue;
+      {
+	/* Mar 11 2004 get ftype for complete directory count
+	 */
+	checkit = sh_unix_get_ftype(tmpcat);
+	if (checkit == SH_FILE_DIRECTORY) 
+	  {
+	    ++theDir.NumDirs;
+	  }
+	continue;
+      }
     
     /* --- May be true, false, or not found. --- 
      */
@@ -1383,10 +1539,12 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 	 */
 	if (dst_ptr && sh.flag.reportonce == S_TRUE)
 	  dummy = dst_ptr->reported;
+
 	checkit = sh_files_filecheck (file_class_next, 
 				      iname, 
 				      dirlist->sh_d_name,
-				      &dummy);
+				      &dummy, 0);
+
 	if (dst_ptr && checked_flag == S_FALSE)
 	  dst_ptr->checked = S_TRUE;
 	/* Thu Mar  7 15:09:40 CET 2002 Propagate the 'reported' flag
@@ -1454,11 +1612,13 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 	    
 	    if (cchecked_flag == S_FALSE)
 	      {
-		sh_files_checkdir (class_next, rdepth_next, tmpcat);
+		sh_files_checkdir (class_next, rdepth_next, tmpcat, 
+				   dirlist->sh_d_name);
 		dst_ptr->childs_checked = S_TRUE;
 	      }
 	    else if (checked_flag == -1)
-	      sh_files_checkdir (class_next, rdepth_next, tmpcat);
+	      sh_files_checkdir (class_next, rdepth_next, tmpcat, 
+				 dirlist->sh_d_name);
 	    
 	  }
       }
@@ -1471,37 +1631,64 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
     
     SH_FREE(tmpcat);
     
-    if (sig_terminate == 1)
-      SL_RETURN((0), _("sh_files_checkdir"));
+    if ((sig_termfast == 1) || (sig_terminate == 1)) 
+      {
+	SL_RETURN((0), _("sh_files_checkdir"));
+      }
     
     dirlist = dirlist->next;
     
   } while (dirlist != NULL);
 
-  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_DSUM,
-		   theDir.NumDirs,
-		   theDir.NumRegular,
-		   theDir.NumSymlinks,
-		   theDir.NumFifos,
-		   theDir.NumSockets,
-		   theDir.NumCDev,
-		   theDir.NumBDev);
-  
+  if (flag_err_info == SL_TRUE)
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_DSUM,
+		       theDir.NumDirs,
+		       theDir.NumRegular,
+		       theDir.NumSymlinks,
+		       theDir.NumFifos,
+		       theDir.NumSockets,
+		       theDir.NumCDev,
+		       theDir.NumBDev);
+    }
+
   kill_sh_dirlist (dirlist_orig);
 
+#if !defined(HOST_IS_DARWIN)
+  /* 
+   * Hardlink check; not done on MacOS X because of resource forks
+   */
+  if ((sh_check_hardlinks == S_TRUE) && (hardlink_num != theDir.NumDirs)) 
+    {
+      if (0 != sh_files_hle_test(hardlink_num-theDir.NumDirs, iname))
+	{
+	  tmpcat = SH_ALLOC(strlen(tmpname) + 256);
+	  sl_snprintf(tmpcat, strlen(tmpname) + 256, 
+		      _("%s: subdirectory count (%d) != hardlinks (%d)"),
+		      tmpname, theDir.NumDirs, hardlink_num);
+	  sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__, 0, 
+			   MSG_E_SUBGEN, tmpcat, _("sh_files_checkdir"));
+	  SH_FREE(tmpcat);
+	}
+    }
+#endif
+
   SH_FREE(tmpname);
+
   SL_RETURN((0), _("sh_files_checkdir"));
 }
 
 int get_the_fd (SL_TICKET ticket);
 
 
-ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
-			       int * reported)
+static ShFileType sh_files_filecheck (int class, char * dirName, 
+				      char * fileName,
+				      int * reported, 
+				      int rsrcflag)
 {
   /* 28 Aug 2001 allow NULL fileName
    */
-  char            fullpath[PATH_MAX+2];
+  char            fullpath[PATH_MAX];
   char            fileHash[KEY_LEN + 1];
   int             status;
   file_type       theFile;
@@ -1523,17 +1710,28 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
       SL_RETURN(SH_FILE_UNKNOWN, _("sh_files_filecheck"));
     }
 
-  if (fileName != NULL && 0 != sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], 
-						    fileName, S_FALSE)) 
+  if ((fileName != NULL) && (class != SH_LEVEL_ALLIGNORE) && 
+      (0 != sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], 
+				 fileName, S_FALSE))) 
     {
-      tmpdir  = sh_util_safe_name (dirName);
-      tmpname = sh_util_safe_name (fileName);
-      
-      sh_error_handle (ShDFLevel[SH_ERR_T_NAME], FIL__, __LINE__, 0,
-		       MSG_FI_OBSC2,
-		       tmpdir, tmpname);
-      SH_FREE(tmpname);
-      SH_FREE(tmpdir);
+      if ((dirName != NULL) && (dirName[0] == '/') && (dirName[1] == '\0')) 
+	{
+	  tmpname = sh_util_safe_name (fileName);
+	  sh_error_handle (ShDFLevel[SH_ERR_T_NAME], FIL__, __LINE__, 0,
+			   MSG_FI_OBSC2,
+			   "", tmpname);
+	  SH_FREE(tmpname);
+	}
+      else
+	{
+	  tmpdir  = sh_util_safe_name (dirName);
+	  tmpname = sh_util_safe_name (fileName);
+	  sh_error_handle (ShDFLevel[SH_ERR_T_NAME], FIL__, __LINE__, 0,
+			   MSG_FI_OBSC2,
+			   tmpdir, tmpname);
+	  SH_FREE(tmpname);
+	  SH_FREE(tmpdir);
+	}
     }    
 
   /* sh_files_fullpath accepts NULL fileName
@@ -1553,7 +1751,7 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
 
   /* stat the file and determine checksum (if a regular file)
    */
-  sl_strlcpy (theFile.fullpath, fullpath, PATH_MAX+1);
+  sl_strlcpy (theFile.fullpath, fullpath, PATH_MAX);
   theFile.check_mask = sh_files_maskof(class);
   theFile.reported   = (*reported);
 
@@ -1562,7 +1760,7 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
   status = sh_unix_getinfo ( (class == SH_LEVEL_ALLIGNORE) ? 
 			     ShDFLevel[class] : ShDFLevel[SH_ERR_T_FILE], 
 			     fileName,
-			     &theFile, fileHash);
+			     &theFile, fileHash, class);
   
   if (status != 0)
     {
@@ -1573,15 +1771,20 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
       SL_RETURN(SH_FILE_UNKNOWN, _("sh_files_filecheck"));
     }
   
+  if (sig_termfast == 1) {
+    goto ret_point;
+  }
+
   /* report
    */
-  if (theFile.c_mode[0] == '-')
+  if ((flag_err_debug == SL_TRUE) && (theFile.c_mode[0] == '-'))
     {
       tmpname = sh_util_safe_name (fullpath); /* fixed in 1.5.4 */
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_CSUM,
 		       fileHash, tmpname);
       SH_FREE(tmpname);
     } 
+  ++sh.statistics.files_checked;
       
   if      ( sh.flag.checkSum == SH_CHECK_INIT && sh.flag.update == S_FALSE ) 
     {
@@ -1589,8 +1792,10 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
     }
   else if (sh.flag.checkSum == SH_CHECK_INIT && sh.flag.update == S_TRUE )
     {
-      sh_hash_compdata (class, &theFile, fileHash);
-      sh_hash_pushdata (&theFile, fileHash);
+      if (0 == sh_hash_compdata (class, &theFile, fileHash))
+	{
+	  sh_hash_pushdata (&theFile, fileHash);
+	}
     }
   else if (sh.flag.checkSum == SH_CHECK_CHECK 
 	   /* && theFile.c_mode[0] == '-' */
@@ -1609,24 +1814,53 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
       utime_buf.actime   = (time_t) theFile.atime;
       utime_buf.modtime  = (time_t) theFile.mtime;
 #if !defined(O_NOATIME)
-      sl_set_suid ();
       retry_aud_utime (FIL__, __LINE__, fullpath, &utime_buf);
-      sl_unset_suid ();
 #endif
     }
   
-  /* yes I know, the 'break' is not needed
+#ifdef HOST_IS_DARWIN
+  /*
+   * Check for resource fork
    */
+  if ( (theFile.c_mode[0] != 'd') && (rsrcflag == 0) )
+    {
+      int  dummy;
+      static int rsrc_init = 0;
+      static char rsrc[5];
+      char testpath[PATH_MAX];
+
+      if (rsrc_init == 0) {
+	sl_strlcpy(rsrc, _("rsrc"), 5);
+	rsrc_init = 1;
+      }
+      sl_strlcpy (testpath, fullpath, PATH_MAX);
+      sl_strlcat (testpath,      "/", PATH_MAX);
+      sl_strlcat (testpath,     rsrc, PATH_MAX);
+
+      if (sl_strlen(testpath) == (5 + sl_strlen(fullpath)))
+	{
+	  if (0 == sh_unix_file_stat (testpath))
+	    {
+	      sh_files_filecheck (class, fullpath, rsrc, &dummy, 1);
+	    }
+	}
+    }
+#else
+      (void) rsrcflag; /* avoid compiler warning */
+#endif
+
+ ret_point:
+
   switch (theFile.c_mode[0]) 
     {
-    case '-': SL_RETURN(SH_FILE_REGULAR, _("sh_files_filecheck"));   break;
-    case 'l': SL_RETURN(SH_FILE_SYMLINK, _("sh_files_filecheck"));   break;
-    case 'd': SL_RETURN(SH_FILE_DIRECTORY, _("sh_files_filecheck")); break;
-    case 'c': SL_RETURN(SH_FILE_CDEV, _("sh_files_filecheck"));      break;
-    case 'b': SL_RETURN(SH_FILE_BDEV, _("sh_files_filecheck"));      break;
-    case '|': SL_RETURN(SH_FILE_FIFO, _("sh_files_filecheck"));      break;
-    case 's': SL_RETURN(SH_FILE_SOCKET, _("sh_files_filecheck"));    break;
-    default:  SL_RETURN(SH_FILE_UNKNOWN, _("sh_files_filecheck"));   break;
+    case '-': SL_RETURN(SH_FILE_REGULAR, _("sh_files_filecheck"));   
+    case 'l': SL_RETURN(SH_FILE_SYMLINK, _("sh_files_filecheck"));   
+    case 'd': SL_RETURN(SH_FILE_DIRECTORY, _("sh_files_filecheck")); 
+    case 'c': SL_RETURN(SH_FILE_CDEV, _("sh_files_filecheck"));      
+    case 'b': SL_RETURN(SH_FILE_BDEV, _("sh_files_filecheck"));      
+    case '|': SL_RETURN(SH_FILE_FIFO, _("sh_files_filecheck"));      
+    case 's': SL_RETURN(SH_FILE_SOCKET, _("sh_files_filecheck"));    
+    default:  SL_RETURN(SH_FILE_UNKNOWN, _("sh_files_filecheck"));   
     }
   
   /* notreached */
@@ -1644,15 +1878,15 @@ static int sh_files_fullpath (char * testdir, char * d_name, char * statpath)
     {
       if ( (llen = sl_strlen(testdir)) > (PATH_MAX-2) ) 
 	SL_RETURN((-1),_("sh_files_fullpath"));
-      sl_strlcpy(statpath, testdir,    PATH_MAX - 2);
+      sl_strlcpy(statpath, testdir,    PATH_MAX - 1);
     }
   if (d_name != NULL) 
     {
       if (llen > 1 || statpath[0] != '/')
-	sl_strlcat(statpath, "/",   PATH_MAX+1);
-      if (sl_strlen(d_name) >= ( PATH_MAX - sl_strlen(statpath) ) ) 
+	sl_strlcat(statpath, "/",   PATH_MAX);
+      if ((sl_strlen(d_name) + sl_strlen(statpath)) >= PATH_MAX)
 	SL_RETURN((-1),_("sh_files_fullpath"));
-      sl_strlcat(statpath, d_name,   PATH_MAX+1);
+      sl_strlcat(statpath, d_name,   PATH_MAX);
     }
   if (statpath == NULL) 
     SL_RETURN((-1),_("sh_files_fullpath"));
