@@ -28,6 +28,9 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <signal.h>
+#if defined(SH_WITH_SERVER)
+#include <pwd.h>
+#endif
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <errno.h>
@@ -52,6 +55,10 @@
 #include "sh_utils.h"
 #include "sh_error.h"
 #include "sh_tiger.h"
+#if defined(SH_WITH_SERVER)
+#include "sh_static.h"
+#endif
+
 
 static struct {
   char     conf_id[SH_MINIBUF+1];
@@ -69,6 +76,40 @@ typedef struct {
 #define SH_GPG_BAD     1
 #define SH_GPG_BADSIGN 2
 
+/* replace #if 0 by #if 1 and set an appropriate path in front of '/pdbg.'
+ * for debugging
+ */
+#if 0
+#define PDGBFILE "/pdbg."
+#endif
+
+#if defined(PDGBFILE)
+FILE * pdbg;
+FILE * pdbgc;
+#define PDBG_OPEN    pdbg = fopen(PDGBFILE"main",  "a")  
+#define PDBG_CLOSE   fclose (pdbg)
+#define PDBG(arg)    fprintf(pdbg,  "PDBG: step %d\n", arg); fflush(pdbg)
+#define PDBG_D(arg)  fprintf(pdbg,  "PDBG: %d\n", arg); fflush(pdbg)
+#define PDBG_S(arg)  fprintf(pdbg,  "PDBG: %s\n", arg); fflush(pdbg)
+
+#define PDBGC_OPEN   pdbgc = fopen(PDGBFILE"child", "a")  
+#define PDBGC_CLOSE  fclose (pdbgc)
+#define PDBGC(arg)   fprintf(pdbgc, "PDBG: step %d\n", arg); fflush(pdbgc)
+#define PDBGC_D(arg) fprintf(pdbgc, "PDBG: %d\n", arg); fflush(pdbgc)
+#define PDBGC_S(arg) fprintf(pdbgc, "PDBG: %s\n", arg); fflush(pdbgc)
+#else
+#define PDBG_OPEN    
+#define PDBG_CLOSE   
+#define PDBG(arg)    
+#define PDBG_D(arg)  
+#define PDBG_S(arg)  
+#define PDBGC_OPEN    
+#define PDBGC_CLOSE   
+#define PDBGC(arg)    
+#define PDBGC_D(arg)  
+#define PDBGC_S(arg)  
+#endif
+
 #undef  FIL__
 #define FIL__  _("sh_gpg.c")
 
@@ -78,6 +119,10 @@ static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
   char * test_gpg;
   char * test_ptr1 = NULL;
   char * test_ptr2 = NULL;
+  char   wstrip1[128];
+  char   wstrip2[128];
+  int    i, k;
+#include "sh_gpg_chksum.h"
 
   SL_ENTER(_("sh_gpg_checksum"));
 
@@ -100,14 +145,55 @@ static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
     test_ptr1 += 2;
   else
     test_ptr1 = GPG_HASH;
-  
-  if (0 != sl_strncmp(test_ptr1, test_ptr2, KEY_LEN + 6))
+
+  /* Tue Jun 24 23:11:54 CEST 2003 (1.7.9) -- strip whitespace
+   */
+  k = 0;
+  for (i = 0; i < 127; ++i)
+    {
+      if (test_ptr1[i] == '\0')
+	break;
+      if (test_ptr1[i] != ' ')
+	{
+	  wstrip1[k] = test_ptr1[i];
+	  ++k;
+	}
+    }
+  wstrip1[k] = '\0';
+
+  for(i = 0; i < KEY_LEN; ++i)
+    {
+      if (gpgchk[i] != wstrip1[i]) 
+	{
+	  sh_error_handle(SH_ERR_SEVERE, FIL__, __LINE__, 0, MSG_E_GPG_CHK, 
+			  gpgchk, wstrip1);
+	  break;
+	}
+    }
+
+  k = 0;
+  for (i = 0; i < 127; ++i)
+    {
+      if (test_ptr2[i] == '\0')
+	break;
+      if (test_ptr2[i] != ' ')
+	{
+	  wstrip2[k] = test_ptr2[i];
+	  ++k;
+	}
+    }
+  wstrip2[k] = '\0';
+
+  if (0 != sl_strncmp(wstrip1, wstrip2, 127))
     {
       TPT(((0), FIL__, __LINE__, _("msg=<pgp checksum: %s>\n"), test_gpg));
       TPT(((0), FIL__, __LINE__, _("msg=<Compiled-in : %s>\n"), GPG_HASH));
+      TPT(((0), FIL__, __LINE__, _("msg=<wstrip1     : %s>\n"), wstrip1));
+      TPT(((0), FIL__, __LINE__, _("msg=<wstrip2     : %s>\n"), wstrip2));
       if (flag == 1)
 	sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_GPG, 
 			GPG_HASH, test_gpg);
+      dlog(1, FIL__, __LINE__, _("The compiled-in checksum of the gpg binary\n(%s)\ndoes not match the actual checksum\n(%s).\nYou need to recompile with the correct checksum."), wstrip1, wstrip2);
       SH_FREE(test_gpg);
       SL_RETURN((-1), _("sh_gpg_checksum"));
     }
@@ -117,7 +203,7 @@ static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
 #endif
 
 static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd, 
-			    int mode, char * id)
+			    int mode, char * id, char * homedir)
 {
   int pipedes[2];
   FILE * outf = NULL;
@@ -131,9 +217,12 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 #endif
 #if defined(WITH_GPG)
   char   cc0[2] = "-";
+  char   cc3[32];
+  char   cc4[SH_PATHBUF+32];
+  char   cc5[32] = "--no-tty";
 #endif
 
-  char * arg[6];
+  char * arg[9];
 
 #if defined(HAVE_GPG_CHECKSUM)
   SL_TICKET   checkfd;
@@ -150,17 +239,97 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 
 #if defined(WITH_GPG)
   /* -- GnuPG -- */
-  sl_strlcpy (path,  DEFAULT_GPG_PATH, 256);
+  sl_strlcpy (path,  DEFAULT_GPG_PATH,  256);
   sl_strlcpy (cc1,   _("--status-fd"),  32);
   sl_strlcpy (cc2,   _("--verify"),     32);
+  sl_strlcpy (cc3,   _("--homedir"),    32);
+  /* sl_strlcpy (cc4,   sh.effective.home, SH_PATHBUF+32); */
+  sl_strlcpy (cc4,   homedir,           SH_PATHBUF+32);
+  sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32); 
 
+  /* fprintf(stderr, "YULE: homedir=%s\n", homedir); */
+
+#if defined(SH_WITH_SERVER)
+  if (0 == sl_ret_euid())   /* privileges not dropped yet */
+    {
+      struct stat lbuf;
+      int         status_stat = 0;
+      struct passwd * tempres = getpwnam(DEFAULT_IDENT);
+      if (!tempres)
+	tempres = sh_getpwnam(DEFAULT_IDENT);
+      if (!tempres)
+	{
+	  dlog(1, FIL__, __LINE__, 
+	       _("User %s does not exist. Please add the user to your system.\n"), 
+	       DEFAULT_IDENT);
+	  status_stat = -1;
+	}
+      if (!tempres->pw_dir || tempres->pw_dir[0] == '\0')
+	{
+	  dlog(1, FIL__, __LINE__, 
+	       _("User %s does not have a home directory.\nPlease add the home directory for this user to your system.\n"), 
+	       DEFAULT_IDENT);
+	  status_stat = -2;
+	}
+      if (status_stat == 0)
+	{
+	  sl_strlcpy (cc4, tempres->pw_dir, SH_PATHBUF+32); 
+	  sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32); 
+	  status_stat =  retry_lstat(FIL__, __LINE__, cc4, &lbuf);
+	  if (status_stat == -1)
+	    {
+	      dlog(1, FIL__, __LINE__, 
+		   _("Gnupg directory %s for user %s\ndoes not exist or is not accessible.\nPlease add the directory and put the keyring (pubring.gpg) there\nto verify the configuration file.\n"),
+		   cc4, DEFAULT_IDENT);
+	      status_stat = -3;
+	    }
+	}
+      if (status_stat == 0 && lbuf.st_uid != tempres->pw_uid)
+	{
+	  dlog(1, FIL__, __LINE__, 
+	       _("Gnupg directory %s\nis not owned by user %s.\n"), 
+	       cc4, DEFAULT_IDENT);
+	  status_stat = -4;
+	}
+      if (status_stat == 0)
+	{
+	  sl_strlcat (cc4,   _("/pubring.gpg"),      SH_PATHBUF+32); 
+	  status_stat =  retry_lstat(FIL__, __LINE__, cc4, &lbuf);
+	  if (status_stat == -1)
+	    {
+	      dlog(1, FIL__, __LINE__, 
+		   _("Gnupg public keyring %s for user %s\ndoes not exist or is not accessible.\nPlease add the directory and put the keyring (pubring.gpg) there\nto verify the configuration file.\n"),
+		   cc4, DEFAULT_IDENT);
+	      status_stat = -5;
+	    }
+	}
+      if (status_stat == 0 && lbuf.st_uid != tempres->pw_uid)
+	{
+	  dlog(1, FIL__, __LINE__, 
+	       _("Gnupg public keyring %s\nis not owned by user %s.\n"), 
+	       cc4, DEFAULT_IDENT);
+	  status_stat = -6;
+	}
+      if (status_stat != 0)
+	{
+	  sh_error_handle((-1), FIL__, __LINE__, status_stat, MSG_EXIT_ABORT1, 
+			  sh.prg_name);
+	  aud_exit (FIL__, __LINE__, EXIT_FAILURE);
+	}
+      sl_strlcpy (cc4, tempres->pw_dir, SH_PATHBUF+32); 
+      sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32); 
+    }
+#endif
 
   arg[0] = path; 
   arg[1] = cc1;
   arg[2] = "1";
   arg[3] = cc2;
-  arg[4] = cc0;
-  arg[5] = NULL;
+  arg[4] = cc3;
+  arg[5] = cc4;
+  arg[6] = cc5;
+  arg[7] = cc0;
+  arg[8] = NULL;
 
   /* catch 'unused parameter' compiler warning
    */
@@ -200,7 +369,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
    */
   if (sh.effective.home != NULL)
     {
-      envp[0] = sl_malloc (sl_strlen(sh.effective.home) + 6);
+      envp[0] = malloc (sl_strlen(sh.effective.home) + 6); /* free() ok   */
       if (envp[0] != NULL)
 	sprintf (envp[0], "HOME=%s",                     /* known to fit  */
 		 sh.effective.home); 
@@ -216,7 +385,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
   if (aud_pipe(FIL__, __LINE__, pipedes) < 0) 
     {
       if (envp[0] != NULL) 
-	my_free(envp[0]);
+	free(envp[0]);
       SL_RETURN( (NULL), _("sh_gpg_popen"));
     }
   
@@ -229,7 +398,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
       close(pipedes[0]);
       close(pipedes[1]);
       if (envp[0] != NULL) 
-	my_free(envp[0]);
+	free(envp[0]);
       SL_RETURN( (NULL), _("sh_gpg_popen"));
     }
 
@@ -242,6 +411,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 			pipedes[STDOUT_FILENO], STDOUT_FILENO) < 0)
 	{
 	  TPT(((0), FIL__, __LINE__, _("msg=<dup2 on pipe failed>\n")));
+	  dlog(1, FIL__, __LINE__, _("Internal error: dup2 failed\n"));
 	  aud__exit(FIL__, __LINE__, EXIT_FAILURE);
 	}
       
@@ -257,6 +427,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 	  if (retry_aud_dup2(FIL__, __LINE__, fd, STDIN_FILENO) < 0)
 	    {
 	      TPT(((0), FIL__, __LINE__, _("msg=<dup2 on fd failed>\n")));
+	      dlog(1, FIL__, __LINE__, _("Internal error: dup2 failed\n"));
 	      aud__exit(FIL__, __LINE__, EXIT_FAILURE);
 	    }
 	}
@@ -264,6 +435,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
       if (retry_aud_dup2(FIL__, __LINE__, fd, STDIN_FILENO) < 0)
 	{
 	  TPT(((0), FIL__, __LINE__, _("msg=<dup2 on fd failed>\n")));
+	  dlog(1, FIL__, __LINE__, _("Internal error: dup2 failed\n"));
 	  aud__exit(FIL__, __LINE__, EXIT_FAILURE);
 	}
 #endif
@@ -281,9 +453,22 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
        * gpg.
        */
       memset (skey, '\0', sizeof(sh_key_t));
-      sl_set_suid(); 
       aud_setuid(FIL__, __LINE__, geteuid());
       
+      PDBGC_OPEN;
+      PDBGC_D((int)getuid());
+      PDBGC_D((int)geteuid());
+
+      {
+	int i = 0;
+	while (arg[i] != NULL)
+	  {
+	    PDBGC_S(arg[i]);
+	    ++i;
+	  }
+      }
+      PDBGC_CLOSE;
+
       /* exec the program */
 
 #if defined(__linux__) && defined(HAVE_GPG_CHECKSUM)
@@ -307,6 +492,8 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 	  fcntl  (pfd, F_SETFD, FD_CLOEXEC);
 	  retry_aud_execve (FIL__, __LINE__,  pname, arg, envp);
 	      
+	  dlog(1, FIL__, __LINE__, _("Unexpected error: execve %s failed\n"),
+	       pname);
 	  /* failed 
 	   */
 	  aud__exit(FIL__, __LINE__, EXIT_FAILURE);
@@ -334,8 +521,9 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 #else
 	  checkfd = sl_open_fastread(DEFAULT_GPG_PATH, SL_NOPRIV);
 #endif
-	  if (0 != sh_gpg_checksum(checkfd, 0))
+	  if (0 != sh_gpg_checksum(checkfd, 0)) {
 	    aud__exit(FIL__, __LINE__, EXIT_FAILURE);
+	  }
 	  sl_close(checkfd);
 	}
 #endif
@@ -343,6 +531,8 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 
 #if defined(WITH_GPG)
       retry_aud_execve (FIL__, __LINE__, DEFAULT_GPG_PATH, arg, envp);
+      dlog(1, FIL__, __LINE__, _("Unexpected error: execve %s failed\n"),
+	   DEFAULT_GPG_PATH);
 #elif defined(WITH_PGP)
       retry_aud_execve (FIL__, __LINE__, DEFAULT_PGP_PATH, arg, envp);
 #endif
@@ -350,6 +540,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
       /* failed 
        */
       TPT(((0), FIL__, __LINE__, _("msg=<execve failed>\n")));
+      dlog(1, FIL__, __LINE__, _("Unexpected error: execve failed\n"));
       aud__exit(FIL__, __LINE__, EXIT_FAILURE);
     }
 
@@ -357,13 +548,13 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
    */
 
   if (envp[0] != NULL) 
-    my_free(envp[0]);
+    free(envp[0]);
 
   close (pipedes[STDOUT_FILENO]);
   retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
   retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFL,  O_NONBLOCK);
 
-  MBLK( outf = fdopen (pipedes[STDIN_FILENO], "r"); )
+  outf = fdopen (pipedes[STDIN_FILENO], "r");
   
   if (outf == NULL) 
     {
@@ -384,7 +575,7 @@ static int sh_gpg_pclose (sh_gpg_popen_t *source)
   
   SL_ENTER(_("sh_gpg_pclose"));
 
-  MBLK( status = fclose(source->pipe); )
+  status = fclose(source->pipe);
   if (status)
     SL_RETURN( (-1), _("sh_gpg_pclose"));
   
@@ -397,7 +588,8 @@ static int sh_gpg_pclose (sh_gpg_popen_t *source)
 }
  
 static
-int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
+int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp, 
+			   char * homedir, int whichfile)
 {
   struct stat buf;
   char line[256];
@@ -427,6 +619,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
 		      sh_error_message(status), DEFAULT_GPG_PATH);
       SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
+
   if (0 != tf_trust_check (DEFAULT_GPG_PATH, SL_YESPRIV))
     SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
 
@@ -479,7 +672,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
 
   fflush(NULL);
  
-  source.pipe   = sh_gpg_popen  ( &source, fd, 0, NULL );
+  source.pipe   = sh_gpg_popen  ( &source, fd, 0, NULL, homedir );
 
   if (NULL == source.pipe)
     {
@@ -498,6 +691,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
   while (NULL != fgets(line, 255, source.pipe))
     {
 
+      TPT(((0), FIL__, __LINE__, _("msg=<gpg out: %s>\n"), line));
       if (line[strlen(line)-1] == '\n')
 	line[strlen(line)-1] = ' ';
       sh_error_handle(SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN, 
@@ -511,8 +705,38 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
        */
       if (0 == sl_strncmp(_("BADSIG"), &line[9], 6) ||
 	  0 == sl_strncmp(_("ERRSIG"), &line[9], 6) ||
+	  0 == sl_strncmp(_("NO_PUBKEY"), &line[9], 6) ||
+	  0 == sl_strncmp(_("NODATA"), &line[9], 6) ||
 	  0 == sl_strncmp(_("SIGEXPIRED"), &line[9], 6))
 	{
+	  if      (0 == sl_strncmp(_("BADSIG"), &line[9], 6)) {
+	    dlog(1, FIL__, __LINE__, 
+		 _("%s file is signed, but the signature is invalid."),
+		 ((whichfile == 1) ? _("Configuration") : _("Database")));
+	  } 
+	  else if (0 == sl_strncmp(_("NO_PUBKEY"), &line[9], 6)) {
+	    dlog(1, FIL__, __LINE__, 
+		 _("%s file is signed, but the public key to verify the signature is not in my keyring %s/.gnupg/pubring.asc."), 
+		 ((whichfile == 1) ? _("Configuration") : _("Database")),
+		 homedir);
+	  }
+	  else if (0 == sl_strncmp(_("ERRSIG"), &line[9], 6)) {
+	    dlog(1, FIL__, __LINE__, 
+		 _("%s file is signed, but the public key to verify the signature is not in my keyring %s/.gnupg/pubring.asc."), 
+		 ((whichfile == 1) ? _("Configuration") : _("Database")),
+		 homedir);
+	  }
+	  else if (0 == sl_strncmp(_("SIGEXPIRED"), &line[9], 6)) {
+	    dlog(1, FIL__, __LINE__, 
+		 _("%s file is signed, but the public key to verify the signature has expired."), 
+		 ((whichfile == 1) ? _("Configuration") : _("Database")));
+	  }
+	  else if (0 == sl_strncmp(_("NODATA"), &line[9], 6)) {
+	    dlog(1, FIL__, __LINE__, 
+		 _("%s file is not signed."), 
+		 ((whichfile == 1) ? _("Configuration") : _("Database")));
+	  }
+
 	  have_fp = BAD; have_id = BAD;
 	  break;
 	}
@@ -564,7 +788,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
 #ifdef WITH_PGP
   /* get the fingerprint */
 
-  source.pipe   = sh_gpg_popen  ( &source, fd, 1,  sign_id);
+  source.pipe   = sh_gpg_popen  ( &source, fd, 1,  sign_id, homedir);
   if (NULL == source.pipe)
     {
       sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN, 
@@ -657,13 +881,23 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
   char  * tmp;
   char  * tmp2;
 
+  char  * homedir = sh.effective.home;
+#if defined(SH_WITH_SERVER)
+  struct passwd * tempres;
+#endif
+#ifdef USE_FINGERPRINT
+#include "sh_gpg_fp.h"
+#endif
+
   SL_ENTER(_("sh_gpg_check_sign"));
 
+
   if (what == 0 || what == 1)
     fd1 = get_the_fd(file_1);
   if (what == 0 || what == 2)
     fd2 = get_the_fd(file_2);
 
+
   if (fd1 < 0 || fd2 < 0)
     {
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD1 = %d>\n"), fd1));
@@ -678,7 +912,17 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
   if (what == 0 || what == 1)
     {
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD1 = %d>\n"), fd1));
-      status = sh_gpg_check_file_sign (fd1, gp.conf_id, gp.conf_fp);
+#if defined(SH_WITH_SERVER)
+      tempres = getpwnam(DEFAULT_IDENT);
+      if (!tempres)
+	tempres = sh_getpwnam(DEFAULT_IDENT);
+      if ((tempres != NULL) && (0 == sl_ret_euid()))
+	{
+	  /* privileges not dropped yet*/
+	  homedir = tempres->pw_dir;
+	}
+#endif
+      status = sh_gpg_check_file_sign(fd1, gp.conf_id, gp.conf_fp, homedir, 1);
       TPT(((0), FIL__, __LINE__, _("msg=<CONF SIGUSR: |%s|>\n"), gp.conf_id));
       TPT(((0), FIL__, __LINE__, _("msg=<CONF SIGFP:  |%s|>\n"), gp.conf_fp));
     }
@@ -686,7 +930,17 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
   if ((what == 0 && SH_GPG_OK == status) || what == 2)
     {
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD2 = %d>\n"), fd2));
-      status = sh_gpg_check_file_sign (fd2, gp.data_id, gp.data_fp);
+#if defined(SH_WITH_SERVER)
+      tempres = getpwnam(DEFAULT_IDENT);
+      if (!tempres)
+	tempres = sh_getpwnam(DEFAULT_IDENT);
+      if ((tempres != NULL) && (0 == sl_ret_euid()))
+	{
+	  /* privileges not dropped yet*/
+	  homedir = tempres->pw_dir;
+	}
+#endif
+      status = sh_gpg_check_file_sign(fd2, gp.data_id, gp.data_fp, homedir, 2);
       TPT(((0), FIL__, __LINE__, _("msg=<DATA SIGUSR: |%s|>\n"), gp.data_id));
       TPT(((0), FIL__, __LINE__, _("msg=<DATA SIGFP:  |%s|>\n"), gp.data_fp));
     }
@@ -696,6 +950,19 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 #ifdef USE_FINGERPRINT
       if ((sl_strcmp(SH_GPG_FP, gp.conf_fp) == 0))
 	{
+	  int i;
+
+	  for(i = 0; i < (int) sl_strlen(gp.conf_fp); ++i)
+	    {
+	      if (gpgfp[i] != gp.conf_fp[i]) 
+		{
+		  sh_error_handle(SH_ERR_SEVERE, FIL__, __LINE__, 0, 
+				  MSG_E_GPG_FP, 
+				  gpgfp, gp.conf_fp);
+		  break;
+		}
+	    }
+
 	  if (smsg == S_FALSE)
 	    {
 	      tmp  = sh_util_safe_name(gp.conf_id);
@@ -767,9 +1034,28 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 
   if (status != SH_GPG_OK) 
     {
+      uid_t   e_uid  = sl_ret_euid();
+      char  * e_home = sh.effective.home;
+
+#if defined(SH_WITH_SERVER)
+      struct passwd * e_tempres = getpwnam(DEFAULT_IDENT);
+      if (!e_tempres)
+	e_tempres = sh_getpwnam(DEFAULT_IDENT);
+      if ((e_tempres != NULL) && (0 == sl_ret_euid()))   
+	{
+	  /* privileges not dropped yet */
+	  e_uid  = e_tempres->pw_uid;
+	  e_home = e_tempres->pw_dir;
+	}
+#endif
       dlog(1, FIL__, __LINE__, 
-	   _("The signature of the configuration file or the file signature database\ncould not be verified.Possible reasons are:\n - the signature key is not in the private keyring of UID %d,\n - there is no keyring in %s/.gnupg, or\n - the file is not signed - did you move /filename.asc to /filename ?\nTo create a signed file, use:\n   gpg -a --clearsign --not-dash-escaped FILE\n   mv FILE.asc FILE\n"), 
-	   (int) sl_ret_euid(), sh.effective.home);
+	   _("The signature of the configuration file or the file signature database\ncould not be verified. Possible reasons are:\n - gpg binary (%s) not found\n - invalid signature\n - the signature key is not in the private keyring of UID %d,\n - there is no keyring in %s/.gnupg, or\n - the file is not signed - did you move /filename.asc to /filename ?\nTo create a signed file, use (remove old signatures before):\n   gpg -a --clearsign --not-dash-escaped FILE\n   mv FILE.asc FILE\n"),
+#if defined(WITH_GPG) 
+	   DEFAULT_GPG_PATH,
+#else
+	   DEFAULT_PGP_PATH,
+#endif
+	   (int) e_uid, e_home);
     }
 
   TPT(((0), FIL__, __LINE__, _("msg=<Status = %d>\n"), status));
