@@ -60,10 +60,25 @@
 
 #undef  FIL__
 #define FIL__  _("sh_mail.c")
+#undef  GOOD
+#undef  BAD
 
-static int failedMail = GOOD;
+static int failedMail = SL_FALSE;
 
+/* MX Resolver Struct
+ */
+typedef struct mx_ {
+  int    pref;
+  char * address;
+} mx;
 
+typedef struct dnsrep_ {
+  int    count;
+  mx   * reply;
+} dnsrep;
+
+static int free_mx (dnsrep * answers);
+static dnsrep * return_mx (char *domain);
 
 /*********************************************
  *  utility function for verifying mails
@@ -103,6 +118,11 @@ int sh_mail_sigverify (char * s)
 
   /* open the file, then check it 
    */
+  if (sl_is_suid())
+    {
+      fprintf(stderr, _("Cannot open file %s in suid mode\n"), s);
+      _exit (EXIT_FAILURE);
+    }
   if ( SL_ISERROR(fd = sl_open_read (s, SL_NOPRIV)))
     {
       fprintf(stderr, _("Could not open file %s\n"), s);
@@ -251,7 +271,6 @@ void reset_count_dev_mail(void)
 
 int sh_mail_setaddress (char * address)
 {
-  unsigned long  numcop;
   char     *     p;
 
   SL_ENTER(_("sh_mail_setaddress"));
@@ -266,19 +285,21 @@ int sh_mail_setaddress (char * address)
       else
 	p = &sh.srvmail.alt[address_num*64];
 
-      numcop = sl_strlcpy (p, address, 64);
+      sl_strlcpy (p, address, 64);
       
-      if (SL_ISERROR(numcop))
+      if ((p == NULL) || ( sl_strlen(address) != sl_strlen(p)))
 	{
 	  memset(p, '\0', 64);
 	  SL_RETURN ( (-1), _("sh_mail_setaddress"));
 	}
       address_list[address_num] = p;
+#if 0
       if (!sl_is_suid())
 	{
 	  TPT(( 0, FIL__, __LINE__, _("msg=<address_list[%d] = %s>\n"), 
 		address_num, address_list[address_num]));
 	}
+#endif
       if (setaddress_compiled == S_TRUE)
 	{
 	  ++address_num;
@@ -307,8 +328,6 @@ int sh_mail_setaddress_int (char * address)
   setaddress_compiled = S_FALSE;
   SL_RETURN(i, _("sh_mail_setaddress_int"));
 }
-  
-
 
 int sh_mail_setNum (char * str)
 {
@@ -345,6 +364,12 @@ int set_mail_subject (char * str)
   if (mail_subject != NULL)
     SH_FREE(mail_subject);
 
+  if (0 == sl_strncmp(str, _("NULL"), 4))
+    {
+      mail_subject = NULL;
+      SL_RETURN( 0, _("set_mail_subject"));
+    }
+
   mail_subject = sh_util_strdup(str);
   SL_RETURN( (0), _("set_mail_subject"));
 }
@@ -372,15 +397,25 @@ void sh_mail_emptystack (void)
   SL_RET0(_("sh_mail_emptystack"));
 }
 
+/* insert "\r\n" after each 998 char
+ */
+static char * split_string(char * str);
+
 void sh_mail_pushstack (char * msg)
 {
+  char * p;
+
   SL_ENTER(_("sh_mail_pushstack"));
 
-  if (msg == NULL || failedMail == BAD || sh.srvmail.name[0] == '\0') 
+  if (msg == NULL || failedMail == SL_TRUE || sh.srvmail.name[0] == '\0') 
     SL_RET0(_("sh_mail_pushstack"));
 
+#if 0
   if (msg != NULL && sl_strlen(msg) > 998)  /* RFC 2822 */
     msg[998] = '\0';
+#endif
+
+  p = split_string(msg);
 
   if (fifo_mail == NULL)
     {
@@ -388,8 +423,10 @@ void sh_mail_pushstack (char * msg)
       fifo_init(fifo_mail);
     }
 
-  push_list (fifo_mail, msg);
-  ++sh.mailNum.alarm_last;
+  if (0 == push_list (fifo_mail, p))
+    ++sh.mailNum.alarm_last;
+
+  SH_FREE(p);
 
   if (sh.mailNum.alarm_last >= sh.mailNum.alarm_interval)
     {
@@ -400,6 +437,7 @@ void sh_mail_pushstack (char * msg)
   SL_RET0(_("sh_mail_pushstack"));
 }
 
+
 /* The mailer.
  */
 static int sh_mail_end_conn (FILE * connfile);
@@ -412,11 +450,13 @@ void sh_mail_get_subject(char * message,
     { 'T', S_FMT_TIME,    0, 0, NULL},
     { 'H', S_FMT_STRING,  0, 0, NULL},
     { 'M', S_FMT_STRING,  0, 0, NULL},
+    { 'S', S_FMT_STRING,  0, 0, NULL},
     {'\0', S_FMT_ULONG,   0, 0, NULL},
   };
 
   char * p;
   char * mptr;
+  char   sev[8];
 
   SL_ENTER(_("sh_mail_get_subject"));
 
@@ -442,6 +482,38 @@ void sh_mail_get_subject(char * message,
   else
     rep_serv_tab[2].data_str   = message;
 
+  mptr = sl_strstr(message, _("sev="));
+  if (mptr)
+    {
+      mptr += 5;
+      sev[0] = *mptr; ++mptr;
+      sev[1] = *mptr; ++mptr;
+      sev[2] = *mptr; ++mptr;
+      sev[3] = *mptr; ++mptr;
+      sev[4] = '\0';
+    }
+  else
+    {
+      mptr = message;
+      sev[0] = *mptr; ++mptr;
+      sev[1] = *mptr; ++mptr;
+      sev[2] = *mptr; ++mptr;
+      sev[3] = *mptr; ++mptr;
+      if (*mptr == ' ') {
+	sev[4] = '\0';
+      } else {
+	sev[4] = *mptr; ++mptr;
+	if (*mptr == ' ') {
+	  sev[5] = '\0';
+	} else {
+	  sev[5] = *mptr;
+	  sev[6] = '\0';
+	}
+      }
+    }
+  rep_serv_tab[3].data_str   = sev;
+
+
   p = sh_util_formatted(mail_subject, rep_serv_tab);
   sl_strlcat(mheader, p, len);
   SH_FREE(p);
@@ -453,6 +525,26 @@ void sh_mail_get_subject(char * message,
 #include <sys/mman.h>
 #endif
 
+static char * sh_mail_realloc (char * inbuf, size_t * insize, int increase)
+{
+  size_t newsize = (*insize) + increase + 1;
+  char * outbuf;
+
+  SL_ENTER(_("sh_mail_realloc"));
+
+  outbuf = SH_ALLOC(newsize);
+  MLOCK(outbuf, newsize);
+  sl_strlcpy(outbuf, inbuf, newsize);
+
+  memset (inbuf, '\0', (*insize));
+  MUNLOCK(inbuf, (*insize));
+  SH_FREE(inbuf);
+
+  *insize = newsize;
+
+  SL_RETURN( (outbuf), _("sh_mail_realloc"));
+}
+
 int sh_mail_msg (char * message)
 {
     char         subject[32+32+SH_MINIBUF+2+3+SH_PATHBUF];
@@ -462,6 +554,7 @@ int sh_mail_msg (char * message)
     char       * popMsg;
     int          status = 0, errcount;
     int          i;
+    int          num_popped = 0;
 
     char       * bufcompress;
     static int   failcount = 0;
@@ -481,6 +574,13 @@ int sh_mail_msg (char * message)
     SH_FIFO * fifo_temp = NULL;
 
     char    * theSig;
+    char    * theMsg = NULL;
+
+    /* #define SH_MAILBUF (256)    */
+#define SH_MAILBUF (8*4096)
+
+    size_t    msgbufsize = SH_MAILBUF;
+    size_t    combufsize = SH_MAILBUF;
 
     SL_ENTER(_("sh_mail_msg"));
 
@@ -489,20 +589,17 @@ int sh_mail_msg (char * message)
 
     /* Return if we cannot mail.
      */
-    if (failedMail == BAD) 
+    if (failedMail == SL_TRUE) 
       SL_RETURN((-1), _("sh_mail_msg"));
 
-    if (failedMail == GOOD && address_list[0] == NULL)
+    if (failedMail == SL_FALSE && address_list[0] == NULL)
       {
 	TPT((0, FIL__, __LINE__, 
 	     _("msg=<Mail error: no recipient address.>\n")));
-	failedMail = BAD;
+	failedMail = SL_TRUE;
 	SL_RETURN((-1), _("sh_mail_msg"));
       }
 
-    if (message != NULL && sl_strlen(message) > 998)  /* RFC 2822 */
-      message[998] = '\0';
-
     if ( (success_time > 0) && (fail_time > 0) &&
 	 (time(NULL) - success_time) > 3600*SH_MAX_FAIL)
       {
@@ -511,23 +608,31 @@ int sh_mail_msg (char * message)
 			 _("mail"), address_list[0]);
 	m_block = 0;
 	sh_mail_emptystack();
-	failedMail = BAD;
+	sh.mailNum.alarm_last = 0;
+	failedMail = SL_TRUE;
 	SL_RETURN((-1), _("sh_mail_msg"));
       }
 
     /* Try at most each hour.
      */
-    if (fail_time > 0 &&  (time(NULL) - fail_time) < 3600)
+    if ((fail_time > 0) && (time(NULL) - fail_time) < 3600)
       {
-	/* -- Save for later. -- 
-	 */
-	sh_mail_pushstack (message);
-	++failcount;
-
-	SL_RETURN((-1), _("sh_mail_msg"));
+	if (failcount > 3)
+	  {
+	    /* -- Save for later. -- 
+	     */
+	    sh_mail_pushstack (message);
+	    ++failcount;
+	    
+	    SL_RETURN((-1), _("sh_mail_msg"));
+	  }
+	else
+	  {
+	    retry_msleep(2, 0);
+	    ++failcount;
+	  }
       }
 
-
     /* -- Reset time of last failure. --
      */
     fail_time = 0;
@@ -541,19 +646,39 @@ int sh_mail_msg (char * message)
 
     /* ---------  Build complete message. ------------------------ */
 
+    theMsg = split_string(message);
 
     /* ---------- Header  ---------------------------------------- */
 
     if (mail_subject == NULL)
       {
-	sl_strlcpy(mheader, _("Subject: "),            sizeof(mheader));
-	sl_strlcat(mheader, sh_unix_time (0),          sizeof(mheader));
-	sl_strlcat(mheader, " ",                       sizeof(mheader));
-	sl_strlcat(mheader, sh.host.name,              sizeof(mheader));
+	sl_strlcpy(mheader, _("Subject: "),            sizeof(mheader)-5);
+	sl_strlcat(mheader, sh_unix_time (0),          sizeof(mheader)-5);
+	sl_strlcat(mheader, " ",                       sizeof(mheader)-5);
+	sl_strlcat(mheader, sh.host.name,              sizeof(mheader)-5);
       }
     else
       {
-	sh_mail_get_subject(message, mheader, sizeof(mheader));
+	
+	if (message == NULL)
+	  {
+	    theMsg  = pop_list(fifo_mail);
+	    message = theMsg;
+	    if (message)
+	      --sh.mailNum.alarm_last;
+	  }
+	
+	if (message)
+	  {
+	    sh_mail_get_subject(message, mheader, sizeof(mheader)-5);
+	  }
+	else
+	  {
+	    sl_strlcpy(mheader, _("Subject: "),            sizeof(mheader)-5);
+	    sl_strlcat(mheader, sh_unix_time (0),          sizeof(mheader)-5);
+	    sl_strlcat(mheader, " ",                       sizeof(mheader)-5);
+	    sl_strlcat(mheader, sh.host.name,              sizeof(mheader)-5);
+	  }
       }
 
     /* RFC 821: Header is terminated by an empty line
@@ -567,26 +692,33 @@ int sh_mail_msg (char * message)
     sl_strlcat(subject, sh.host.name,              sizeof(subject));
     sl_strlcat(subject, "\r\n",                    sizeof(subject));
 
-#define SH_MAILBUF (4*4096)
 
-    mailMsg     = (char *) SH_ALLOC (SH_MAILBUF);
-    bufcompress = (char *) SH_ALLOC (SH_MAILBUF);
+    mailMsg     = (char *) SH_ALLOC (msgbufsize);
+    bufcompress = (char *) SH_ALLOC (combufsize);
 
-    MLOCK(mailMsg     , SH_MAILBUF);
-    MLOCK(bufcompress , SH_MAILBUF);
+    MLOCK(mailMsg     , msgbufsize);
+    MLOCK(bufcompress , combufsize);
 
-    sl_strlcpy(mailMsg, mheader, SH_MAILBUF);
+    sl_strlcpy(mailMsg, mheader, msgbufsize);
 
 
-    sl_strlcat(mailMsg, _("-----BEGIN MESSAGE-----\r\n"), SH_MAILBUF);
-    sl_strlcat(mailMsg, subject, SH_MAILBUF);
-    sh_util_compress (bufcompress, subject, SH_MAILBUF-KEY_LEN-1);
+    sl_strlcat(mailMsg, _("-----BEGIN MESSAGE-----\r\n"), msgbufsize);
+    sl_strlcat(mailMsg, subject, msgbufsize);
+    sh_util_compress (bufcompress, subject, (combufsize - KEY_LEN - 1));
     if (message != NULL)
       {
-	sl_strlcat(mailMsg, message, SH_MAILBUF-(4*KEY_LEN));
-	sl_strlcat(mailMsg,  "\r\n", SH_MAILBUF-(4*KEY_LEN));
+	if ((sl_strlen(theMsg) + sl_strlen(mailMsg) + 1) > 
+	    (msgbufsize-(4*KEY_LEN)))
+	  {
+	    mailMsg     = sh_mail_realloc(mailMsg,     
+					  &msgbufsize, sl_strlen(theMsg)+2);
+	    bufcompress = sh_mail_realloc(bufcompress, 
+					  &combufsize, sl_strlen(theMsg));
+	  }
+	sl_strlcat(mailMsg,  theMsg, msgbufsize-(4*KEY_LEN));
+	sl_strlcat(mailMsg,  "\r\n", msgbufsize-(4*KEY_LEN));
 
-	sh_util_compress (bufcompress, message, SH_MAILBUF-KEY_LEN-1);
+	sh_util_compress (bufcompress,  theMsg, combufsize-KEY_LEN-1);
       }
 
     if (sh.mailNum.alarm_last > 0)
@@ -597,18 +729,31 @@ int sh_mail_msg (char * message)
 	while ( NULL != (popMsg = pop_list(fifo_mail)) )
 	  {
 	    push_list (fifo_temp, popMsg);
-	    sl_strlcat(mailMsg, popMsg, SH_MAILBUF-(4*KEY_LEN));
-	    sl_strlcat(mailMsg, "\r\n", SH_MAILBUF-(4*KEY_LEN));
-	    sh_util_compress (bufcompress, popMsg, SH_MAILBUF-KEY_LEN-1);
+
+	    if ((sl_strlen(popMsg) + sl_strlen(mailMsg) + 1) > 
+		(msgbufsize-(4*KEY_LEN)))
+	      {
+		mailMsg     = sh_mail_realloc(mailMsg,     
+					      &msgbufsize, 
+					      sl_strlen(popMsg)+2);
+		bufcompress = sh_mail_realloc(bufcompress, 
+					      &combufsize, 
+					      sl_strlen(popMsg));
+	      }
+
+	    sl_strlcat(mailMsg, popMsg, msgbufsize-(4*KEY_LEN));
+	    sl_strlcat(mailMsg, "\r\n", msgbufsize-(4*KEY_LEN));
+	    sh_util_compress (bufcompress, popMsg, combufsize-KEY_LEN-1);
 
 	    SH_FREE(popMsg);
 	    --sh.mailNum.alarm_last;
+	    ++num_popped;
 	  }
       }
 
     /* ------ signature block ------------------------------------ */
     
-    sl_strlcat(mailMsg, _("-----BEGIN SIGNATURE-----\r\n"), SH_MAILBUF);
+    sl_strlcat(mailMsg, _("-----BEGIN SIGNATURE-----\r\n"), msgbufsize);
 
     /* Generate new signature key.
      */
@@ -632,7 +777,7 @@ int sh_mail_msg (char * message)
 				 bufcompress, sl_strlen(bufcompress));
 	sl_strlcat (mailMsg, 
 		    theSig,
-		    SH_MAILBUF);
+		    msgbufsize);
       }
     else
       {
@@ -645,11 +790,11 @@ int sh_mail_msg (char * message)
 	BREAKEXIT(sh_util_encode);
         sh_util_encode(skey->crypt, bufcompress, 0, 'A');
 
-        sl_strlcat (mailMsg, skey->crypt, SH_MAILBUF);
+        sl_strlcat (mailMsg, skey->crypt, msgbufsize);
         memset (skey->crypt, '\0', KEY_LEN);
         isfirst = 0;
       }
-    sl_strlcat (mailMsg, "\r\n", SH_MAILBUF);
+    sl_strlcat (mailMsg, "\r\n", msgbufsize);
 
     /* X(n) -> X(n-1)
      */
@@ -658,10 +803,10 @@ int sh_mail_msg (char * message)
     sprintf(subject, _("%06d %010ld::%s\r\n"),         /* known to fit  */
             mailcount, (long) id_audit, sh.host.name);
 
-    sl_strlcat (mailMsg, subject, SH_MAILBUF);
+    sl_strlcat (mailMsg, subject, msgbufsize);
     ++mailcount;
 
-    sl_strlcat (mailMsg, _("-----END MESSAGE-----"), SH_MAILBUF);
+    sl_strlcat (mailMsg, _("-----END MESSAGE-----"), msgbufsize);
 
 
 
@@ -697,8 +842,17 @@ int sh_mail_msg (char * message)
 				 _("mail"), address_list[i]);
 		m_block = 0;
 		++errcount;
-		if (connfile != NULL) 
-		  fclose (connfile);
+		++sh.statistics.mail_failed;
+	      }
+	    else
+	      {
+		++sh.statistics.mail_success;
+	      }
+
+	    if (connfile != NULL)
+	      {
+		fclose (connfile);
+		connfile = NULL;
 	      }
 	    ++i;
 	  }
@@ -721,18 +875,27 @@ int sh_mail_msg (char * message)
 			     _("mail"), address_list[0]);
 	    m_block = 0;
 	    errcount = address_num;
-	    if (connfile != NULL) 
-	      fclose (connfile);
+	    ++sh.statistics.mail_failed;
+	  }
+	else
+	  {
+	    ++sh.statistics.mail_success;
+	  }
+
+	if (connfile != NULL)
+	  {
+	    fclose (connfile);
+	    connfile = NULL;
 	  }
       }
 
     
-    memset (bufcompress, '\0', SH_MAILBUF);
-    MUNLOCK(bufcompress , SH_MAILBUF);
+    memset (bufcompress, '\0', combufsize);
+    MUNLOCK(bufcompress , combufsize);
     SH_FREE(bufcompress);
 
-    memset (mailMsg, '\0', SH_MAILBUF);
-    MUNLOCK(mailMsg , SH_MAILBUF);
+    memset (mailMsg, '\0', msgbufsize);
+    MUNLOCK(mailMsg , msgbufsize);
     SH_FREE(mailMsg);
 
     /* --- Stay responsible for delivery in case of failure --- */
@@ -741,14 +904,17 @@ int sh_mail_msg (char * message)
       {
         while ( (NULL != (popMsg = pop_list(fifo_temp))) )
           {
-            push_list (fifo_mail, popMsg);
-	    ++sh.mailNum.alarm_last;
+            if (0 == push_list (fifo_mail, popMsg))
+	      ++sh.mailNum.alarm_last;
             SH_FREE(popMsg);
           }
 	if (message != NULL)
 	  {
+#if 0
 	    push_list (fifo_mail, message);
-	    ++sh.mailNum.alarm_last;
+#endif
+	    if (0 == push_list (fifo_mail,  theMsg))
+	      ++sh.mailNum.alarm_last;
 	  }
       }
     else if (fifo_temp != NULL)
@@ -761,7 +927,13 @@ int sh_mail_msg (char * message)
     if (fifo_temp != NULL)
       SH_FREE(fifo_temp);
 
-    if (connfile != NULL) fclose (connfile);
+    /*
+    if (connfile != NULL) 
+      fclose (connfile);
+    */
+
+    if (theMsg != NULL)
+      SH_FREE(theMsg);
 
     /* --- Reset signal. ---
      */
@@ -773,6 +945,7 @@ int sh_mail_msg (char * message)
 	SL_RETURN((-1), _("sh_mail_msg"));
       }
     success_time = time(NULL);
+    failcount = 0;
 
     SL_RETURN((0), _("sh_mail_msg"));
 }
@@ -786,12 +959,18 @@ int sh_mail_msg (char * message)
  */
 
 #include <ctype.h>
+#ifdef  HOST_IS_HPUX
+#define _XOPEN_SOURCE_EXTENDED
+#endif
 #include <netdb.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
+#define SH_NEED_GETHOSTBYXXX
+#include "sh_static.h"
+
 /* missing on HP-UX 10.20 */
 #ifndef IPPORT_SMTP
 #define IPPORT_SMTP 25
@@ -811,10 +990,16 @@ int sh_mail_set_relay (char * str_s)
     SL_RETURN( -1, _("sh_mail_set_relay"));
 
   if (relay_host != NULL)
-    my_free (relay_host);
+    SH_FREE (relay_host);
+
+  if (0 == sl_strncmp(str_s, _("NULL"), 4))
+    {
+      relay_host = NULL;
+      SL_RETURN( 0, _("sh_mail_set_relay"));
+    }
 
   i = sl_strlen(str_s) + 1;
-  relay_host = sl_malloc (i);
+  relay_host = SH_ALLOC (i);
   if (relay_host != NULL)
     sl_strlcpy(relay_host, str_s, i);
   else
@@ -851,6 +1036,9 @@ static FILE * sh_mail_start_conn (int aFlag)
 
   int                  fd;
 
+  dnsrep     * answers;
+  mx         * result;
+
   SL_ENTER(_("sh_mail_start_conn"));
 
   time_wait = 300;
@@ -897,14 +1085,44 @@ static FILE * sh_mail_start_conn (int aFlag)
 
 
   if (relay_host != NULL) 
-    sl_strlcpy (m_machine, relay_host, sizeof(m_machine));
- 
-  TPT(( 0, FIL__, __LINE__, _("msg=<user %s machine %s>\n"), 
-	m_user, m_machine)); 
+    {
+      sl_strlcpy (m_machine, relay_host, sizeof(m_machine));
+      TPT((0, FIL__, __LINE__, _("msg=<user %s machine %s>\n"), 
+	   m_user, m_machine)); 
+      fd = connect_port (m_machine, IPPORT_SMTP, 
+			 error_call, &error_num, error_msg, 256);
+    }
+  else
+    {
+      answers = return_mx (m_machine);
+      if (answers)
+	{
+	  result = answers->reply;
+	  fd     = -1;
+ 	  for (i = 0; i < answers->count; ++i)
+	    {
+	      sl_strlcpy(m_machine, result[i].address, sizeof(m_machine));
+	      TPT((0, FIL__, __LINE__, 
+		   _("msg=<user %s mx %s pref %d>\n"), 
+		   m_user, m_machine, result[i].pref));
+	      fd = connect_port (m_machine, IPPORT_SMTP, 
+				 error_call, &error_num, error_msg, 256);
+	      if (fd >= 0)
+		break;
+	    }
+	  free_mx(answers);
+	}
+      else
+	{
+	  sl_strlcpy(error_call, _("return_mx"), SH_MINIBUF);
+	  sl_strlcpy(error_msg, _("The specified host is unknown: "), 256);
+	  sl_strlcat(error_msg, m_machine, 256); 
 
-  fd = connect_port (m_machine, IPPORT_SMTP, 
-		     error_call, &error_num, error_msg, 256);
+	  fd = -1;
+	}
+    }
 
+  
   if (fd < 0)
     {
       sh_error_handle ((-1), FIL__, __LINE__, error_num, 
@@ -915,7 +1133,7 @@ static FILE * sh_mail_start_conn (int aFlag)
 
   /* associate a FILE structure with it
    */
-  MBLK( connFile = fdopen (fd, "r+"); )
+  connFile = fdopen (fd, "r+");
   if (connFile == NULL) 
     {
       TPT(( 0, FIL__, __LINE__, _("msg=<fdopen() failed>\n")));
@@ -933,7 +1151,7 @@ static FILE * sh_mail_start_conn (int aFlag)
 		      _("sh_mail_start_conn"), 
 		      _("mail"), sh.host.name);
       TPT(( 0, FIL__, __LINE__, _("msg=<Timeout>\n")));
-      MBLK( fclose(connFile); )
+      fclose(connFile);
       SL_RETURN( NULL, _("sh_mail_start_conn"));
     }
 
@@ -963,7 +1181,7 @@ static FILE * sh_mail_start_conn (int aFlag)
 		      _("mail"), sh.host.name);
 
       TPT(( 0, FIL__, __LINE__, _("msg=<Timeout.>\n")));
-      MBLK( fclose(connFile); )
+      fclose(connFile);
       SL_RETURN( NULL, _("sh_mail_start_conn"));
     }
 
@@ -989,7 +1207,7 @@ static FILE * sh_mail_start_conn (int aFlag)
 		      _("MAIL FROM failed"), _("sh_mail_start_conn"), 
 		      _("mail"), this_address);
       TPT(( 0, FIL__, __LINE__, _("msg=<Timeout.>\n")));
-      MBLK( fclose(connFile); )
+      fclose(connFile);
       SL_RETURN( NULL, _("sh_mail_start_conn"));
     }
 
@@ -1010,7 +1228,7 @@ static FILE * sh_mail_start_conn (int aFlag)
 			  _("RCPT TO failed"), _("sh_mail_start_conn"), 
 			  _("mail"), address);
 	  TPT(( 0, FIL__, __LINE__, _("msg=<Timeout.>\n")));
-	  MBLK( fclose(connFile); )
+	  fclose(connFile);
 	  SL_RETURN( NULL, _("sh_mail_start_conn"));
 	}
     }
@@ -1040,7 +1258,7 @@ static FILE * sh_mail_start_conn (int aFlag)
 	}
       if (ecount == address_num)
 	{
-	  MBLK( fclose(connFile); )
+	  fclose(connFile);
 	  SL_RETURN( NULL, _("sh_mail_start_conn"));
 	}
     }
@@ -1059,7 +1277,7 @@ static FILE * sh_mail_start_conn (int aFlag)
 		      _("DATA failed"), _("sh_mail_start_conn"), 
 		      _("mail"), address);
       TPT(( 0, FIL__, __LINE__, _("msg=<Timeout.>\n")));
-      MBLK( fclose(connFile); )
+      fclose(connFile);
       SL_RETURN( NULL, _("sh_mail_start_conn"));
     }
 
@@ -1105,6 +1323,7 @@ static int sh_mail_end_conn (FILE * connFile)
     {  
       fflush(connFile);
       fprintf(connFile, _("QUIT%c%c"), 13, 10);
+      fflush(connFile);
       TPT(( 0, FIL__, __LINE__, _("msg=<exit>\n")));
 
       SL_RETURN (0, _("sh_mail_end_conn"));
@@ -1132,11 +1351,9 @@ static jmp_buf wait_timeout;
 
 static void sh_mail_alarmhandle (int mysignal)
 {
-  if (mysignal == SIGALRM)   /* use 'signal' to avoid compiler warning */
-    {  
-      alarm(0);
-      longjmp(wait_timeout, 1);
-    }
+  (void) mysignal; /* avoid compiler warning */
+
+  longjmp(wait_timeout, 1);
 }
 
 static int sh_mail_wait(int code, FILE * m_socket)
@@ -1154,15 +1371,15 @@ static int sh_mail_wait(int code, FILE * m_socket)
 
   time_t waited_time;
 
-  struct  sigaction  def_act;
+  struct   sigaction          old_act;
+  volatile unsigned int       old_alarm = 0;
+
   struct  sigaction  new_act;
   sigset_t           unblock;
 
   sigemptyset(&unblock);
   sigaddset  (&unblock, SIGALRM);
 
-  def_act.sa_handler = SIG_IGN;
-
   new_act.sa_handler = sh_mail_alarmhandle;
   sigemptyset( &new_act.sa_mask );       /* set an empty mask       */
   new_act.sa_flags = 0;                  /* init sa_flags           */
@@ -1174,7 +1391,8 @@ static int sh_mail_wait(int code, FILE * m_socket)
   if (setjmp(wait_timeout) != 0)
     {
       alarm(0);
-      sigaction (SIGALRM, &def_act, NULL);
+      sigaction (SIGALRM, &old_act, NULL);
+      alarm(old_alarm);
       sigprocmask(SIG_UNBLOCK, &unblock, NULL);
       TPT((0, FIL__, __LINE__, _("msg=<mail_wait: timeout>\n"))); 
       SL_RETURN( 0, _("mail_wait"));
@@ -1184,8 +1402,8 @@ static int sh_mail_wait(int code, FILE * m_socket)
 
   /* timeout after 5 minutes
    */
-  alarm(0);
-  sigaction (SIGALRM, &new_act, &def_act);
+  old_alarm = alarm(0);
+  sigaction (SIGALRM, &new_act, &old_act);
   alarm(time_wait);
 
   rcode = 0;
@@ -1196,8 +1414,10 @@ static int sh_mail_wait(int code, FILE * m_socket)
     if ( (g=fgetc(m_socket)) == EOF)
       {
 	alarm(0);
+	sigaction (SIGALRM, &old_act, NULL); 
+	alarm(old_alarm);
+	sigprocmask(SIG_UNBLOCK, &unblock, NULL);
 	TPT((0, FIL__, __LINE__, _("msg=<mail_wait: EOF>\n"))); 
-	sigaction (SIGALRM, &def_act, NULL); 
 	SL_RETURN( 0, _("mail_wait")); 
       }
 
@@ -1229,15 +1449,18 @@ static int sh_mail_wait(int code, FILE * m_socket)
     case WAIT_NL:
       if (g != '\n')
 	break;
-      alarm(0);                            /* Disable alarm       */
-      sigaction (SIGALRM, &def_act, NULL); /* Reset handler       */
+      alarm(0);
+      sigaction (SIGALRM, &old_act, NULL); 
+      alarm(old_alarm);
+      sigprocmask(SIG_UNBLOCK, &unblock, NULL);
+
       TPT((0, FIL__, __LINE__, 
 	   _("msg=<mail_wait: OK got %d (%d) need %d (%d)>\n"),
 	   rcode, (int)(rcode/100), code, (int)(code/100) ));
       g = ((int)(rcode/100) == (int)(code/100)) ? 1 : 0;
       if (g != 1)
 	{
-	  sprintf(errmsg,                         /* known to fit  */
+          sprintf(errmsg,                              /* known to fit  */
 		  _("Bad response (%d), expected %d"), rcode, code);
 	  sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_NET, 
 			  errmsg, _("sh_mail_wait"), 
@@ -1258,8 +1481,11 @@ static int sh_mail_wait(int code, FILE * m_socket)
       break; 
       
     default:
-      alarm(0);                            /* Disable alarm       */
-      sigaction (SIGALRM, &def_act, NULL); /* Reset handler       */
+      alarm(0);
+      sigaction (SIGALRM, &old_act, NULL); 
+      alarm(old_alarm);
+      sigprocmask(SIG_UNBLOCK, &unblock, NULL);
+
       TPT((0, FIL__, __LINE__, _("msg=<mail_wait: bad>\n"))); 
       SL_RETURN( 0, _("mail_wait")); 
       
@@ -1267,7 +1493,9 @@ static int sh_mail_wait(int code, FILE * m_socket)
   }
 
   alarm(0);                            /* Disable alarm       */
-  sigaction (SIGALRM, &def_act, NULL); /* Reset handler       */
+  sigaction (SIGALRM, &old_act, NULL); 
+  alarm(old_alarm);
+  sigprocmask(SIG_UNBLOCK, &unblock, NULL);
 
   TPT((0, FIL__, __LINE__, _("msg=<mail_wait: failed>\n"))); 
 
@@ -1275,6 +1503,422 @@ static int sh_mail_wait(int code, FILE * m_socket)
   SL_RETURN( 0, _("mail_wait")); 
 }
 
+/* -- function to insert "\r\n" after each 998 chars --
+ */
+
+#define SPLIT_AT 998
+
+static char * split_string(char * str)
+{
+  int size;
+  int blocks;
+  int count = 0;
+
+  char * p, * p0;
+  char * q;
+
+  if (!str)
+    return NULL;
+
+  size   = strlen(str);
+  /* fprintf(stderr, "orig = %d\n", size); */
+  blocks = 1 + (size / SPLIT_AT);
+  
+  size   = size + (2*blocks) + 1;
+  p = SH_ALLOC(size);
+  memset(p, '\0', size);
+  /* fprintf(stderr, "alloc = %d\n", size); */
+  p0 = p;
+
+  q = str;
+  while (*q) {
+    *p = *q;
+    ++p;
+    ++q;
+    ++count;
+    if (0 == (count % SPLIT_AT)) {
+      count = 0;
+      *p = '\r';
+      ++p;
+      *p = '\n';
+      ++p;
+    }
+  }
+  /* fprintf(stderr, "used = %d\n", strlen(p0)); */
+
+  return p0;
+}
+
+
+
+/*****************************************************************
+ *
+ * MX Resolver Routines
+ *
+ *****************************************************************/
+
+#include <netinet/in.h>
+#ifdef __APPLE__
+#define BIND_8_COMPAT 1
+#endif
+#include <arpa/nameser.h>
+#include <resolv.h>
+#include <netdb.h>
+#include <sys/socket.h>
+#include <arpa/inet.h>
+
+#include "sh_tools.h"
+
+#ifndef HFIXEDSZ
+#define HFIXEDSZ 12
+#endif
+#ifndef QFIXEDSZ
+#define QFIXEDSZ  4
+#endif
+
+
+static unsigned int get_short (unsigned char * loc)
+{
+  unsigned int retval = 0;
+  if (loc)
+    {
+      /* byte order: MSB first
+       */
+      retval = (((unsigned char) * loc) * 256) | ((unsigned char) * (loc + 1));
+    }
+  return (retval);
+}
+
+static dnsrep * get_mx (char *hostname)
+{
+  int  ret, length, status;
+  mx * result;
+
+  typedef union
+  {
+    HEADER head;
+    unsigned char buffer[4096];
+  } querybuf;
+
+  querybuf reply;
+  char expanded[1024];
+  unsigned char * comp_dn, * eom;
+  HEADER * header;
+  int      count, index, type, rdlength, pref;
+  dnsrep * retval;
+
+  SL_ENTER(_("get_mx"));
+
+  if (0 != res_init ())
+    SL_RETURN (NULL, _("get_mx"));
+
+  errno = 0;
+  length = res_query (hostname, C_IN, T_MX, 
+		      (unsigned char *) &reply, 4095);
+  if (length < 1)
+    {
+      /* error handling
+       */
+      if (length == -1)
+	{
+	  if (errno == ECONNREFUSED)
+	    status = ECONNREFUSED;
+	  else
+	    status = h_errno;
+
+#ifdef FIL__
+	  sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, status, MSG_E_SUBGEN,
+			   (errno == ECONNREFUSED) ? 
+			   sh_error_message (status) : 
+			   sh_tools_errmessage(status),
+			   _("res_query"));
+#else
+	  if (errno == ECONNREFUSED)
+	    fprintf(stderr, " ERROR: %s: \n", strerror(errno));
+	  else
+	    fprintf(stderr, "HERROR: %s\n", hstrerror(h_errno));
+#endif
+	}
+      SL_RETURN (NULL, _("get_mx"));
+    }
+
+  ret = 0;
+  header  = (HEADER *) &reply;
+
+  /* start of data section
+   */
+  comp_dn = (unsigned char *) &reply + HFIXEDSZ;
+
+  /* end-of-message
+   */
+  eom     = (unsigned char *) &reply + length;
+
+  /* HEADER NAME  -- must be skipped or decompressed
+   * TYPE         -- type of data we got back, 16 bit integer
+   * CLASS        -- class we got back, also a 16 bit integer 
+   * TTL          -- 32 bit time-to-live. just skip this 
+   * RDLENGTH     -- length of the data to follow 
+   * RDATA        -- the data:
+   *                 PREF  -- 16 bit preference 
+   *                 MX    -- name of mail exchanger, must be decompressed
+   */
+
+  /* Skip the query data. 
+   * QDCOUNT is the number of entries (unsigned 16 bit int). 
+   */
+  count = ntohs (header->qdcount); 
+  for (index = 0; index < count; ++index)
+    {
+      ret = dn_skipname (comp_dn, eom);
+      comp_dn += ret + QFIXEDSZ;
+      if (ret < 1 || comp_dn >= eom)
+	SL_RETURN (NULL, _("get_mx"));
+    }
+  count         = ntohs (header->ancount);
+  if (count < 1)
+    SL_RETURN (NULL, _("get_mx"));
+
+  retval        = SH_ALLOC (sizeof (dnsrep));
+  if (!retval)
+    SL_RETURN (NULL, _("get_mx"));
+  retval->count = count;
+
+  /* allocate space for the results */
+
+  result        = SH_ALLOC (count * sizeof (mx));
+  if (!result)
+    {
+      SH_FREE   (retval);
+      SL_RETURN (NULL, _("get_mx"));
+    }
+  retval->reply = result;
+
+  do
+    {
+      /* HEADER NAME 
+       */
+      ret = dn_expand ((unsigned char *) &reply, eom, comp_dn, 
+		       (char *) expanded, 1023);
+      comp_dn += ret;
+      if (ret < 1 || comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* TYPE
+       */
+      type = get_short (comp_dn);
+      comp_dn += 2;
+      if (type != T_MX || comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* CLASS (re-use 'type' var)
+       */
+      type = get_short (comp_dn);
+      comp_dn += 2;
+      if (comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* TTL
+       */
+      comp_dn += 4;
+      if (comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* RDLENGTH
+       */
+      rdlength = get_short (comp_dn);
+      comp_dn += 2;
+      if (rdlength < 1 || comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* RDATA
+       */
+      pref = get_short (comp_dn);
+      comp_dn += 2;
+      if (comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      ret = dn_expand ((unsigned char *) &reply, eom, comp_dn, 
+		       (char *) expanded, 1023);
+      comp_dn += ret;
+      if (ret < 1)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+      count--;
+
+      /* fill in the struct 
+       */
+      result[count].pref = pref;
+      result[count].address = SH_ALLOC (strlen (expanded) + 1);
+      strcpy (result[count].address, expanded);        /* known to fit  */
+    }
+  while (ret > 0 && comp_dn < eom && count);
+
+  SL_RETURN (retval, _("get_mx"));
+}
+
+static int comp_mx_pref (const void * a, const void * b)
+{
+  const mx * ax = (const mx *) a;
+  const mx * bx = (const mx *) b;
+  
+  if      (ax->pref > bx->pref)
+    return 1;
+  else if (ax->pref < bx->pref)
+    return -1;
+  else
+    return 0;
+}
+
+/*
+ * return_mx returns a list of valid mail exchangers for domain
+ */
+static dnsrep * return_mx (char *domain)
+{
+  struct hostent *host = NULL;
+  dnsrep * answers = NULL;
+  mx     * result;
+  dnsrep * retval;
+  char     errmsg[128]; 
+
+  SL_ENTER(_("return_mx"));
+
+  if (domain != NULL)
+    answers = get_mx (domain);
+
+  if (answers != NULL && answers->count > 0)
+    {
+      qsort(answers->reply, answers->count, sizeof(mx),
+            comp_mx_pref);
+      SL_RETURN (answers, _("return_mx"));
+    }
+  else
+    {
+#ifdef FIL__
+      if (domain != NULL)
+	{
+	  sl_strlcpy (errmsg, _("No MX record for domain "), 127);
+	  sl_strlcat (errmsg, domain, 127);
+	  sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN,
+			   errmsg,
+			   _("get_mx"));
+#else
+	  strcpy  (errmsg,                               /* known to fit  */
+		   _("No MX record for domain "));
+	  strncat (errmsg, domain, 100);
+	  errmsg[122] = '\0';
+	  fprintf(stderr, "Warning: %s\n", errmsg);
+#endif
+	}
+      if (domain != NULL)
+	host = sh_gethostbyname (domain);
+      if (!host)
+	{
+#ifdef FIL__
+	  sl_strlcpy (errmsg, _("Unknown host "), 127);
+	  sl_strlcat (errmsg, domain, 127);
+	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN,
+			   errmsg,
+			   _("return_mx"));
+#endif
+	  SL_RETURN (NULL, _("return_mx"));
+	}
+      result       = SH_ALLOC (sizeof (mx));
+      retval       = SH_ALLOC (sizeof (dnsrep));
+      retval->reply = result;
+      retval->count = 1;
+      result->pref  = 0;
+      result->address = SH_ALLOC (strlen (host->h_name) + 1);
+      strcpy (result->address, host->h_name);          /* known to fit  */
+      SL_RETURN (retval, _("return_mx"));
+    }
+}
+
+static int free_mx (dnsrep * answers)
+{
+  mx     * result;
+  int      i;
+
+  SL_ENTER(_("free_mx"));
+  if (!answers)
+    SL_RETURN (0, _("return_mx"));
+
+  result = answers->reply;  
+  for (i = 0;  i < answers->count; ++i)
+    {
+      SH_FREE (result[i].address);
+    }
+  SH_FREE(result);
+  SH_FREE(answers);
+  SL_RETURN (0, _("return_mx"));
+}
+
+#ifdef TEST_ONLY
+int main(int argc, char * argv[])
+{
+  int      i;
+  dnsrep * answers;
+  mx     * result;
+
+  if (argc < 2)
+    {
+      fprintf(stderr, "Usage: dns <hostname>\n");
+      return -1;
+    }
+  answers = return_mx(argv[1]);
+
+  if (!answers)
+    {
+      fprintf(stderr, "No answer\n");
+      return -1;
+    }
+
+  if (answers->count > 0)
+    {
+      result = answers->reply;
+      for (i = 0; i < answers->count; ++i)
+	{
+	  fprintf(stderr, "Record %3d: [%3d] %s\n", i, 
+		  result[i].pref, result[i].address);
+	}	  
+    }
+  else
+    {
+      fprintf(stderr, "No answer\n");
+      free_mx(answers);
+      return -1;
+    }
+  free_mx(answers);
+  return (0);
+}
+#endif
+
   
 
 /* if defined(SH_WITH_MAIL) */
