@@ -23,6 +23,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/types.h>
 #include <unistd.h>
 
 
@@ -35,6 +36,8 @@
 #include "sh_utils.h"
 #include "sh_mem.h"
 
+extern int safe_logger (int signal, int method, pid_t thepid);
+
 #undef  FIL__
 #define FIL__  _("sh_mem.c")
 static int eblock = 0;
@@ -48,6 +51,7 @@ static int eblock = 0;
  */
 #define SH_MEMMULT 16
 
+
 typedef struct mem_struct {
   struct mem_struct *next;        /* link to next struct    */
   char * real_address;            /* address assigned       */
@@ -60,13 +64,70 @@ typedef struct mem_struct {
 memlist_t   * memlist       = NULL;
 
 int           Free_Count  = 0, Alloc_Count = 0;
+int           Now_Alloc_Count = 0, Max_Alloc_Count = 0;
 unsigned long Mem_Current = 0, Mem_Max = 0;
 
+#if 0
+#define MEM_DETAILS 
+#endif
+
+#ifdef MEM_DETAILS
+int           max_upto_032 = 0;
+int           max_upto_064 = 0;
+int           max_upto_128 = 0;
+int           max_upto_256 = 0;
+int           max_upto_512 = 0;
+int           max_upto_1024 = 0;
+int           max_upto_4096 = 0;
+int           max_upto_inf = 0;
+
+int           now_upto_032 = 0;
+int           now_upto_064 = 0;
+int           now_upto_128 = 0;
+int           now_upto_256 = 0;
+int           now_upto_512 = 0;
+int           now_upto_1024 = 0;
+int           now_upto_4096 = 0;
+int           now_upto_inf = 0;
+
+int           tot_upto_032 = 0;
+int           tot_upto_064 = 0;
+int           tot_upto_128 = 0;
+int           tot_upto_256 = 0;
+int           tot_upto_512 = 0;
+int           tot_upto_1024 = 0;
+int           tot_upto_4096 = 0;
+int           tot_upto_inf = 0;
+#endif
+
+#ifdef MEM_LOG
+void sh_mem_dump ()
+{
+  memlist_t   * this = memlist;
+
+
+  FILE * fd = fopen(MEM_LOG, "w");
+
+  while (this != NULL)
+    {
+      fprintf (fd, "%20s %5d %ld\n",  this->file, this->line, this->size);
+      this = this->next;
+    }
+  fclose(fd);
+  return;
+}
+#else
+void sh_mem_dump ()
+{
+  return;
+}
+#endif
 
 void sh_mem_stat ()
 {
   memlist_t   * this;
 
+
   SL_ENTER(_("sh_mem_stat"));
 
   if (Alloc_Count == Free_Count) 
@@ -77,10 +138,24 @@ void sh_mem_stat ()
     }
     
   sh_error_handle (SH_ERR_INFO, FIL__, __LINE__, 0, MSG_MSTAMP2,
-		   Alloc_Count, Free_Count);
+		   Alloc_Count, Free_Count, Max_Alloc_Count);
   sh_error_handle (SH_ERR_INFO, FIL__, __LINE__, 0, MSG_MSTAMP,
 		   Mem_Max, Mem_Current);
 
+#ifdef MEM_DETAILS
+  fprintf(stderr, "\n");
+  fprintf(stderr, "__SIZE_____TOTAL___MAXIMUM___\n");
+  fprintf(stderr, "    32    %6d    %6d\n", tot_upto_032, max_upto_032);
+  fprintf(stderr, "    64    %6d    %6d\n", tot_upto_064, max_upto_064);
+  fprintf(stderr, "   128    %6d    %6d\n", tot_upto_128, max_upto_128);
+  fprintf(stderr, "   256    %6d    %6d\n", tot_upto_256, max_upto_256);
+  fprintf(stderr, "   512    %6d    %6d\n", tot_upto_512, max_upto_512);
+  fprintf(stderr, "  1024    %6d    %6d\n", tot_upto_1024, max_upto_1024);
+  fprintf(stderr, "  4096    %6d    %6d\n", tot_upto_4096, max_upto_4096);
+  fprintf(stderr, "   inf    %6d    %6d\n", tot_upto_inf, max_upto_inf);
+  fprintf(stderr, "\n");
+#endif
+
   this = memlist;
 
   while (this != NULL) 
@@ -128,15 +203,18 @@ void * sh_mem_malloc (size_t size, char * file, int line)
 
   SL_ENTER(_("sh_mem_malloc"));
 
-  the_realAddress = sl_malloc(size + 2 * SH_MEMMULT);
+  the_realAddress = malloc(size + 2 * SH_MEMMULT);
   
   if ( the_realAddress  == NULL ) 
     {
       if (eblock == 0)
 	{
 	  eblock = 1;
+	  safe_logger (0, 0, getpid());
+	  /*
 	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_MMEM,
 			   file, line);
+	  */
 	  eblock = 0;
 	}
       /* use _exit() rather than exit() - we malloc() in atexit() functions 
@@ -153,19 +231,82 @@ void * sh_mem_malloc (size_t size, char * file, int line)
   memset(theAddress,      0,         1);
 
   ++Alloc_Count;
+  ++Now_Alloc_Count;
+
+  if (Max_Alloc_Count < Now_Alloc_Count)
+    Max_Alloc_Count = Now_Alloc_Count;
+
+#ifdef MEM_DETAILS
+  if (size <= 32)
+    { 
+      ++now_upto_032;
+      ++tot_upto_032;
+      if (now_upto_032 > max_upto_032) max_upto_032 = now_upto_032;
+    }
+  else if  (size <= 64)
+    { 
+      ++now_upto_064;
+      ++tot_upto_064;
+      if (now_upto_064 > max_upto_064) max_upto_064 = now_upto_064;
+    }
+  else if  (size <= 128)
+    { 
+      ++now_upto_128;
+      ++tot_upto_128;
+      if (now_upto_128 > max_upto_128) max_upto_128 = now_upto_128;
+    }
+  else if  (size <= 256)
+    { 
+      ++now_upto_256;
+      ++tot_upto_256;
+      if (now_upto_256 > max_upto_256) max_upto_256 = now_upto_256;
+    }
+  else if  (size <= 512)
+    { 
+      ++now_upto_512;
+      ++tot_upto_512;
+      if (now_upto_512 > max_upto_512) max_upto_512 = now_upto_512;
+    }
+  else if  (size <= 1024)
+    { 
+      ++now_upto_1024;
+      ++tot_upto_1024;
+      if (now_upto_1024 > max_upto_1024) max_upto_1024 = now_upto_1024;
+    }
+  else if  (size <= 4096)
+    { 
+      ++now_upto_4096;
+      ++tot_upto_4096;
+      if (now_upto_4096 > max_upto_4096) max_upto_4096 = now_upto_4096;
+    }
+  else 
+    { 
+      ++now_upto_inf;
+      ++tot_upto_inf;
+      if (now_upto_inf > max_upto_inf) max_upto_inf = now_upto_inf;
+
+      fprintf(stderr, "\n___BIGSIZE___");
+      fprintf(stderr, "   %6d  -> %16s  %10d \n", size, file, line);
+      fprintf(stderr, "\n");
+
+    }
+#endif
 
   Mem_Current += size;
   Mem_Max = ( (Mem_Current > Mem_Max) ? Mem_Current : Mem_Max);
 
-  this = (memlist_t *) sl_malloc (sizeof(memlist_t));
+  this = (memlist_t *) malloc (sizeof(memlist_t));
 
   if ( this == NULL) 
     {
       if (eblock == 0)
 	{
 	  eblock = 1;
-	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_MNULL,
+	  safe_logger(0, 0, getpid());
+	  /*
+	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_MMEM,
 			   file, line);
+	  */
 	  eblock = 0;
 	}
       SL_RETURN( the_realAddress, _("sh_mem_malloc"));
@@ -182,7 +323,7 @@ void * sh_mem_malloc (size_t size, char * file, int line)
   this->next = memlist;
   memlist = this;
 
-  SL_RETURN( theAddress, _(""));
+  SL_RETURN( theAddress, _("sh_mem_malloc"));
 }
 
 
@@ -238,10 +379,31 @@ void sh_mem_free (void * a, char * file, int line)
 	before->next = this->next;
     }
 
-  my_free(a);
+  free(a);
   if (this)
-    my_free(this);
+    free(this);
   ++Free_Count;
+  --Now_Alloc_Count;
+
+#ifdef MEM_DETAILS
+  if (size <= 32)
+    --now_upto_032;
+  else if  (size <= 64)
+    --now_upto_064;
+  else if  (size <= 128)
+    --now_upto_128;
+  else if  (size <= 256)
+    --now_upto_256;
+  else if  (size <= 512)
+    --now_upto_512;
+  else if  (size <= 1024)
+    --now_upto_1024;
+  else if  (size <= 4096)
+    --now_upto_4096;
+  else 
+    --now_upto_inf;
+#endif
+
   Mem_Current -= size;
   SL_RET0(_("sh_mem_free"));
 }
@@ -252,12 +414,13 @@ void sh_mem_free (void * a)
 {
   SL_ENTER(_("sh_mem_free"));
 
-  if ( a == NULL ) 
-    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_MNULL);
+  if (a)
+    {
+      free(a);
+    }
   else
     {
-      my_free(a);
-      a = NULL;
+      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_MNULL);
     }
   SL_RET0(_("sh_mem_free"));
 }
@@ -268,14 +431,17 @@ void * sh_mem_malloc (size_t size)
 
   SL_ENTER(_("sh_mem_malloc"));
 
-  theAddress = sl_malloc(size);
+  theAddress = malloc(size);
 
   if ( theAddress == NULL ) 
     {
       if (eblock == 0)
 	{
 	  eblock = 1;
+	  safe_logger(0, 0, getpid());
+	  /*
 	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_MMEM);
+	  */
 	  eblock = 0;
 	}
       /* use _exit() rather than exit() - we malloc() in atexit()  
