@@ -38,6 +38,8 @@
 #include "sh_forward.h"
 #include "sh_modules.h"
 #include "sh_gpg.h"
+#include "sh_hash.h"
+#include "sh_ignore.h"
 #ifdef WITH_EXTERNAL
 #include "sh_extern.h"
 #endif
@@ -121,7 +123,7 @@ struct str_ListSections tab_ListSections[] = {
 };
 
    
-static void sh_readconfig_line (char * line);
+static int sh_readconfig_line (char * line);
 
 static ShSectionType read_mode = SH_SECTION_NONE;
 
@@ -223,7 +225,7 @@ int sh_readconf_read (void)
 	{
 	  sl_get_euid(&euid);
 	  dlog(1, FIL__, __LINE__, 
-	       _("Could not open the local file signature database for reading because\nof the following error: %s\nIf this is a permission problem, you need to change file permissions\nto make the file readable for the effective UID: %d\n"), 
+	       _("Could not open the local configuration file for reading because\nof the following error: %s\nIf this is a permission problem, you need to change file permissions\nto make the file readable for the effective UID: %d\n"), 
 	       sl_get_errmsg(), (int) euid);
 	  sh_error_handle ((-1), FIL__, __LINE__, EACCES, MSG_NOACCESS, 
 			   (long) euid, 
@@ -257,7 +259,7 @@ int sh_readconf_read (void)
 	sl_write_line(fdTmp, line_in, sl_strlen(line_in));
       }
 #if defined(WITH_GPG) || defined(WITH_PGP)
-    if (0 == sl_strncmp(line, _("-----END PGP SIGNATURE-----"), 25))
+    if (0 == sl_strncmp(line_in, _("-----END PGP SIGNATURE-----"), 25))
       break;
 #else
     if (0 == sl_strncmp(line_in, _("[EOF]"), 5))
@@ -279,6 +281,8 @@ int sh_readconf_read (void)
 
   /* ---  Start reading lines.  ---
    */
+  conf_line = 0;
+
   while ( sh_unix_getline (fd, line_in, 512) > 0) {
 
     ++conf_line;
@@ -334,7 +338,8 @@ int sh_readconf_read (void)
 
     /* Skip header etc. 
      */
-    if (line[0] == '#' || line[0] == '\0' || line[0] == ';' || line[0] == '/')
+    if (line[0] == '#' || line[0] == '\0' || line[0] == ';' || 
+	(line[0] == '/' && line[1] == '/'))
       continue; 
   
     /* Clip off trailing white space.                 
@@ -382,8 +387,10 @@ int sh_readconf_read (void)
 	  }
 #endif
 	if (read_mode == SH_SECTION_NONE)
-	  sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 0, MSG_EINVALL,
-			   _("config file"), (long) conf_line);
+	  {
+	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALHEAD,
+			     (long) conf_line);
+	  }
       } 
 
     /* ---  an @host directive -------------- */
@@ -405,7 +412,8 @@ int sh_readconf_read (void)
 	  {
 	    if (0 == seen_host)
 	      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALD,
-			       _("config file"), (long) conf_line);
+			       _("config file"), 
+			       (long) conf_line);
 
 	    else if (host_int == 0)
 	      {
@@ -484,7 +492,8 @@ int sh_readconf_read (void)
 	  {
 	    if (0 == seen_system)
 	      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALD,
-			       _("config file"), (long) conf_line);
+			       _("config file"),
+			       (long) conf_line);
 
 	    else if (sys_int == 0)
 	      {
@@ -530,7 +539,11 @@ int sh_readconf_read (void)
 	     some_other_system[seen_system] == 0      && 
 	     read_mode != SH_SECTION_NONE) 
       { 
-	sh_readconfig_line (line);
+	if (0 != sh_readconfig_line (line))
+	  {
+	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALCONF,
+			     (long) conf_line);
+	  }
       }
 
   }
@@ -539,7 +552,8 @@ int sh_readconf_read (void)
 	   
   if (0 != seen_host || 0 != seen_system)
     sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALDD,
-		     _("config file"), (long) conf_line);
+		     _("config file"), 
+		     (long) conf_line);
 
 #if defined(WITH_GPG) || defined(WITH_PGP)
   /* Validate signature of open file.
@@ -690,6 +704,11 @@ typedef struct _cfg_options {
 extern int sh_set_schedule_one(char * str);
 extern int sh_set_schedule_two(char * str);
 #endif
+#if defined (SH_WITH_SERVER)
+extern int sh_socket_use (char * c);
+extern int sh_socket_uid (char * c);
+extern int sh_socket_password (char * c);
+#endif
 
 cfg_options ext_table[] = {
 #if defined(WITH_EXTERNAL)
@@ -715,13 +734,15 @@ cfg_options ext_table[] = {
     sh_ext_add_and },
   { N_("setfilteror"),     SH_SECTION_EXTERNAL, SH_SECTION_NONE,  
     sh_ext_add_or },
-  { N_("externalseverity"),SH_SECTION_LOG,      SH_SECTION_NONE,  
+  { N_("externalseverity"),SH_SECTION_LOG,      SH_SECTION_EXTERNAL,  
     sh_error_set_external },
-  { N_("externalclass"),   SH_SECTION_LOG,      SH_SECTION_NONE,  
+  { N_("externalclass"),   SH_SECTION_LOG,      SH_SECTION_EXTERNAL,  
     sh_error_external_mask },
 #endif
 
 #if defined(WITH_DATABASE)
+  { N_("usepersistent"),   SH_SECTION_DATABASE, SH_SECTION_NONE,  
+    sh_database_use_persistent },
   { N_("setdbname"),       SH_SECTION_DATABASE, SH_SECTION_NONE,  
     sh_database_set_database },
   { N_("setdbtable"),      SH_SECTION_DATABASE, SH_SECTION_NONE,  
@@ -734,9 +755,9 @@ cfg_options ext_table[] = {
     sh_database_set_password },
   { N_("addtodbhash"),     SH_SECTION_DATABASE, SH_SECTION_NONE,  
     sh_database_add_to_hash },
-  { N_("databaseseverity"),SH_SECTION_LOG,      SH_SECTION_NONE,  
+  { N_("databaseseverity"),SH_SECTION_LOG,      SH_SECTION_DATABASE,  
     sh_error_set_database },
-  { N_("databaseclass"),   SH_SECTION_LOG,      SH_SECTION_NONE,  
+  { N_("databaseclass"),   SH_SECTION_LOG,      SH_SECTION_DATABASE,  
     sh_error_database_mask },
   { N_("setdbservertstamp"), SH_SECTION_DATABASE,      SH_SECTION_NONE,  
     set_enter_wrapper },
@@ -777,11 +798,22 @@ cfg_options ext_table[] = {
   { N_("file"),           SH_SECTION_USER1,      SH_SECTION_NONE, 
     sh_files_pushfile_user1 },
 
+  { N_("ignoreadded"), SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_ignore_add_new },
+  { N_("ignoremissing"), SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_ignore_add_del },
+
   { N_("filecheckscheduleone"), SH_SECTION_MISC,   SH_SECTION_NONE, 
     sh_set_schedule_one },
   { N_("filecheckscheduletwo"), SH_SECTION_MISC,   SH_SECTION_NONE, 
     sh_set_schedule_two },
 
+  { N_("usehardlinkcheck"),   SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_files_check_hardlinks },
+  { N_("hardlinkoffset"),     SH_SECTION_MISC,   SH_SECTION_NONE,
+    sh_files_hle_reg },
+  { N_("addokchars"),         SH_SECTION_MISC,   SH_SECTION_NONE,
+    sh_util_obscure_ok },
   { N_("setrecursionlevel"),  SH_SECTION_MISC,   SH_SECTION_NONE, 
     sh_files_setrecursion },
   { N_("checksumtest"),       SH_SECTION_MISC,   SH_SECTION_NONE, 
@@ -801,6 +833,9 @@ cfg_options ext_table[] = {
     sh_unix_set_io_limit },
 #endif
 
+  { N_("versionstring"),        SH_SECTION_MISC,   SH_SECTION_NONE,
+    sh_hash_version_string },
+
   { N_("digestalgo"),           SH_SECTION_MISC,   SH_SECTION_NONE, 
     sh_tiger_hashtype },
 
@@ -835,6 +870,18 @@ cfg_options ext_table[] = {
   { N_("setudpactive"),        SH_SECTION_SRV,  SH_SECTION_MISC, 
     set_syslog_active },
 #endif
+  { N_("setusesocket"),        SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_socket_use },
+  { N_("setsocketallowuid"),   SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_socket_uid },
+  { N_("setsocketpassword"),   SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_socket_password },
+  { N_("setstripdomain"),      SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_forward_set_strip },
+  { N_("useseparatelogs"),     SH_SECTION_SRV,  SH_SECTION_MISC, 
+    set_flag_sep_log },
+  { N_("setchrootdir"),        SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_unix_set_chroot },
   { N_("setclienttimelimit"),  SH_SECTION_SRV,  SH_SECTION_MISC, 
     sh_forward_set_time_limit },
   { N_("useclientseverity"),   SH_SECTION_SRV,  SH_SECTION_MISC, 
@@ -880,7 +927,8 @@ cfg_options ext_table[] = {
   { N_("mailsubject"),       SH_SECTION_MAIL,  SH_SECTION_MISC, 
     set_mail_subject },
 #endif
-
+  { N_("setbindaddress"),    SH_SECTION_MISC,  SH_SECTION_NONE,
+    sh_calls_set_bind_addr },
   { N_("daemon"),            SH_SECTION_MISC,  SH_SECTION_NONE, 
     sh_unix_setdeamon },
   { N_("samhainpath"),       SH_SECTION_MISC,  SH_SECTION_NONE, 
@@ -904,6 +952,12 @@ cfg_options ext_table[] = {
     sh_error_set_syslog },
   { N_("syslogclass"),       SH_SECTION_LOG,   SH_SECTION_NONE, 
     sh_error_syslog_mask },
+#ifdef HAVE_LIBPRELUDE
+  { N_("preludeseverity"),   SH_SECTION_LOG,   SH_SECTION_NONE, 
+    sh_error_set_prelude },
+  { N_("preludeclass"),      SH_SECTION_LOG,   SH_SECTION_NONE, 
+    sh_error_prelude_mask },
+#endif
 
   { N_("logcalls"),          SH_SECTION_LOG,   SH_SECTION_NONE, 
     sh_aud_set_functions },
@@ -915,9 +969,6 @@ cfg_options ext_table[] = {
     sh_log_set_console },
 
 #ifdef WITH_MESSAGE_QUEUE
-  { N_("setconsole"),        SH_SECTION_MISC,  SH_SECTION_NONE, 
-    sh_log_set_console },
-
   { N_("messagequeueactive"),SH_SECTION_MISC,  SH_SECTION_NONE, 
     enable_msgq },
 #endif
@@ -949,11 +1000,13 @@ cfg_options ext_table[] = {
 
 
 
-static void sh_readconfig_line (char * line)
+static int sh_readconfig_line (char * line)
 {    
   char * c;
   char * tmp;
   int    i;
+  int    good_opt = -1;
+
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
   int    modnum, modkey;
 #endif
@@ -999,11 +1052,13 @@ static void sh_readconfig_line (char * line)
 	  TPT(( 0, FIL__, __LINE__, _("msg=<ConfigFile: not key=value: %s>\n"),
 		line));
 	}
-      SL_RET0(_("sh_readconf_line"));
+      SL_RETURN(good_opt, _("sh_readconf_line"));
     }
   else
     ++c;
 
+  /* skip leading whitespace
+   */
   while ((*c) == ' ' || (*c) == '\t')
     ++c;
 
@@ -1014,7 +1069,7 @@ static void sh_readconfig_line (char * line)
 	  TPT(( 0, FIL__, __LINE__, _("msg=<ConfigFile: not key=value: %s>\n"),
 		line));
 	}
-      SL_RET0(_("sh_readconf_line"));
+      SL_RETURN(good_opt, _("sh_readconf_line"));
     }
 
   /* convert to lowercase                              */
@@ -1040,10 +1095,11 @@ static void sh_readconfig_line (char * line)
 	  for (modkey = 0; modList[modnum].conf_table[modkey].the_opt != NULL; 
 	       ++modkey) 
 	    {
-	      if (sl_strncmp (line,  
+	      if (sl_strncmp (line,   
 			      _(modList[modnum].conf_table[modkey].the_opt),
-			      sl_strlen(modList[modnum].conf_table[modkey].the_opt)) == 0)
+			      sl_strlen(modList[modnum].conf_table[modkey].the_opt) ) == 0)
 		{
+		  good_opt = 0;
 		  if (0 != modList[modnum].conf_table[modkey].func(c))
 		    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALS,
 				     _(modList[modnum].conf_table[modkey].the_opt), c);
@@ -1053,10 +1109,12 @@ static void sh_readconfig_line (char * line)
 			    _("msg=<line = %s, option = %s>\n"), line,
 			    _(modList[modnum].conf_table[modkey].the_opt)));
 		    }
+		  goto outburst;
 		}
 	    }
 	}
     }
+  outburst:
 #endif
 
 
@@ -1065,7 +1123,11 @@ static void sh_readconfig_line (char * line)
       i = 0;
       while (ident[i] != NULL) {
 	if (sl_strncmp (line, _(ident[i]), sl_strlen(ident[i])-1) == 0)
-	  sh_error_set_iv (identnum[i], c);
+	  {
+	    good_opt = 0;
+	    sh_error_set_iv (identnum[i], c);
+	    break;
+	  }
 	++i;
       }
     }
@@ -1079,6 +1141,7 @@ static void sh_readconfig_line (char * line)
 	      sl_strncmp (line, _(ext_table[i].optname), 
 			  sl_strlen(ext_table[i].optname)) == 0)
 	    {
+	      good_opt = 0;
 	      if (0 != ext_table[i].func (c))
 		sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALS,
 				 _(ext_table[i].optname), c);
@@ -1087,8 +1150,8 @@ static void sh_readconfig_line (char * line)
 	  ++i;
 	}
     }
-     
-  SL_RET0(_("sh_readconf_line"));
+
+  SL_RETURN(good_opt, _("sh_readconf_line"));
 }
   
     
