@@ -17,6 +17,8 @@
 /*  along with this program; if not, write to the Free Software            */
 /*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */
 
+#include "config_xor.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -31,7 +33,6 @@
  */
 #ifndef TESTONLY
 
-#include "config_xor.h"
 
 #undef  FIL__
 #define FIL__  _("sh_schedule.c")
@@ -120,7 +121,7 @@ int test_val (int i, int min, int max, int min_step,
   return 1;
 }
 
-int test_sched (sh_schedule_t * isched)
+int test_sched_int (sh_schedule_t * isched)
 {
   time_t now;
   struct tm * tval;
@@ -155,8 +156,25 @@ int test_sched (sh_schedule_t * isched)
   return 0;
 }
 
-char DayNames[7][3] = { "sun", "mon", "tue", "wed", "thu", "fri", "sat" };
-char MonNames[12][3] = { "jan", "feb", "mar", "apr", "may", "jun", 
+/* test a linked list of schedules
+ */
+int test_sched (sh_schedule_t * isched)
+{
+  sh_schedule_t * intern = isched;
+  int             retval = 0;
+
+  while (intern != NULL)
+    {
+      if (test_sched_int(intern) == 1)
+	retval = 1;
+      intern = intern->next;
+    }
+  return retval;
+}
+
+
+char DayNames[7][4] = { "sun", "mon", "tue", "wed", "thu", "fri", "sat" };
+char MonNames[12][4] = { "jan", "feb", "mar", "apr", "may", "jun", 
 		       "jul", "aug", "sep", "oct", "nov", "dec" };
 
 
@@ -265,8 +283,10 @@ int parse_token(int i, sh_schedule_t * isched, char * p)
     case 1:
       if (isched->max[i] == isched->min[i])
 	{
+	  /* fix for daylight saving time: subtract 3600 sec 
+	   */
 	  if (isched->min_step == 3599)
-	    isched->min_step = 86399;
+	    isched->min_step = 86399 - 3600;
 	}
       else
 	{
@@ -291,7 +311,7 @@ int parse_sched (const char * ssched, sh_schedule_t * isched)
     return -1;
 
 #ifdef TESTONLY
-  copy = malloc(strlen(ssched)+1);
+  copy = malloc(strlen(ssched)+1);                 /* testonly code */
 #else
   copy = SH_ALLOC(strlen(ssched)+1);
 #endif
@@ -320,6 +340,7 @@ int parse_sched (const char * ssched, sh_schedule_t * isched)
 
   isched->last_exec = (time_t)-1;
   isched->first     = 0;
+  isched->next      = NULL;
 
 #ifdef TESTONLY
   free(copy);
@@ -391,7 +412,7 @@ int main(int argc, char * argv[])
     {
       if (test_sched(&isched))
 	printf("EXECUTE  at: %s", ctime(&(isched.last_exec)));
-      sleep (1);
+      sleep (1); /* TESTONLY */
     }
   return 0;
 }
