@@ -29,10 +29,19 @@
 #include <unistd.h>
 #include <errno.h>
 #include <limits.h>
-
+#ifdef HAVE_LIBGEN_H
+#include <libgen.h>
+#endif
+#ifdef HAVE_SCHED_H
+#include <sched.h>
+#endif
 
 #ifdef SH_USE_SUIDCHK
 
+#ifndef HAVE_BASENAME
+#define basename(a) ((NULL == strrchr(a, '/')) ? (a) : (strrchr(a, '/')))
+#endif
+
 #undef  FIL__
 #define FIL__  _("sh_suidchk.c")
 
@@ -102,6 +111,22 @@ sh_rconf sh_suidchk_table[] = {
     N_("suidcheckfps"),
     sh_suidchk_set_fps
   },
+  {
+    N_("suidcheckyield"),
+    sh_suidchk_set_yield
+  },
+  {
+    N_("suidcheckquarantinefiles"),
+    sh_suidchk_set_quarantine
+  },
+  {
+    N_("suidcheckquarantinemethod"),
+    sh_suidchk_set_qmethod
+  },
+  {
+    N_("suidcheckquarantinedelete"),
+    sh_suidchk_set_qdelete
+  },
   {
     NULL,
     NULL
@@ -113,6 +138,10 @@ static time_t  lastcheck         = (time_t) 0;
 static int     ShSuidchkActive   = S_TRUE;
 static long    ShSuidchkInterval = 7200;
 static long    ShSuidchkFps      = 0;
+static int     ShSuidchkYield    = S_FALSE;
+static int     ShSuidchkQEnable  = S_FALSE;
+static int     ShSuidchkQMethod  = SH_Q_CHANGEPERM;
+static int     ShSuidchkQDelete  = S_FALSE;
 static int     ShSuidchkSeverity = SH_ERR_SEVERE;
 static char *  ShSuidchkExclude  = NULL;
 static int     ExcludeLen        = 0;
@@ -128,22 +157,232 @@ static char *
 filesystem_type (char * path, char * relpath, struct stat * statp);
 
 #ifndef PATH_MAX
-#define PATH_MAX 1023
+#define PATH_MAX 1024
 #endif
 
 extern unsigned long sh_files_maskof (int class);
 
+/* Recursively descend into the directory to make sure that
+ * there is no symlink in the path.
+ */
+static int do_truncate_int (char * path, int depth)
+{
+  char      * q;
+  struct stat one; 
+  struct stat two;
+  int         fd;
+
+  if (depth > 99)
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, EINVAL,
+		       MSG_SUID_ERROR,
+		       _("do_truncate: max depth 99 exceeded"));
+      return -1;
+    }
+  ++depth;
+  if (path[0] != '/')
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, EINVAL,
+		       MSG_SUID_ERROR,
+		       _("do_truncate: not an absolute path"));
+      return -1;
+    }
+  ++path;
+  q = strchr(path, '/');
+  if (q)
+    {
+      *q = '\0';
+      if (0 != lstat(path, &one))
+	{ 
+	  sh_error_handle ((-1), FIL__, __LINE__, errno,
+			   MSG_SUID_ERROR,
+			   sh_error_message(errno));
+	  *q = '/'; 
+	  return -1; 
+	}
+      if (!S_ISDIR(one.st_mode))
+	{ 
+	  sh_error_handle ((-1), FIL__, __LINE__, EINVAL,
+			   MSG_SUID_ERROR,
+			   _("Possible race: not a directory"));
+	  *q = '/'; 
+	  return -1; 
+	}
+
+
+      if (0 != chdir(path))
+	{
+	  sh_error_handle ((-1), FIL__, __LINE__, errno,
+			   MSG_SUID_ERROR,
+			   sh_error_message(errno));
+	  *q = '/';
+	  return -1;
+	}
+      *q = '/';
+      if (0 != lstat(".", &two))
+	{ 
+	  sh_error_handle ((-1), FIL__, __LINE__, errno,
+			   MSG_SUID_ERROR,
+			   sh_error_message(errno));
+	  return -1; 
+	}
+      if ((one.st_dev != two.st_dev) || (one.st_ino != two.st_ino) || 
+	  (!S_ISDIR(two.st_mode)))
+	{ 
+	  sh_error_handle ((-1), FIL__, __LINE__, EINVAL,
+			   MSG_SUID_ERROR,
+			   _("Possible race: lstat(dir) != lstat(.)"));
+	  return -1;
+	}
+
+
+      return (do_truncate_int(q, depth));
+    }
+  else
+    {
+      /* no more '/', so this is the file 
+       */
+      if (*path == '\0')
+	return -1;
+      if (0 != lstat(path, &one))
+	{
+	  sh_error_handle ((-1), FIL__, __LINE__, errno,
+			   MSG_SUID_ERROR,
+			   sh_error_message(errno));
+	  return -1;
+	} 
+      fd = open(path, O_RDWR);
+      if (-1 == fd)
+	{
+	  sh_error_handle ((-1), FIL__, __LINE__, errno,
+			   MSG_SUID_ERROR,
+			   sh_error_message(errno));
+	  return -1;
+	} 
+      if (0 != fstat(fd, &two))
+	{ 
+	  sh_error_handle ((-1), FIL__, __LINE__, errno,
+			   MSG_SUID_ERROR,
+			   sh_error_message(errno));
+	  close(fd);
+	  return -1; 
+	}
+      if ((one.st_dev != two.st_dev) || (one.st_ino != two.st_ino))
+	{ 
+	  sh_error_handle ((-1), FIL__, __LINE__, EINVAL,
+			   MSG_SUID_ERROR,
+			   _("Possible race: lstat != fstat"));
+	  close(fd); 
+	  return -1;
+	}
+      if (!S_ISREG(two.st_mode))
+	{ 
+	  sh_error_handle ((-1), FIL__, __LINE__, EINVAL,
+			   MSG_SUID_ERROR,
+			   _("Possible race: not a regular file"));
+	  close(fd); 
+	  return -1;
+	}
+      if ((0 == (two.st_mode & S_ISUID)) && (0 == (two.st_mode & S_ISGID)))
+	{ 
+	  sh_error_handle ((-1), FIL__, __LINE__, EINVAL,
+			   MSG_SUID_ERROR,
+			   _("Possible race: not a suid/sgid file"));
+	  close(fd); 
+	  return -1;
+	}
+      if (ShSuidchkQDelete == S_FALSE)
+	{
+	  if ((two.st_mode & S_ISUID) > 0)
+	    two.st_mode -= S_ISUID;
+	  if ((two.st_mode & S_ISGID) > 0)
+	    two.st_mode -= S_ISGID;
+	  if (-1 == fchmod(fd, two.st_mode))
+	    {
+	      sh_error_handle ((-1), FIL__, __LINE__, errno,
+			       MSG_SUID_ERROR,
+			       sh_error_message(errno));
+	      close(fd); 
+	      return -1;
+	    }
+	  if (two.st_nlink > 1)
+	    {
+	      sh_error_handle ((-1), FIL__, __LINE__, 0,
+			       MSG_SUID_ERROR,
+			       _("Not truncated because hardlink count gt 1"));
+	      close(fd); 
+	      return -1;
+	    }
+	  if (-1 == ftruncate(fd, 0))
+	    {
+	      sh_error_handle ((-1), FIL__, __LINE__, errno,
+			       MSG_SUID_ERROR,
+			       sh_error_message(errno));
+	      close(fd); 
+	      return -1;
+	    }
+	}
+      else
+	{
+	  if (-1 == unlink(path))
+	    {
+	      sh_error_handle ((-1), FIL__, __LINE__, errno,
+			       MSG_SUID_ERROR,
+			       sh_error_message(errno));
+	      close(fd); 
+	      return -1;
+	    }
+	}
+      close (fd);
+      return (0);
+    }
+}
+
+static int do_truncate (char * path)
+{
+  int result;
+
+  if (0 != chdir("/"))
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, errno,
+		       MSG_SUID_ERROR,
+		       sh_error_message(errno));
+    }
+
+  result = do_truncate_int (path, 0);
+
+  if (0 != chdir("/"))
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, errno,
+		       MSG_SUID_ERROR,
+		       sh_error_message(errno));
+    }
+  return result;
+}
 
 int sh_suidchk_check_internal (char * iname)
 {
   DIR *           thisDir = NULL;
+  FILE *          filePtr = NULL;
   struct dirent * thisEntry;
   char          * tmpcat;
   char          * tmp;
+  char          * msg;
+  char          * dir;
+  char          * filetmp;
+  char          * basetmp;
+  char            timestrc[32];
+  char            timestra[32];
+  char            timestrm[32];
+  char            buffer[1024];
   struct stat     buf;
   int             status;
+  int             count;
+  int             readFile = -1;
+  int             writeFile = -1;
   char          * fs;
-  long            sl_status;
+  long            sl_status = SL_ENONE;
+  struct stat     fileInfo;
 
   file_type       theFile;
   char            fileHash[KEY_LEN + 1];
@@ -156,18 +395,17 @@ int sh_suidchk_check_internal (char * iname)
       SL_RETURN( (-1), _("sh_suidchk_check_internal"));
     }
 
-  if (sig_terminate == 1)
+  if (sig_urgent == 1) {
     SL_RETURN( (0), _("sh_suidchk_check_internal"));
+  }
 
-  sl_set_suid();
   thisDir = opendir (iname);
-  sl_unset_suid();
 
   if (thisDir == NULL)
     {
       status = errno;
       tmp = sh_util_safe_name(iname);
-      sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_OPENDIR,
+      sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_SUBGEN,
 		       sh_error_message(status),
 		       tmp );
       SH_FREE(tmp);
@@ -176,8 +414,9 @@ int sh_suidchk_check_internal (char * iname)
 
   do {
 
-    if (sig_terminate == 1)
+    if (sig_urgent == 1) {
       SL_RETURN( (0), _("sh_suidchk_check_internal"));
+    }
 
 
     thisEntry = readdir (thisDir);
@@ -190,12 +429,18 @@ int sh_suidchk_check_internal (char * iname)
       if (sl_strcmp (thisEntry->d_name, "..") == 0)
 	continue;
 
-      tmpcat = SH_ALLOC(PATH_MAX+1);
-      sl_status = sl_strlcpy(tmpcat, iname,                 PATH_MAX+1);
-      if (! SL_ISERROR(sl_status) && tmpcat[1] != '\0')
-	sl_status = sl_strlcat(tmpcat, "/",                 PATH_MAX+1);
+      tmpcat = SH_ALLOC(PATH_MAX);
+      sl_strlcpy(tmpcat, iname,                 PATH_MAX);
+
+      if ((sl_strlen(tmpcat) != sl_strlen(iname)) || (tmpcat[0] == '\0'))
+	sl_status = SL_ETRUNC;
+      else
+	{
+	  if (tmpcat[1] != '\0') 
+	    sl_status = sl_strlcat(tmpcat, "/",                 PATH_MAX);
+	}
       if (! SL_ISERROR(sl_status))
-	sl_status = sl_strlcat(tmpcat, thisEntry->d_name,   PATH_MAX+1);
+	sl_status = sl_strlcat(tmpcat, thisEntry->d_name,   PATH_MAX);
       if (SL_ISERROR(sl_status))
 	{
 	  tmp = sh_util_safe_name(tmpcat);
@@ -210,20 +455,20 @@ int sh_suidchk_check_internal (char * iname)
       ++FileLimNum;
       ++FileLimTotal;
 
-      if (ShSuidchkFps > 0 && FileLimNum > ShSuidchkFps && FileLimTotal > 0)
+      if ((ShSuidchkFps > 0 && FileLimNum > ShSuidchkFps && FileLimTotal > 0)&&
+	  (ShSuidchkYield == S_FALSE))
 	{
 	  FileLimNum  = 0;
 	  FileLimNow  = time(NULL);
  
 	  if ( (FileLimNow  - FileLimStart) > 0 && 
 	       FileLimTotal/(FileLimNow  - FileLimStart) > ShSuidchkFps )
-	    sleep( (int)(FileLimTotal/(FileLimNow  - FileLimStart)) /
-		   ShSuidchkFps );
+	    retry_msleep( (int)(FileLimTotal/(FileLimNow  - FileLimStart)) /
+		   ShSuidchkFps , 0);
 	}
 	      
-      sl_set_suid();
       status = retry_lstat(FIL__, __LINE__, tmpcat, &buf);
-      sl_unset_suid();
+
       if (status != 0)
 	{
 	  status = errno;
@@ -260,12 +505,12 @@ int sh_suidchk_check_internal (char * iname)
 		   (0 !=(S_ISUID & buf.st_mode)||0 !=(S_ISGID & buf.st_mode)))
 	    {
 	      
-	      sl_strlcpy (theFile.fullpath, tmpcat, PATH_MAX+1);
+	      sl_strlcpy (theFile.fullpath, tmpcat, PATH_MAX);
 	      theFile.check_mask = sh_files_maskof(SH_LEVEL_READONLY);
 	      theFile.reported   = S_FALSE;
 	      status = sh_unix_getinfo (ShDFLevel[SH_ERR_T_RO], 
 					thisEntry->d_name,
-					&theFile, fileHash);
+					&theFile, fileHash, 0);
 
 	      tmp = sh_util_safe_name(tmpcat);
 
@@ -281,11 +526,23 @@ int sh_suidchk_check_internal (char * iname)
 		  if (sh.flag.checkSum == SH_CHECK_INIT  && 
 		      sh.flag.update == S_TRUE )
 		    {
-		      sh_hash_compdata (SH_LEVEL_READONLY, &theFile, fileHash);
-		      sh_hash_pushdata (&theFile, fileHash);
-		      sh_error_handle ((-1), FIL__, __LINE__, 
-				       0, MSG_SUID_FOUND,
-				       tmp );
+		      if (-1 == sh_hash_have_it (tmpcat))
+			{
+			  sh_error_handle ((-1), FIL__, __LINE__, 
+					   0, MSG_SUID_FOUND,
+					   tmp );
+			}
+		      else
+			{
+			  sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 
+					   0, MSG_SUID_FOUND,
+					   tmp );
+			}
+		      if (0 == sh_hash_compdata (SH_LEVEL_READONLY, 
+						 &theFile, fileHash))
+			{
+			  sh_hash_pushdata (&theFile, fileHash);
+			}
 		    }
 		  else if (sh.flag.checkSum == SH_CHECK_INIT  && 
 			   sh.flag.update == S_FALSE )
@@ -301,13 +558,248 @@ int sh_suidchk_check_internal (char * iname)
 				       0, MSG_SUID_FOUND,
 				       tmp );
 		      if (-1 == sh_hash_have_it (tmpcat))
-			sh_error_handle (ShSuidchkSeverity, FIL__, __LINE__, 
-					 0, MSG_SUID_POLICY,
-					 _("suid/sgid file not in database"),
-					 tmp );
+			{
+			  msg = SH_ALLOC(SH_BUFSIZE);
+			  msg[0] = '\0';
+			  sl_strlcpy (timestrc, 
+				      sh_unix_gmttime (theFile.ctime), 32);
+			  sl_strlcpy (timestra, 
+				      sh_unix_gmttime (theFile.atime), 32);
+			  sl_strlcpy (timestrm, 
+				      sh_unix_gmttime (theFile.mtime), 32);
+#ifdef SH_USE_XML
+			  sl_snprintf(msg, SH_BUFSIZE, _("owner_new=\"%s\" iowner_new=\"%ld\" group_new=\"%s\" igroup_new=\"%ld\" size_new=\"%ld\" ctime_new=\"%s\" atime_new=\"%s\" mtime_new=\"%s\""), 
+				      theFile.c_owner, theFile.owner, 
+				      theFile.c_group, theFile.group, 
+				      (unsigned long) theFile.size, 
+				      timestrc, timestra, timestrm);
+#else
+			  sl_snprintf(msg, SH_BUFSIZE, _("owner_new=<%s>, iowner_new=<%ld>, group_new=<%s>, igroup_new=<%ld>, filesize=<%ld>, ctime=<%s>, atime=<%s>, mtime=<%s>"), 
+				      theFile.c_owner, theFile.owner, 
+				      theFile.c_group, theFile.group, 
+				      (unsigned long) theFile.size, 
+				      timestrc, timestra, timestrm);
+#endif
+
+			  sh_error_handle (ShSuidchkSeverity, FIL__, __LINE__, 
+					   0, MSG_SUID_POLICY,
+					   _("suid/sgid file not in database"),
+					   tmp, msg );
+			  SH_FREE(msg);
+
+			  /* Quarantine file according to configured method
+                          */
+			  if (ShSuidchkQEnable == S_TRUE)
+			    {
+			      switch (ShSuidchkQMethod)
+				{
+				  case SH_Q_DELETE:
+				    /* if (unlink (theFile.fullpath) == -1) */
+				    if (do_truncate (theFile.fullpath) == -1)
+				      {
+					status = errno;
+					msg = SH_ALLOC(SH_BUFSIZE);
+					sl_snprintf(msg, SH_BUFSIZE, _("Problem quarantining file.  File NOT quarantined.  errno = %ld"), status);
+					sh_error_handle (ShSuidchkSeverity,
+							 FIL__, __LINE__, 
+							 status,
+							 MSG_SUID_QREPORT, msg,
+							 tmp );
+					SH_FREE(msg);
+				      }
+				    else
+				      {
+				        sh_error_handle (ShSuidchkSeverity,
+							 FIL__, __LINE__, 0,
+							 MSG_SUID_QREPORT,
+							 _("Quarantine method applied"),
+							 tmp );
+				      }
+				    break;
+				  case SH_Q_CHANGEPERM:
+				    if (stat(tmp, &fileInfo) == -1)
+				      {
+					status = errno;
+					msg = SH_ALLOC(SH_BUFSIZE);
+					sl_snprintf(msg, SH_BUFSIZE, _("I/O error.  errno = %ld"), status);
+					sh_error_handle (ShSuidchkSeverity,
+							 FIL__, __LINE__, 
+							 status,
+							 MSG_SUID_QREPORT, msg,
+							 tmp );
+					SH_FREE(msg);
+				      }
+				    else
+				      {
+					if ((fileInfo.st_mode & S_ISUID) > 0)
+					  fileInfo.st_mode -= S_ISUID;
+					if ((fileInfo.st_mode & S_ISGID) > 0)
+					  fileInfo.st_mode -= S_ISGID;
+					if (chmod(tmp, fileInfo.st_mode) == -1)
+					  {
+					    status = errno;
+					    msg = SH_ALLOC(SH_BUFSIZE);
+					    sl_snprintf(msg, SH_BUFSIZE, _("Problem quarantining file.  File NOT quarantined.  errno = %ld"), status);
+					    sh_error_handle (ShSuidchkSeverity,
+							     FIL__, __LINE__, 
+							     status,
+							     MSG_SUID_QREPORT,
+							     msg, tmp );
+					    SH_FREE(msg);
+					  }
+					else
+					  {
+					    sh_error_handle (ShSuidchkSeverity,
+							     FIL__, __LINE__, 
+							     0,
+							     MSG_SUID_QREPORT,
+							     _("Quarantine method applied"),
+							     tmp );
+					  }
+					}
+				    break;
+				  case SH_Q_MOVE:
+				    dir = SH_ALLOC(PATH_MAX+1);
+				    sl_strlcpy (dir, DEFAULT_QDIR, PATH_MAX+1);
+				    if (access (dir, F_OK) != 0)
+				      {
+					status = errno;
+					msg = SH_ALLOC(SH_BUFSIZE);
+					sl_snprintf(msg, SH_BUFSIZE, _("Problem quarantining file.  File NOT quarantined.  errno = %ld"), status);
+					sh_error_handle (ShSuidchkSeverity,
+							 FIL__, __LINE__, 
+							 status,
+							 MSG_SUID_QREPORT, msg,
+							 tmp );
+					SH_FREE(msg);
+				      }
+				    else
+				      {
+					if (stat (tmp, &fileInfo) == -1)
+					  {
+					    status = errno;
+					    msg = SH_ALLOC(SH_BUFSIZE);
+					    sl_snprintf(msg, SH_BUFSIZE, _("I/O error.  errno = %ld"), status);
+					    sh_error_handle (ShSuidchkSeverity,
+							     FIL__, __LINE__, 
+							     status,
+							     MSG_SUID_QREPORT,
+							     msg, tmp );
+					    SH_FREE(msg);
+					  }
+					else
+					  {
+					    basetmp = SH_ALLOC(1 + sl_strlen(theFile.fullpath));
+					    sl_strlcpy(basetmp, theFile.fullpath, 
+						       1 + sl_strlen(theFile.fullpath));
+					    filetmp = SH_ALLOC(PATH_MAX+1);
+					    sl_snprintf(filetmp, PATH_MAX+1, "%s/%s", 
+							DEFAULT_QDIR, basename(basetmp));
+					    SH_FREE(basetmp);
+					    
+					    readFile  = open (theFile.fullpath, O_RDONLY);
+					    if (readFile != -1)
+					      writeFile = open (filetmp, O_WRONLY|O_CREAT);
+
+					    if ((readFile == -1) || (writeFile == -1))
+					      {
+						status = errno;
+						msg = SH_ALLOC(SH_BUFSIZE);
+						sl_snprintf(msg, SH_BUFSIZE, _("Problem quarantining file.  File NOT quarantined.  errno = %ld"), status);
+						sh_error_handle (ShSuidchkSeverity,
+								 FIL__, __LINE__, status,
+								 MSG_SUID_QREPORT,
+								 msg, tmp );
+						SH_FREE(msg);
+					      }
+					    else
+					      {
+						while ((count = read (readFile, buffer, sizeof (buffer))) > 0)
+						  if (write (writeFile, buffer, count) != count)
+						    {
+						      status = errno;
+						      msg = SH_ALLOC(SH_BUFSIZE);
+					    	      sl_snprintf(msg, SH_BUFSIZE, _("I/O error.  errno = %ld"), status);
+						      sh_error_handle (ShSuidchkSeverity,
+								       FIL__,
+								       __LINE__,
+								       status,
+								       MSG_SUID_QREPORT,
+								       msg, tmp );
+						      SH_FREE(msg);
+						    }
+					      }
+					    close (readFile);
+					    close (writeFile);
+					    /* if (unlink (theFile.fullpath) == -1) */
+					    if (do_truncate (theFile.fullpath) == -1)
+					      {
+						status = errno;
+						msg = SH_ALLOC(SH_BUFSIZE);
+						sl_snprintf(msg, SH_BUFSIZE, _("Problem quarantining file.  File NOT quarantined.  errno = %ld"), status);
+						sh_error_handle (ShSuidchkSeverity,
+								 FIL__, __LINE__, status,
+								 MSG_SUID_QREPORT,
+								 msg, tmp );
+						SH_FREE(msg);
+					      }
+					    else
+					      {
+						sl_snprintf(filetmp, PATH_MAX+1, "%s/%s.info", 
+							    DEFAULT_QDIR, 
+							    basename(theFile.fullpath));
+						filePtr = fopen (filetmp, "w+");
+						fprintf(filePtr, _("File Info:\n filename=%s\n size=%ld\n owner=%s(%d)\n group=%s(%d)\n ctime=%s\n atime=%s\n mtime=%s\n"), 
+							theFile.fullpath, 
+							(unsigned long) theFile.size, 
+							theFile.c_owner, (int) theFile.owner, 
+							theFile.c_group, (int) theFile.group, 
+							timestrc, timestra, timestrm);
+						fclose (filePtr);
+					
+						sh_error_handle (ShSuidchkSeverity,
+							 	 FIL__,__LINE__,
+								 0, MSG_SUID_QREPORT,
+								 _("Quarantine method applied"),
+								 tmp );
+						if (chmod(filetmp, S_IRUSR | S_IWUSR) == -1)
+					  	{
+					    	  status = errno;
+					    	  msg = SH_ALLOC(SH_BUFSIZE);
+					    	  sl_snprintf(msg, SH_BUFSIZE, _("Problem setting permissions on quarantined file.  errno = %ld"), status);
+					    	  sh_error_handle (ShSuidchkSeverity,
+							     	   FIL__,__LINE__,
+								   status, MSG_SUID_QREPORT,
+							     	   msg, tmp );
+					    	  SH_FREE(msg);
+					  	}
+					      }
+					    SH_FREE(filetmp);
+					  }
+				      }
+				    SH_FREE(dir);
+				    break;
+				  default:
+				    sh_error_handle (ShSuidchkSeverity, FIL__,
+						     __LINE__, 0, MSG_SUID_QREPORT,
+						     _("Bad quarantine method"),
+						     tmp);
+				    break;
+				}
+			    }
+			  else
+			    {
+			      /* 1.8.1 push file to in-memory database
+			       */
+			      sh_hash_compdata (SH_LEVEL_READONLY,
+						&theFile, fileHash);
+			    }
+			}
 		      else
-			sh_hash_compdata (SH_LEVEL_READONLY, 
-					  &theFile, fileHash);
+			{
+			  sh_hash_compdata (SH_LEVEL_READONLY, 
+					    &theFile, fileHash);
+			}
 		    }
 		}
 	      SH_FREE(tmp);
@@ -316,6 +808,19 @@ int sh_suidchk_check_internal (char * iname)
 	}
       SH_FREE(tmpcat);
     }
+#ifdef HAVE_SCHED_YIELD
+    if (ShSuidchkYield == S_TRUE)
+      {
+	if (sched_yield() == -1)
+	  {
+	    status = errno;
+	    sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_SUBGEN,
+		             _("Failed to release time slice"),
+			     _("sh_suidchk_check_internal") );
+	    
+	  }
+      }
+#endif
   }  while (thisEntry != NULL);
 
   closedir (thisDir);
@@ -335,6 +840,7 @@ int sh_suidchk_init ()
   return (0);
 }
 
+
 /*************
  *
  * module cleanup
@@ -353,6 +859,9 @@ int sh_suidchk_end ()
  *************/
 int sh_suidchk_timer (unsigned long tcurrent)
 {
+  if (sh.flag.checkSum == SH_CHECK_INIT)
+    return -1;
+
   if (ShSuidchkSched != NULL)
     {
       return test_sched(ShSuidchkSched);
@@ -416,12 +925,32 @@ int sh_suidchk_set_exclude (char * c)
 {
   SL_ENTER(_("sh_suidchk_set_exclude"));
   if (c == NULL)
-    return -1;
+    {
+      SL_RETURN(-1, _("sh_suidchk_set_exclude"));
+    }
+
+  if (0 == sl_strncmp(c, _("NULL"), 4))
+    {
+      if (ShSuidchkExclude != NULL)
+	SH_FREE(ShSuidchkExclude);
+      ShSuidchkExclude = NULL;
+      SL_RETURN(0, _("sh_suidchk_set_exclude"));
+    }
 
   if (ShSuidchkExclude != NULL)
     SH_FREE(ShSuidchkExclude);
-  ShSuidchkExclude = sh_util_strdup (c);
-  ExcludeLen       = sl_strlen (ShSuidchkExclude);
+
+  /* 1.8.1 add trailing slash
+   */
+  ExcludeLen       = sl_strlen(c);
+  if (c[ExcludeLen-1] != '/')
+    {
+      ExcludeLen++;
+    }
+  ShSuidchkExclude = SH_ALLOC(ExcludeLen + 1);
+  sl_strlcpy(ShSuidchkExclude, c, (ExcludeLen + 1));
+  ShSuidchkExclude[ExcludeLen-1] = '/';
+
   SL_RETURN(0, _("sh_suidchk_set_exclude"));
 }
 
@@ -442,25 +971,55 @@ int sh_suidchk_set_timer (char * c)
   SL_RETURN( 0, _("sh_suidchk_set_timer"));
 }
 
+
+int sh_suidchk_free_schedule ()
+{
+  sh_schedule_t * current = ShSuidchkSched;
+  sh_schedule_t * next    = NULL;
+
+  while (current != NULL)
+    {
+      next = current->next;
+      SH_FREE(current);
+      current = next;
+    }
+  ShSuidchkSched = NULL;
+  return 0;
+}
+
 int sh_suidchk_set_schedule (char * str)
 {
   int status;
+  sh_schedule_t * newSched = NULL;
 
   SL_ENTER(_("sh_suidchk_set_schedule"));
 
+  /*
   if (ShSuidchkSched != NULL)
     {
       SH_FREE(ShSuidchkSched);
       ShSuidchkSched = NULL;
     }
-  ShSuidchkSched = SH_ALLOC(sizeof(sh_schedule_t));
-  status = create_sched(str, ShSuidchkSched);
-  if (status != 0)
+  */
+
+  if (0 == sl_strncmp(str, _("NULL"), 4))
     {
-      SH_FREE(ShSuidchkSched);
-      ShSuidchkSched = NULL;
+      sh_suidchk_free_schedule ();
+      return 0;
     }
 
+  newSched = SH_ALLOC(sizeof(sh_schedule_t));
+  status = create_sched(str, newSched);
+  if (status != 0)
+    {
+      SH_FREE(newSched);
+      newSched = NULL;
+    }
+  else
+    {
+      newSched->next = ShSuidchkSched;
+      ShSuidchkSched = newSched;
+    }
   SL_RETURN( status, _("sh_suidchk_set_schedule"));
 }
 
@@ -483,6 +1042,19 @@ int sh_suidchk_set_fps (char * c)
   SL_RETURN( 0, _("sh_suidchk_set_fps"));
 }
 
+int sh_suidchk_set_yield (char * c)
+{
+  int i;
+  SL_ENTER(_("sh_suidchk_set_yield"));
+#ifdef HAVE_SCHED_YIELD
+  i = sh_util_flagval(c, &ShSuidchkYield);
+#else
+  (void) c; /* cast to void to avoid compiler warning */
+  i = -1;
+#endif
+  SL_RETURN(i, _("sh_suidchk_set_yield"));
+}
+
 int sh_suidchk_set_activate (char * c)
 {
   int i;
@@ -491,6 +1063,70 @@ int sh_suidchk_set_activate (char * c)
   SL_RETURN(i, _("sh_suidchk_set_activate"));
 }
 
+int sh_suidchk_set_quarantine (char * c)
+{
+  int i;
+  SL_ENTER(_("sh_suidchk_set_quarantine"));
+  i = sh_util_flagval(c, &ShSuidchkQEnable);
+  SL_RETURN(i, _("sh_suidchk_set_quarantine"));
+}
+
+int sh_suidchk_set_qdelete (char * c)
+{
+  int i;
+  SL_ENTER(_("sh_suidchk_set_qdelete"));
+  i = sh_util_flagval(c, &ShSuidchkQDelete);
+  SL_RETURN(i, _("sh_suidchk_set_qdelete"));
+}
+
+int sh_suidchk_set_qmethod (char * c)
+{
+  long val;
+  int  ret = 0;
+
+  SL_ENTER(_("sh_suidchk_set_qmethod"));
+
+  val = strtol (c, (char **)NULL, 10);
+  if (val < 0)
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS,
+		       _("suidchk qmethod"), c);
+      ret = -1;
+    }
+  else
+    {
+      switch (val)
+      {
+        case SH_Q_DELETE:
+          ShSuidchkQMethod = SH_Q_DELETE;
+          break;
+        case SH_Q_CHANGEPERM:
+          ShSuidchkQMethod = SH_Q_CHANGEPERM;
+          break;
+        case SH_Q_MOVE:
+          if (access (DEFAULT_QDIR, F_OK) != 0)
+	    {
+	      if (mkdir (DEFAULT_QDIR, 0640) == -1)
+		{
+		  sh_error_handle ((-1), FIL__, __LINE__, EINVAL,
+				   MSG_SUID_ERROR,
+				   _("Unable to create quarantine directory"));
+		}
+	    }
+          ShSuidchkQMethod = SH_Q_MOVE;
+          break;
+        default:
+	  sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS,
+			   _("suidchk qmethod"), c);
+          ShSuidchkQMethod = -1;
+	  ret = -1;
+          break;
+      }
+    }
+
+  SL_RETURN( ret, _("sh_suidchk_set_qmethod"));
+}
+
 #if defined(FSTYPE_STATFS) || defined(FSTYPE_AIX_STATFS)
 /* dirname.c -- return all but the last element in a path
    Copyright (C) 1990 Free Software Foundation, Inc.
@@ -761,13 +1397,61 @@ fstype_to_string (t)
   switch (t)
     {
     case MNT_AIX:
-      return _("aix");	/* NFS filesystems are actually MNT_AIX. */
+      return _("aix");	/* AIX 4.3: NFS filesystems are actually MNT_AIX. */
+#ifdef MNT_NAMEFS
+    case MNT_NAMEFS:
+      return _("namefs");
+#endif
     case MNT_NFS:
       return _("nfs");
     case MNT_JFS:
       return _("jfs");
     case MNT_CDROM:
       return _("cdrom");
+#ifdef MNT_PROCFS
+    case MNT_PROCFS:
+      return _("procfs");
+#endif
+#ifdef MNT_SFS
+    case MNT_SFS:
+      return _("sfs");
+#endif
+#ifdef MNT_CACHEFS
+    case MNT_CACHEFS:
+      return _("cachefs");
+#endif
+#ifdef MNT_NFS3
+    case MNT_NFS3:
+      return _("nfs3");
+#endif
+#ifdef MNT_AUTOFS
+    case MNT_AUTOFS:
+      return _("autofs");
+#endif
+#ifdef MNT_VXFS
+    case MNT_VXFS:
+      return _("vxfs");
+#endif
+#ifdef MNT_VXODM
+    case MNT_VXODM:
+      return _("veritasfs");
+#endif
+#ifdef MNT_UDF
+    case MNT_UDF:
+      return _("udfs");
+#endif
+#ifdef MNT_NFS4
+    case MNT_NFS4:
+      return _("nfs4");
+#endif
+#ifdef MNT_RFS4
+    case MNT_RFS4:
+      return _("nfs4");
+#endif
+#ifdef MNT_CIFS
+    case MNT_CIFS:
+      return _("cifs");
+#endif
     default:
       return "?";
     }
@@ -988,12 +1672,13 @@ filesystem_type_uncached (path, relpath, statp)
        if( fss.f_flag & ST_NOSUID )
          type = _("nosuid");
     }
+  (void) statp; /* fix compiler warning */
 #endif
 
 #ifdef FSTYPE_STATFS		/* 4.4BSD.  */
   struct statfs fss;
   char *p;
-#ifdef MNT_VISFLAGMASK
+#if defined(MNT_VISFLAGMASK) && defined(HAVE_STRUCT_STATFS_F_FLAGS)
   int flags;
 #endif
   /* char * sh_dirname(const char *path); */
