@@ -92,14 +92,12 @@ SL_TICKET tiger_fd = (-1);
 
 #ifdef HAVE_LONG_LONG
 static
-word64 * sh_tiger_hash_val (char * filename, 
-			    TigerType what, 
-			    unsigned long Length)
+word64 * sh_tiger_hash_val (char * filename, TigerType what, 
+			    unsigned long Length, int timeout)
 #else
 static
-sh_word32 * sh_tiger_hash_val (char * filename, 
-			       TigerType what, 
-			       unsigned long Length)
+sh_word32 * sh_tiger_hash_val (char * filename, TigerType what, 
+			       unsigned long Length, int timeout)
 #endif
 {
   SL_TICKET  fd;
@@ -108,14 +106,19 @@ sh_word32 * sh_tiger_hash_val (char * filename,
   int  blk;
   char    * tmp;
   sh_byte * bptr;
-  sh_byte buffer[PRIV_MAX+2];
+  sh_byte buffer[PRIV_MAX + 72];
   sh_byte bbuf[64];
 
+  unsigned long pages_read;
   uid_t   euid;
 
   unsigned long ncount = 0, nblocks = 0;
   unsigned long  t, msb, lsb;
 
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+  extern long IO_Limit;
+#endif
+
 #ifdef HAVE_LONG_LONG
 #define TIGER_CAST (word64*)
   static word64 res[3];
@@ -161,33 +164,43 @@ sh_word32 * sh_tiger_hash_val (char * filename,
 	}
       
 #if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
-    if (skey->mlock_failed == GOOD) 
+    if (skey->mlock_failed == SL_FALSE) 
       {
-	sl_set_suid ();
-        if ( (-1) == mlock((char *) buffer, (PRIV_MAX+2) * sizeof(sh_byte))) 
-          skey->mlock_failed = BAD;
-	sl_unset_suid ();
+        if ( (-1) == sh_unix_mlock((char *)buffer,(PRIV_MAX)*sizeof(sh_byte))) 
+          skey->mlock_failed = SL_TRUE;
       }
 #else
-    if (skey->mlock_failed == GOOD)
-      skey->mlock_failed = BAD;
+    if (skey->mlock_failed == SL_FALSE)
+      skey->mlock_failed = SL_TRUE;
 #endif
 
 #ifdef TIGER_DBG
     tiger_dbg (res, 0, nblocks, ncount);
 #endif
 
+    pages_read = 0;
+
     while (1) 
       {
-	count = sl_read (fd, buffer, PRIV_MAX);
+	if (timeout > 0)
+	  count = sl_read_timeout (fd, buffer, PRIV_MAX, timeout);
+	else
+	  count = sl_read         (fd, buffer, PRIV_MAX);
+
+	++pages_read;
 
 	if (SL_ISERROR (count)) 
 	  {
 	    TPT((0, FIL__ , __LINE__ , _("msg=<SL_ISERROR (%ld)>\n"), count)); 
 	    tmp = sh_util_safe_name (filename);
-	    sh_error_handle (ShDFLevel[SH_ERR_T_FILE], FIL__, __LINE__, 0,
-			     MSG_E_READ, tmp);
+	    if (count == SL_TIMEOUT)
+	      sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, 0, MSG_E_TIMEOUT,
+			       timeout, tmp);
+	    else
+	      sh_error_handle (ShDFLevel[SH_ERR_T_FILE], FIL__, __LINE__, 0,
+			       MSG_E_READ, tmp);
 	    SH_FREE(tmp);
+	    sh_unix_munlock((char*)buffer, (PRIV_MAX)*sizeof(sh_byte));
 	    SL_RETURN( NULL, _("sh_tiger_hash_val"));
 	  }
 
@@ -195,7 +208,8 @@ sh_word32 * sh_tiger_hash_val (char * filename,
 
 	blk      = (count / 64); /* number of 64-byte words */
 
-	/* nblocks += blk; number of 64-byte words */
+	/* nblocks += blk; number of 64-byte words 
+	 */
 	ncount = count - ((count/64)*64);
 	
 	bptr = buffer;
@@ -205,15 +219,30 @@ sh_word32 * sh_tiger_hash_val (char * filename,
 	    
 	    tiger_t(TIGER_CAST bptr, 64, res);
 	    ++nblocks;
-	    sh.stat.bytes_hashed += 64;
+	    sh.statistics.bytes_hashed += 64;
 	    
 #ifdef TIGER_DBG
 	    tiger_dbg (res, 3, nblocks, ncount);
 #endif
 	  }
 	
-	if (blk < 8) 
+	if (blk < 64) /* this must be (PRIV_MAX / 64) */
 	  break;
+
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+	if (sig_termfast == 1) 
+	  {
+	    memset (bbuf,   0, 64);
+	    memset (buffer, 0, PRIV_MAX);
+	    sh_unix_munlock( (char *) buffer, (PRIV_MAX) * sizeof(sh_byte));
+	    SL_RETURN( NULL, _("sh_tiger_hash_val"));
+	  }
+	if ((IO_Limit) > 0 && (pages_read == 32)) /* check for I/O limit */
+	  {
+	    sh_unix_io_pause ();
+	    pages_read = 0;
+	  }
+#endif 
       }
 
     TPT((0, FIL__, __LINE__ , _("msg=<Last block.>\n")));
@@ -254,7 +283,7 @@ sh_word32 * sh_tiger_hash_val (char * filename,
         while( j < 64 )
 	  bbuf[j++] = 0;
 	tiger_t(TIGER_CAST bbuf, 64, res);
-	sh.stat.bytes_hashed += 64;
+	sh.statistics.bytes_hashed += 64;
 	++nblocks; ncount = 0;
         memset(bbuf, 0, 56 ); 
       }
@@ -273,7 +302,7 @@ sh_word32 * sh_tiger_hash_val (char * filename,
     bbuf[63] = msb >> 24;
 
     tiger_t(TIGER_CAST bbuf, 64, res);
-    sh.stat.bytes_hashed += 64;
+    sh.statistics.bytes_hashed += 64;
     
 #ifdef TIGER_DBG
     tiger_dbg (res, 7, nblocks, ncount);
@@ -282,7 +311,7 @@ sh_word32 * sh_tiger_hash_val (char * filename,
     memset (bbuf,   0, 64);
     memset (buffer, 0, PRIV_MAX);
 
-    MUNLOCK( (char *) buffer, (PRIV_MAX+2) * sizeof(sh_byte));
+    sh_unix_munlock( (char *) buffer, (PRIV_MAX) * sizeof(sh_byte));
 
     if (what == TIGER_FILE)
       sl_close (fd);
@@ -320,6 +349,10 @@ sh_word32 * sh_tiger_hash_val (char * filename,
 
 typedef sh_word32     uint32;
 typedef unsigned char uint8;
+/* this typedef clashes with the Rijndael implementation (namespace ...) */
+#ifndef SH_ENCRYPT
+typedef unsigned char        byte;
+#endif
 
 /* Structure to save state of computation between the single steps.  */
 typedef struct
@@ -343,7 +376,19 @@ int md5Reset(register md5Param* p)
         return 0;
 }
 
+#if defined(__GNUC__) && defined(__i386__)
+static inline UINT32
+ROTL32( UINT32 x, int s)
+{
+        __asm__("roll %%cl,%0"
+                :"=r" (x)
+                :"0" (x),"c" (s));
+        return x;
+}
+#else
 #define ROTL32(x, s) (((x) << (s)) | ((x) >> (32 - (s))))
+#endif
+
 
 #define FF(a, b, c, d, w, s, t) \
         a += ((b&(c^d))^d) + w + t;     \
@@ -397,22 +442,24 @@ void md5Process(md5Param* p)
 
         a = p->h[0]; b = p->h[1]; c = p->h[2]; d = p->h[3];
 
-        FF(a, b, c, d, w[ 0],  7, 0xd76aa478);
-        FF(d, a, b, c, w[ 1], 12, 0xe8c7b756);
-        FF(c, d, a, b, w[ 2], 17, 0x242070db);
-        FF(b, c, d, a, w[ 3], 22, 0xc1bdceee);
-        FF(a, b, c, d, w[ 4],  7, 0xf57c0faf);
-        FF(d, a, b, c, w[ 5], 12, 0x4787c62a);
-        FF(c, d, a, b, w[ 6], 17, 0xa8304613);
-        FF(b, c, d, a, w[ 7], 22, 0xfd469501);
-        FF(a, b, c, d, w[ 8],  7, 0x698098d8);
-        FF(d, a, b, c, w[ 9], 12, 0x8b44f7af);
-        FF(c, d, a, b, w[10], 17, 0xffff5bb1);
-        FF(b, c, d, a, w[11], 22, 0x895cd7be);
-        FF(a, b, c, d, w[12],  7, 0x6b901122);
-        FF(d, a, b, c, w[13], 12, 0xfd987193);
-        FF(c, d, a, b, w[14], 17, 0xa679438e);
-        FF(b, c, d, a, w[15], 22, 0x49b40821);
+        FF(a, b, c, d, (*w++),  7, 0xd76aa478);
+        FF(d, a, b, c, (*w++), 12, 0xe8c7b756);
+        FF(c, d, a, b, (*w++), 17, 0x242070db);
+        FF(b, c, d, a, (*w++), 22, 0xc1bdceee);
+        FF(a, b, c, d, (*w++),  7, 0xf57c0faf);
+        FF(d, a, b, c, (*w++), 12, 0x4787c62a);
+        FF(c, d, a, b, (*w++), 17, 0xa8304613);
+        FF(b, c, d, a, (*w++), 22, 0xfd469501);
+        FF(a, b, c, d, (*w++),  7, 0x698098d8);
+        FF(d, a, b, c, (*w++), 12, 0x8b44f7af);
+        FF(c, d, a, b, (*w++), 17, 0xffff5bb1);
+        FF(b, c, d, a, (*w++), 22, 0x895cd7be);
+        FF(a, b, c, d, (*w++),  7, 0x6b901122);
+        FF(d, a, b, c, (*w++), 12, 0xfd987193);
+        FF(c, d, a, b, (*w++), 17, 0xa679438e);
+        FF(b, c, d, a, (*w++), 22, 0x49b40821);
+
+	w = p->data;
 
         GG(a, b, c, d, w[ 1],  5, 0xf61e2562);
         GG(d, a, b, c, w[ 6],  9, 0xc040b340);
@@ -572,17 +619,22 @@ int md5Digest(md5Param* p, uint32* data)
 /* Compute MD5 message digest for bytes read from STREAM.  The
    resulting message digest number will be written into the 16 bytes
    beginning at RESBLOCK.  */
-static int md5_stream(char * filename, void *resblock)
+static int md5_stream(char * filename, void *resblock, int timeout)
 {
   /* Important: BLOCKSIZE must be a multiple of 64.  */
-  static const int BLOCKSIZE = 4096;
+  static const int BLOCKSIZE = 8192;
   md5Param ctx;
-  char buffer[4168]; /* BLOCKSIZE + 72  AIX compiler chokes */
+  char buffer[8264]; /* BLOCKSIZE + 72  AIX compiler chokes */
   off_t sum;
   SL_TICKET  fd;
   char * tmp;
   uid_t   euid;
 
+  unsigned long pages_read;
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+  extern long IO_Limit;
+#endif
+
   /* Initialize the computation context.  */
   md5Reset (&ctx);
 
@@ -599,6 +651,8 @@ static int md5_stream(char * filename, void *resblock)
       return -1;
     }
 
+  pages_read = 0;
+
   /* Iterate over full file contents.  */
   while (1) {
     /* We read the file in blocks of BLOCKSIZE bytes.  One call of the
@@ -609,14 +663,19 @@ static int md5_stream(char * filename, void *resblock)
 
     /* Read block.  Take care for partial reads.  */
     do {
-      n = sl_read(fd, buffer + sum, BLOCKSIZE - sum);
+
+      n = sl_read_timeout (fd, buffer + sum, BLOCKSIZE - sum, timeout);
 
       if (SL_ISERROR (n))
 	{
 	  TPT((0, FIL__ , __LINE__ , _("msg=<SL_ISERROR (%ld)>\n"), n));
 	  tmp = sh_util_safe_name (filename);
-	  sh_error_handle (ShDFLevel[SH_ERR_T_FILE], FIL__, __LINE__, 0,
-			   MSG_E_READ, tmp);
+	  if (n == SL_TIMEOUT)
+	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, 0, MSG_E_TIMEOUT,
+			     timeout, tmp);
+	  else
+	    sh_error_handle (ShDFLevel[SH_ERR_T_FILE], FIL__, __LINE__, 0,
+			     MSG_E_READ, tmp);
 	  SH_FREE(tmp);
 	  return -1;
 	}
@@ -626,6 +685,8 @@ static int md5_stream(char * filename, void *resblock)
     while (sum < BLOCKSIZE 
 	   && n != 0);
 
+    ++pages_read;
+
     /* If end of file is reached, end the loop.  */
     if (n == 0)
       break;
@@ -633,15 +694,27 @@ static int md5_stream(char * filename, void *resblock)
     /* Process buffer with BLOCKSIZE bytes.  Note that
        BLOCKSIZE % 64 == 0
     */
-    md5Update(&ctx, buffer, BLOCKSIZE);
-    sh.stat.bytes_hashed += BLOCKSIZE;
+    md5Update(&ctx, (byte*) buffer, BLOCKSIZE);
+    sh.statistics.bytes_hashed += BLOCKSIZE;
+
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+    if ((IO_Limit) > 0 && (pages_read == 32)) /* check for I/O limit */
+      {
+	sh_unix_io_pause ();
+	pages_read = 0;
+      }
+    if (sig_termfast == 1) 
+      {
+	return -1;
+      }
+#endif 
   }
 
   /* Add the last bytes if necessary.  */
   if (sum > 0)
     {
-      md5Update(&ctx, buffer, sum);
-      sh.stat.bytes_hashed += BLOCKSIZE;
+      md5Update(&ctx, (byte*) buffer, sum);
+      sh.statistics.bytes_hashed += BLOCKSIZE;
     }
 
   /* Construct result in desired memory.  */
@@ -650,8 +723,9 @@ static int md5_stream(char * filename, void *resblock)
   return 0;
 }
 
+
 char * sh_tiger_md5_hash  (char * filename, TigerType what, 
-			   unsigned long Length)
+			   unsigned long Length, int timeout)
 {
   int cnt = (int) Length;
   static char out[KEY_LEN+1];
@@ -665,10 +739,10 @@ char * sh_tiger_md5_hash  (char * filename, TigerType what,
       return out;
     }
 
-  md5_stream (filename, md5buffer);
+  md5_stream (filename, md5buffer, timeout);
 
   for (cnt = 0; cnt < 16; ++cnt)
-    sprintf (&out[cnt*2], _("%02X"), md5buffer[cnt]);
+    sprintf (&out[cnt*2], _("%02X"), md5buffer[cnt]);/* known to fit  */
   for (cnt = 32; cnt < KEY_LEN; ++cnt)
     out[cnt] = '0';
   out[KEY_LEN] = '\0';
@@ -1044,7 +1118,7 @@ void sha_digest(struct sha_ctx *ctx, sha_word8 *s)
 /* Compute SHA1 message digest for bytes read from STREAM.  The
    resulting message digest number will be written into the 16 bytes
    beginning at RESBLOCK.  */
-static int sha1_stream(char * filename, void *resblock)
+static int sha1_stream(char * filename, void *resblock, int timeout)
 {
   /* Important: BLOCKSIZE must be a multiple of 64.  */
   static const int BLOCKSIZE = 4096;
@@ -1055,6 +1129,11 @@ static int sha1_stream(char * filename, void *resblock)
   char * tmp;
   uid_t   euid;
 
+  unsigned long pages_read;
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+  extern long IO_Limit;
+#endif
+
   /* Initialize the computation context.  */
   sha_init(&ctx);
 
@@ -1072,6 +1151,9 @@ static int sha1_stream(char * filename, void *resblock)
     }
 
   /* Iterate over full file contents.  */
+
+  pages_read = 0;
+
   while (1) {
     /* We read the file in blocks of BLOCKSIZE bytes.  One call of the
        computation function processes the whole buffer so that with the
@@ -1081,14 +1163,18 @@ static int sha1_stream(char * filename, void *resblock)
 
     /* Read block.  Take care for partial reads.  */
     do {
-      n = sl_read(fd, buffer + sum, BLOCKSIZE - sum);
+      n = sl_read_timeout(fd, buffer + sum, BLOCKSIZE - sum, timeout);
 
       if (SL_ISERROR (n))
 	{
 	  TPT((0, FIL__ , __LINE__ , _("msg=<SL_ISERROR (%ld)>\n"), n));
 	  tmp = sh_util_safe_name (filename);
-	  sh_error_handle (ShDFLevel[SH_ERR_T_FILE], FIL__, __LINE__, 0,
-			   MSG_E_READ, tmp);
+	  if (n == SL_TIMEOUT)
+	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, 0, MSG_E_TIMEOUT,
+			     timeout, tmp);
+	  else
+	    sh_error_handle (ShDFLevel[SH_ERR_T_FILE], FIL__, __LINE__, 0,
+			     MSG_E_READ, tmp);
 	  SH_FREE(tmp);
 	  return -1;
 	}
@@ -1098,6 +1184,8 @@ static int sha1_stream(char * filename, void *resblock)
     while (sum < BLOCKSIZE 
 	   && n != 0);
 
+    ++pages_read;
+
     /* If end of file is reached, end the loop.  */
     if (n == 0)
       break;
@@ -1105,15 +1193,28 @@ static int sha1_stream(char * filename, void *resblock)
     /* Process buffer with BLOCKSIZE bytes.  Note that
        BLOCKSIZE % 64 == 0
     */
-    sha_update(&ctx, buffer, BLOCKSIZE);
-    sh.stat.bytes_hashed += BLOCKSIZE;
+    sha_update(&ctx, (sha_word8*) buffer, BLOCKSIZE);
+    sh.statistics.bytes_hashed += BLOCKSIZE;
+
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+    if ((IO_Limit) > 0 && (pages_read == 32)) /* check for I/O limit */
+      {
+	sh_unix_io_pause ();
+	pages_read = 0;
+      }
+    if (sig_termfast == 1) 
+      {
+	return -1;
+      }
+#endif 
+
   }
 
   /* Add the last bytes if necessary.  */
   if (sum > 0)
     {
-      sha_update(&ctx, buffer, sum);
-      sh.stat.bytes_hashed += sum;
+      sha_update(&ctx, (sha_word8*) buffer, sum);
+      sh.statistics.bytes_hashed += sum;
     }
 
   sha_final (&ctx);
@@ -1125,7 +1226,7 @@ static int sha1_stream(char * filename, void *resblock)
 
 
 char * sh_tiger_sha1_hash  (char * filename, TigerType what, 
-			    unsigned long Length)
+			    unsigned long Length, int timeout)
 {
   int cnt = (int) Length;  /* fix compiler warning */
   static char out[KEY_LEN+1];
@@ -1139,10 +1240,10 @@ char * sh_tiger_sha1_hash  (char * filename, TigerType what,
       return out;
     }
 
-  sha1_stream (filename, sha1buffer);
+  sha1_stream (filename, sha1buffer, timeout);
 
   for (cnt = 0; cnt < 20; ++cnt)
-    sprintf (&out[cnt*2], _("%02X"), sha1buffer[cnt]);
+    sprintf (&out[cnt*2], _("%02X"),sha1buffer[cnt]);/* known to fit  */
   for (cnt = 40; cnt < KEY_LEN; ++cnt)
     out[cnt] = '0';
   out[KEY_LEN] = '\0';
@@ -1155,6 +1256,11 @@ char * sh_tiger_sha1_hash  (char * filename, TigerType what,
 
 static int hash_type = 0;
 
+int sh_tiger_gethashtype ()
+{
+  return hash_type;
+}
+
 int sh_tiger_hashtype (char * c)
 {
   SL_ENTER( _("sh_tiger_hashtype"));
@@ -1181,24 +1287,34 @@ int sh_tiger_hashtype (char * c)
   SL_RETURN( 0, _("sh_tiger_hashtype"));
 }
 
+static char * sh_tiger_hash_internal (char * filename, TigerType what, 
+				      unsigned long Length, int timeout);
+
+char * sh_tiger_hash (char * filename, TigerType what, 
+		      unsigned long Length)
+{
+  return sh_tiger_hash_internal (filename, what, Length, 0);
+}
+
 char * sh_tiger_generic_hash (char * filename, TigerType what, 
-			      unsigned long Length)
+			      unsigned long Length, int timeout)
 {
 #ifdef USE_SHA1
   if (hash_type == 1)
-    return sh_tiger_sha1_hash (filename, what, Length);
+    return sh_tiger_sha1_hash    (filename, what, Length, timeout);
 #endif
 #ifdef USE_MD5
   if (hash_type == 2)
-    return sh_tiger_md5_hash  (filename, what, Length);
+    return sh_tiger_md5_hash     (filename, what, Length, timeout);
 #endif
-  return sh_tiger_hash      (filename, what, Length);
+  return sh_tiger_hash_internal  (filename, what, Length, timeout);
 }
 
 /*
  * -------   end new ---------  */
   
-char * sh_tiger_hash (char * filename, TigerType what, unsigned long Length)
+static char * sh_tiger_hash_internal (char * filename, TigerType what, 
+				      unsigned long Length, int timeout)
 {
 #ifdef HAVE_LONG_LONG
   word64 * res;
@@ -1209,9 +1325,8 @@ char * sh_tiger_hash (char * filename, TigerType what, unsigned long Length)
 
   SL_ENTER( _("sh_tiger_hash"));
 
-  res = sh_tiger_hash_val (filename, 
-			   what, 
-			   Length);
+  res = sh_tiger_hash_val (filename, what, Length, timeout);
+
   if (res != NULL)
     {
 #ifdef HAVE_LONG_LONG
@@ -1256,9 +1371,7 @@ char * sh_tiger_hash_gpg (char * filename, TigerType what,
 
   SL_ENTER(_("sh_tiger_hash_gpg"));
 
-  res = sh_tiger_hash_val (filename, 
-			   what, 
-			   Length);
+  res = sh_tiger_hash_val (filename, what, Length, 0);
   if (res != NULL)
     {
 #ifdef HAVE_LONG_LONG
@@ -1325,9 +1438,8 @@ UINT32 * sh_tiger_hash_uint32 (char * filename,
 
   memset(out, 0, 6 * sizeof(UINT32));
 
-  res = sh_tiger_hash_val (filename, 
-			   what, 
-			   Length);
+  res = sh_tiger_hash_val (filename,  what,  Length, 0);
+
   if (res != NULL)
     {
 #ifdef HAVE_LONG_LONG
