@@ -53,6 +53,10 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+
+#ifdef  HOST_IS_HPUX
+#define _XOPEN_SOURCE_EXTENDED
+#endif
 #include <netdb.h>
 
 #ifndef FD_SET
@@ -79,6 +83,8 @@
 #include "sh_tools.h"
 #include "sh_utils.h"
 #include "sh_tiger.h"
+#define SH_NEED_GETHOSTBYXXX
+#include "sh_static.h"
 
 #undef  FIL__
 #define FIL__  _("sh_tools.c")
@@ -109,90 +115,6 @@ char * errorExplain (int err_num)
 
 #endif
 
-/* ----------   old version, obsolete  -------------
- */
-#if 0
-char * sh_tools_safe_name(const char * str, int flag)
-{
-  register int  i = 0;
-  char       *  retval;
-  unsigned char c;
-  const char *  p;
-  char          oct[32];
-  int           len = flag;  /* 'unused parameter' warning */
-
-  SL_ENTER(_("sh_tools_safe_name"));
-
-  len = sl_strlen(str);
-
-#ifdef SH_USE_XML
-  retval = SH_ALLOC(len * 6 + 2);
-#else
-  retval = SH_ALLOC(len * 4 + 2);
-#endif
-
-  p = str;
-  while (*p) 
-    {
-      c = *p;
-      if (*p == '\n')
-	{
-	  retval[i] = ' ';
-	}
-      else if (c < 31 || c > 126
-#ifdef WITH_DATABASE
-	       || (*p) == '\''
-#endif
-	       ) 
-	{
-	  sprintf(oct, _("%c%03o"), '\\', c);            /* known to fit  */
-	  retval[i] = oct[0]; ++i;
-	  retval[i] = oct[1]; ++i;
-	  retval[i] = oct[2]; ++i;
-	  retval[i] = oct[3]; 
-	}
-#ifdef SH_USE_XML
-      else if (flag == 1 && (*p) == '"')
-	{ 
-	  retval[i]   = '&'; retval[i+1] = 'q'; 
-	  retval[i+2] = 'u'; retval[i+3] = 'o'; 
-	  retval[i+4] = 't'; retval[i+5] = ';';
-	  i += 5;
-	} 
-      else if (flag == 1 && (*p) == '&')
-	{ 
-	  retval[i]   = '&'; retval[i+1] = 'a'; 
-	  retval[i+2] = 'm'; retval[i+3] = 'p'; 
-	  retval[i+4] = ';';
-	  i += 4;
-	} 
-      else if (flag == 1 && (*p) == '<') 
-	{     /* left angle       */
-	  retval[i] = '&'; ++i; 
-	  retval[i] = 'l'; ++i;
-	  retval[i] = 't'; ++i;
-	  retval[i] = ';';
-	} 
-      else if (flag == 1 && (*p) == '>') 
-	{     /* right angle      */
-	  retval[i] = '&'; ++i; 
-	  retval[i] = 'g'; ++i;
-	  retval[i] = 't'; ++i;
-	  retval[i] = ';';
-	}
-#endif
-      else 
-	{
-	  retval[i] = *p;
-	}
-      ++i; ++p;
-    }
-  retval[i] = '\0';
-  SL_RETURN( retval, _("sh_tools_safe_name"));
-}
-#else
-/* ----------   new version  -------------
- */
 /* --- recode all \blah escapes to '=XX' format, and also code all
  *     remaining unprintable chars                                 ---
  */
@@ -233,31 +155,33 @@ char * sh_tools_safe_name (const char * instr, int flag)
 #ifdef SH_USE_XML
       if (flag == 1 && (*p) == '"')
 	{ 
-	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;
+	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;  /* known to fit  */
 	  continue;
 	} 
       else if (flag == 1 && (*p) == '&')
 	{ 
-	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;
+	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;  /* known to fit  */
 	  continue;
 	} 
       else if (flag == 1 && (*p) == '<') 
 	{     /* left angle       */
-	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;
+	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;  /* known to fit  */
 	  continue;
 	} 
       else if (flag == 1 && (*p) == '>') 
 	{     /* right angle      */
-	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;
+	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;  /* known to fit  */
 	  continue;
-	} 
+	}
+#else
+      (void) flag; /* fix compiler warning */
 #endif
 
       if ( (*p) != '\\' && (*p) != '&' && (*p) != '='  && (*p) != '\'') 
         {
 	  if (c < 31 || c > 126)
 	    {
-	      sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;
+	      sprintf(&outstr[i], "=%02x", c); i+=3; ++p;/* known to fit  */
 	    }
 	  else
 	    {
@@ -268,12 +192,14 @@ char * sh_tools_safe_name (const char * instr, int flag)
 	}
       else if ((*p) == '\'')
 	{
-	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;
+	  sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;  /* known to fit  */
 	}
       else if (*p == '=')
 	{
-	  if (p[1] != '"')
-	    { sprintf(&outstr[i], "=%02x", c);   i+=3; ++p; }
+	  if (p[1] != '"' && p[1] != '<')
+	    { 
+	      sprintf(&outstr[i], "=%02x", c); i+=3; ++p;/* known to fit  */
+	    }
 	  else
 	    { outstr[i] = *p; ++i; ++p; }
 	}
@@ -289,34 +215,43 @@ char * sh_tools_safe_name (const char * instr, int flag)
 
 	  switch (*p) {
 	  case '\\':
-	    sprintf(&outstr[i], "=%02x", c);   i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", c);  i+=3; ++p; /* known to fit  */
 	    break;
 	  case 'n':
-	    sprintf(&outstr[i], "=%02x", '\n'); i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", '\n');          /* known to fit  */
+	    i+=3; ++p;
 	    break;
 	  case 'b':
-	    sprintf(&outstr[i], "=%02x", '\b'); i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", '\b');          /* known to fit  */
+	    i+=3; ++p;
 	    break;		       
 	  case 'r':		       
-	    sprintf(&outstr[i], "=%02x", '\r'); i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", '\r');          /* known to fit  */
+	    i+=3; ++p;
 	    break;		       
 	  case 't':		       
-	    sprintf(&outstr[i], "=%02x", '\t'); i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", '\t');          /* known to fit  */
+	    i+=3; ++p;
 	    break;		       
 	  case 'v':		       
-	    sprintf(&outstr[i], "=%02x", '\v'); i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", '\v');          /* known to fit  */
+	    i+=3; ++p;
 	    break;		       
 	  case 'f':		       
-	    sprintf(&outstr[i], "=%02x", '\f'); i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", '\f');          /* known to fit  */
+	    i+=3; ++p;
 	    break;		       
 	  case '\'':		       
-	    sprintf(&outstr[i], "=%02x", '\''); i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", '\'');          /* known to fit  */
+	    i+=3; ++p;
 	    break;		       
 	  case '\"':	/* also encode quoted '"' */ 	       
-	    sprintf(&outstr[i], "=%02x", '\"'); i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", '\"');          /* known to fit  */
+	    i+=3; ++p;
 	    break;		       
 	  case ' ':		       
-	    sprintf(&outstr[i], "=%02x", ' ');  i+=3; ++p;
+	    sprintf(&outstr[i], "=%02x", ' ');           /* known to fit  */
+	    i+=3; ++p;
 	    break;
 	  default:
 	    if (strlen(p) < 3)
@@ -329,7 +264,8 @@ char * sh_tools_safe_name (const char * instr, int flag)
 		val_octal = (unsigned char) strtoul(tmp, (char **)NULL, 8);
 		if (val_octal != '\0') { 
 		  c = val_octal;
-		  sprintf(&outstr[i], "=%02x", c); i+=3;
+		  sprintf(&outstr[i], "=%02x", c);       /* known to fit  */
+		  i+=3;
 		} 
 		p += 3;
 	      }
@@ -343,20 +279,24 @@ char * sh_tools_safe_name (const char * instr, int flag)
 
 	  if (p[0] == 'a' && p[1] == 'm' && p[2] == 'p' && p[3] == ';')
 	    {
-	      sprintf(&outstr[i], "=%02x", '&');  i+=3; p += 4;
+	      sprintf(&outstr[i], "=%02x", '&');         /* known to fit  */
+	      i+=3; p += 4;
 	    }
 	  else if (p[0] == 'q' && p[1] == 'u' && p[2] == 'o' && p[3] == 't' &&
 		   p[4] == ';')
 	    {
-	      sprintf(&outstr[i], "=%02x", '"');  i+=3; p += 5;
+	      sprintf(&outstr[i], "=%02x", '"');         /* known to fit  */
+	      i+=3; p += 5;
 	    }
 	  else if (p[0] == 'l' && p[1] == 't' && p[2] == ';')
 	    {
-	      sprintf(&outstr[i], "=%02x", '<');  i+=3; p += 3;
+	      sprintf(&outstr[i], "=%02x", '<');         /* known to fit  */
+	      i+=3; p += 3;
 	    }
 	  else if (p[0] == 'g' && p[1] == 't' && p[2] == ';')
 	    {
-	      sprintf(&outstr[i], "=%02x", '>');  i+=3; p += 3;
+	      sprintf(&outstr[i], "=%02x", '>');         /* known to fit  */
+	      i+=3; p += 3;
 	    }
 	  else /* conserve the '&' */
 	    {
@@ -366,11 +306,9 @@ char * sh_tools_safe_name (const char * instr, int flag)
     } /* while (p && *p) */
   
   outstr[i] = '\0';
-  flag = 0; /* fix compiler warning */
   SL_RETURN( outstr, _("sh_tools_safe_name"));
 }
 
-#endif
 
 /* extern int h_errno; */ 
 
@@ -524,12 +462,12 @@ int connect_port (char * address, int port,
 #endif
 	{
       
-	  host_entry = gethostbyname(address);
+	  host_entry = sh_gethostbyname(address);
 
 	  if (host_entry == NULL || host_entry->h_addr == NULL) 
 	    {
 	      sl_strlcpy(ecall, _("gethostbyname"), SH_MINIBUF);
-#ifndef HOST_IS_HPUX
+#ifndef NO_H_ERRNO
 	      *errnum = h_errno;
 #else
 	      *errnum = 666;
@@ -558,12 +496,13 @@ int connect_port (char * address, int port,
 		    {
 		      host_name = SH_ALLOC(sl_strlen(host_entry->h_name) + 1);
 		      if (sl_strlen(host_entry->h_name) > 0)
-			strcpy(host_name, host_entry->h_name);
+			strcpy(host_name,                /* known to fit  */
+			       host_entry->h_name);
 		      else
 			host_name[0] = '\0';
 		    }
 
-		  host_entry = gethostbyaddr ((char *) &sinr.sin_addr, 
+		  host_entry = sh_gethostbyaddr ((char *) &sinr.sin_addr, 
 					      sizeof(struct in_addr),
 					      AF_INET);
 		  if (host_entry == NULL || host_entry->h_name == NULL)
@@ -982,12 +921,19 @@ void get_header (unsigned char * head, unsigned long * bytes, char * u)
 #endif
 
 #if defined(SH_WITH_CLIENT) || defined(SH_WITH_SERVER)
+
+#ifdef  SH_ENCRYPT_2
+#define TRANS_BYTES 65120
+#else
+#define TRANS_BYTES 65280
+#endif
+
 void put_header (unsigned char * head, int protocol, 
 		 unsigned long * length, char * u)
 {
 
   /* static long transfer_limit = (8 * SH_BUFSIZE); V0.8 */
-  static unsigned long transfer_limit = 65280 + 6 + KEY_LEN;
+  static unsigned long transfer_limit = TRANS_BYTES + 6 + KEY_LEN;
 
   SL_ENTER(_("put_header"));
 
@@ -1019,6 +965,327 @@ void put_header (unsigned char * head, int protocol,
 }
 #endif
 
+/* ------------------------------------------
+ *
+ *  version 2 client/server protocol
+ *
+ * ------------------------------------------ 
+ *
+ * header :  flag size[2]
+ *
+ * payload:  random_pad[8] protocol[4] size[4] payload[payload_size] padding
+ *
+ * full_size <= 8192; payload_size <= 8176 (511*16); msg_size <= 8128 (508*16)
+ * (msg_size = payload_size - key_len = payload_size - 48)
+ */ 
+#ifdef SH_WITH_SERVER
+#define SH_V2_FULLSIZE  240
+#define SH_V2_PAYLOAD   224
+#define SH_V2_MESSAGE   176
+#else
+#define SH_V2_FULLSIZE 1024
+#define SH_V2_PAYLOAD  1008
+#define SH_V2_MESSAGE   960
+#endif
+
+#ifdef SH_ENCRYPT
+#include "rijndael-api-fst.h"
+#endif
+
+void sh_tools_show_header (unsigned char * head, char sign)
+{
+#define SH_IS_ASCII(c) (((c) & ~0x7f) == 0)
+
+
+  int    msg_size = (256 * (unsigned int)head[1] + (unsigned int)head[2]);
+  char   code[32]; 
+  char * p = &code[0];
+
+  memset (code, ' ', 32); /* space */
+ 
+  if ((head[0] & SH_PROTO_SRP) != 0) { p[0]='S';p[1]='R';p[2]='P';}
+  p += 4;
+  if ((head[0] & SH_PROTO_MSG) != 0) { p[0]='M';p[1]='S';p[2]='G';}
+  p += 4;
+  if ((head[0] & SH_PROTO_BIG) != 0) { p[0]='B';p[1]='I';p[2]='G';}
+  p += 4;
+  if ((head[0] & SH_PROTO_END) != 0) { p[0]='E';p[1]='N';p[2]='D';}
+  p += 4;
+  if ((head[0] & SH_PROTO_ENC) != 0) { p[0]='E';p[1]='N';p[2]='C';}
+  p += 4;
+  if ((head[0] & SH_PROTO_EN2) != 0) { p[0]='E';p[1]='N';p[2]='2';}
+  code[23] = '\0';
+
+  if (SH_IS_ASCII(head[3]) && isalpha(head[3]) &&
+      SH_IS_ASCII(head[4]) && isalpha(head[4]) &&			   
+      SH_IS_ASCII(head[5]) && isalpha(head[5]) &&			   
+      SH_IS_ASCII(head[6]) && isalpha(head[6])) {
+    fprintf(stderr, "%c %3o %s %5d  %c  %c  %c  %c\n", sign,
+	    head[0], code, msg_size, head[3], head[4], head[5], head[6]); 
+  } else {
+    fprintf(stderr, "%c %3o %s %5d %2X %2X %2X %2X\n", sign,
+	    head[0], code, msg_size, head[3], head[4], head[5], head[6]); 
+  }
+  return;
+}
+
+#ifdef SH_ENCRYPT
+/*
+ * #define DEBUG_EN2
+ *
+ * ingest version 1 7-byte header and payload, return version2 header/payload
+ * last 4 bytes of outgoing header are set to dummy value
+ */
+char * sh_tools_makePack (unsigned char * header, 
+			  char * payload, int payload_size,
+			  keyInstance * keyInstE)
+{
+  UINT32 rpad[3];
+  unsigned char   head[16];
+  double epad;
+  int    i_epad = 0;
+  int    i_blk = payload_size / 16;
+  int    i_blkmax = SH_V2_FULLSIZE / 16;
+  int    pads = 0;
+  int    full_size;
+  char * full_ret;
+
+  char                  * p;
+  RIJ_BYTE                    inBlock[B_SIZ]; 
+  RIJ_BYTE                    outBlock[B_SIZ];
+  int                     j;
+  cipherInstance          cipherInst;
+  int                     err_num;
+  int                     blkfac;
+
+  /* 
+     SL_REQUIRE (i_blk*16 == payload_size, _("payload_size % 16 != 0"));
+  */
+  if ((i_blk * 16) != payload_size) ++i_blk;
+#ifdef DEBUG_EN2
+  fprintf(stderr, "SEND <%d> blocks <%d>\n", payload_size, i_blk);
+#endif
+  /* random_pad
+   */
+  rpad[1] = taus_get (&(skey->rng0[0]), &(skey->rng1[0]), &(skey->rng2[0]));
+  memcpy (head,      &rpad[1],    4);
+  rpad[0] = taus_get (&(skey->rng0[0]), &(skey->rng1[0]), &(skey->rng2[0]));
+  memcpy (&head[4],  &rpad[0],    4);
+  rpad[2] = taus_get (&(skey->rng0[0]), &(skey->rng1[0]), &(skey->rng2[0]));
+  memcpy (&head[8],  &rpad[2],    4);
+
+  /* protocol
+   */
+  /* memcpy (&head[8],  &header[3], 4); */
+
+  /* size (payload)
+   */ 
+  head[12] = header[1];
+  head[13] = header[2];
+  head[14] = '\0';
+  head[15] = '\0';
+
+  if (i_blk < i_blkmax) 
+  {
+    pads = i_blkmax - i_blk;
+    /* memcpy((char *) &rpad[2], &head[12], 4); */
+    epad = taus_get_double (&rpad);
+#ifdef DEBUG_EN2
+    fprintf(stderr, "PAD1 <%d> <%f>\n", pads, epad);
+#endif
+    i_epad = (int) (pads * epad);
+#ifdef DEBUG_EN2
+    fprintf(stderr, "PAD2 <%d> <%d>\n", i_epad, (i_epad*16));
+#endif
+  }
+
+  full_size = 
+  /* head     */ 16 + 
+  /* payload  */ (i_blk*16) + /* payload_size + */ 
+  /* pad      */ (i_epad * 16);
+
+  full_ret = SH_ALLOC(full_size);
+  memcpy(full_ret,                   head,    16);
+  if (payload != NULL)
+    {
+      memcpy(&full_ret[16],              payload, payload_size);
+    }
+  if ((i_blk*16) > payload_size) 
+    {
+#ifdef DEBUG_EN2
+      fprintf(stderr, "SEN2 <%d>\n", (i_blk*16) - payload_size);
+#endif
+      memset(&full_ret[16+payload_size], '\0', (i_blk*16) - payload_size);
+      payload_size = i_blk * 16;
+    }
+  memset(&full_ret[16+payload_size], '\0', i_epad*16);
+#ifdef DEBUG_EN2
+  fprintf(stderr, "SEN3 <%d> <%d>\n", full_size, i_epad*16);
+#endif
+
+  /* rewrite header
+   */
+  header[1]   = (unsigned int)(full_size/256);
+  header[2]   = (unsigned int)(full_size - (256 * header[1]));
+  /* don't erase protocol from header 
+     memset(&header[3], '\0', 4);
+  */
+  p = full_ret; blkfac = full_size / 16;
+
+  err_num = cipherInit (&cipherInst, MODE_CBC, NULL);
+  
+  if (err_num < 0) 
+    {
+      sh_error_handle((-1), FIL__, __LINE__, -1, MSG_E_SUBGEN,
+		      errorExplain(err_num), 
+		      _("sh_tools_makePack: cipherInit"));
+    }
+  for (j = 0; j < blkfac; ++j)
+    {
+      memcpy(inBlock, p, B_SIZ);
+      err_num = blockEncrypt(&cipherInst, keyInstE, 
+			     inBlock, 128 * BNUM, outBlock);
+      if (err_num < 0)
+	{
+	  sh_error_handle((-1), FIL__, __LINE__, -1, MSG_E_SUBGEN,
+			  errorExplain(err_num), 
+			  _("sh_tools_makePack: blockEncrypt"));
+	}
+      memcpy(p, outBlock, B_SIZ);
+      p += B_SIZ;
+    }
+
+  return full_ret;
+}
+
+/* write a 7-byte header and return payload as expected by version 1
+ * last 4 bytes of incoming header are dummy
+ */
+char * sh_tools_revertPack (unsigned char * header, char * message, 
+			    keyInstance * keyInstD,
+			    unsigned long message_size)
+{
+  unsigned long   msg_size;
+  char          * msg_ret;
+
+  char                  * p;
+  RIJ_BYTE                    inBlock[B_SIZ]; 
+  RIJ_BYTE                    outBlock[B_SIZ];
+  int                     j;
+  cipherInstance          cipherInst;
+  int                     err_num;
+  int                     blkfac;
+
+  msg_size = (256 * (unsigned int)header[1] + (unsigned int)header[2]);
+#ifdef DEBUG_EN2
+  fprintf(stderr, "RECV <%lu>\n", msg_size);
+#endif
+  if (msg_size > message_size) {
+    msg_size = message_size;
+#ifdef DEBUG_EN2
+    fprintf(stderr, "RECV TRUNC1 <%lu>\n", msg_size);
+#endif
+  }
+
+  p = message; blkfac = msg_size / 16;
+
+  err_num = cipherInit (&cipherInst, MODE_CBC, NULL);
+  
+  if (err_num < 0) 
+    {
+      sh_error_handle((-1), FIL__, __LINE__, -1, MSG_E_SUBGEN,
+		      errorExplain(err_num), 
+		      _("sh_tools_revertPack: cipherInit"));
+    }
+  for (j = 0; j < blkfac; ++j)
+    {
+      memcpy(inBlock, p, B_SIZ);
+      err_num = blockDecrypt(&cipherInst, keyInstD, 
+			     inBlock, 128 * BNUM, outBlock);
+      if (err_num < 0)
+	{
+	  sh_error_handle((-1), FIL__, __LINE__, -1, MSG_E_SUBGEN,
+			  errorExplain(err_num), 
+			  _("sh_tools_revertPack: blockDecrypt"));
+	}
+      memcpy(p, outBlock, B_SIZ);
+      p += B_SIZ;
+    }
+  
+  /* rewrite size in header
+   */
+  header[1]   = message[12];
+  header[2]   = message[13];
+  msg_size = (256 * (unsigned int)header[1] + (unsigned int)header[2]);
+
+  if (msg_size > (message_size-16)) 
+    {
+      msg_size    = message_size-16;
+      header[1]   = (unsigned int)(msg_size/256);
+      header[2]   = (unsigned int)(msg_size - (256 * header[1]));
+#ifdef DEBUG_EN2
+      fprintf(stderr, "RECV TRUNC2 <%lu>\n", msg_size);
+#endif
+    }
+#ifdef DEBUG_EN2
+  fprintf(stderr, "REC2 <%lu>\n", msg_size);
+#endif
+  /* protocol
+   */
+  /* memcpy(&header[3], &message[8], 4); */
+
+  /* payload 
+   */
+  msg_ret = SH_ALLOC(msg_size+1);
+  if (msg_size > 0)
+    {
+      memcpy(msg_ret, &message[16], msg_size);
+    }
+  msg_ret[msg_size] = '\0';
+#ifdef DEBUG_EN2
+  fprintf(stderr, "REC3 <%lu>\n", msg_size);
+#endif
+  SH_FREE(message);
+
+  return msg_ret;
+}
+#endif
+
+int sh_tools_hash_add(char * key, char * buf, int buflen)
+{
+  char         * theSig;
+
+  SL_ENTER(_("sh_tools_hash_add"));
+
+  theSig = sh_util_siggen (key, buf, buflen);
+  sl_strlcat(buf, theSig, buflen + KEY_LEN + 1);
+      
+  SL_RETURN((0), _("sh_tools_hash_add"));
+}
+
+
+/* return 0 (== FALSE) if no match, else 1 (== TRUE)
+ */
+int sh_tools_hash_vfy(char * key, char * buf, int buflen)
+{
+  char           hash[KEY_LEN+1];
+  register int   i;
+  char         * theSig;
+
+  SL_ENTER(_("sh_tools_hash_vfy"));
+
+  theSig = sh_util_siggen (key, buf, buflen);
+  sl_strlcpy(hash, theSig, KEY_LEN+1);
+
+  for (i = 0; i < KEY_LEN; ++i)
+    {
+      if (buf[buflen + i] != hash[i])
+	SL_RETURN((0), _("sh_tools_hash_vfy"));
+    }
+      
+  SL_RETURN((1), _("sh_tools_hash_vfy"));
+}
+
 /* ------------------------------------------ */
 
 #if defined (SH_WITH_SERVER)
@@ -1036,9 +1303,13 @@ char * hash_me (char * key, char * buf,   int buflen)
 
   SL_ENTER(_("hash_me"));
 
+#ifdef DEBUG_EN2
+  fprintf(stderr, "hash_me    <%s> <%d>\n", 
+	  (key == NULL) ? "NULL" : key, buflen);
+#endif
   /* key = H(NSRV,NCLT,SK)
    */
-  ASSERT_RET((key != NULL), _("key != NULL"), (NULL))
+  ASSERT_RET((key != NULL), _("key != NULL"), (NULL));
 
   theSig = sh_util_siggen (key, buf, buflen);
   sl_strlcpy(hash, theSig, KEY_LEN+1);
@@ -1058,6 +1329,7 @@ char * hash_me (char * key, char * buf,   int buflen)
 #endif
 
 #if defined (SH_WITH_CLIENT)
+
 /* verify the checksum of a buffer; checksum comes first
  */
 int hash_check(char * key, 
@@ -1069,6 +1341,10 @@ int hash_check(char * key,
 
   SL_ENTER(_("hash_check"));
 
+#ifdef DEBUG_EN2
+  fprintf(stderr, "hash_check <%s> <%d>\n", 
+	  (key == NULL) ? "NULL" : key, buflen);
+#endif
   theSig = sh_util_siggen (key, &buf[KEY_LEN], buflen-KEY_LEN);
   sl_strlcpy(hash, theSig, KEY_LEN+1);
       
@@ -1079,6 +1355,7 @@ int hash_check(char * key,
     }
   SL_RETURN((0), _("hash_check"));
 }
+
 #endif
 
 #if defined (SH_WITH_SERVER)
@@ -1104,9 +1381,7 @@ char * get_client_conf_file (char * peer, unsigned long * length)
   sl_strlcat(ret, _("/rc."), size);
   sl_strlcat(ret, peer, size);
   
-  sl_set_suid();
-  status = retry_lstat (FIL__, __LINE__, ret, &buf);
-  sl_unset_suid();
+  status = retry_stat (FIL__, __LINE__, ret, &buf);
 
   if (status == 0)
     goto lab_end;
@@ -1117,9 +1392,7 @@ char * get_client_conf_file (char * peer, unsigned long * length)
   sl_strlcpy(ret, base, size);
   sl_strlcat(ret, "/rc", size);
   
-  sl_set_suid();
-  status = retry_lstat (FIL__, __LINE__, ret, &buf);
-  sl_unset_suid();
+  status = retry_stat (FIL__, __LINE__, ret, &buf);
 
   if (status == 0)
     goto lab_end;
@@ -1132,6 +1405,13 @@ char * get_client_conf_file (char * peer, unsigned long * length)
   SL_RETURN(NULL, _("get_client_conf_file"));
 
  lab_end:
+  if (buf.st_size > 0x7fffffff)
+    {
+      sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, status, MSG_E_SUBGEN,
+		    _("File too large"), _("get_client_conf_file"));
+      SH_FREE(base);
+      SL_RETURN(NULL, _("get_client_conf_file"));
+    }
   *length = (unsigned long) buf.st_size;
   SH_FREE(base);
   SL_RETURN(ret, _("get_client_conf_file"));
@@ -1160,9 +1440,7 @@ char * get_client_data_file (char * peer, unsigned long * length)
   sl_strlcat(ret, _("/file."), size);
   sl_strlcat(ret, peer, size);
   
-  sl_set_suid();
-  status = retry_lstat (FIL__, __LINE__, ret, &buf);
-  sl_unset_suid();
+  status = retry_stat (FIL__, __LINE__, ret, &buf);
 
   if (status == 0)
     goto lab1_end;
@@ -1174,9 +1452,7 @@ char * get_client_data_file (char * peer, unsigned long * length)
   sl_strlcpy(ret, base, size);
   sl_strlcat(ret, _("/file"), size);
   
-  sl_set_suid();
-  status = retry_lstat (FIL__, __LINE__, ret, &buf);
-  sl_unset_suid();
+  status = retry_stat (FIL__, __LINE__, ret, &buf);
 
   if (status == 0)
     goto lab1_end;
@@ -1190,6 +1466,13 @@ char * get_client_data_file (char * peer, unsigned long * length)
   SL_RETURN(NULL, _("get_client_data_file"));
 
  lab1_end:
+  if (buf.st_size > 0x7fffffff)
+    {
+      sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, status, MSG_E_SUBGEN,
+		    _("File too large"), _("get_client_data_file"));
+      SH_FREE(base);
+      SL_RETURN(NULL, _("get_client_data_file"));
+    }
   *length = (unsigned long) buf.st_size;
   SH_FREE(base);
   SL_RETURN(ret, _("get_client_data_file"));
@@ -1218,10 +1501,16 @@ SL_TICKET open_tmp ()
   len        = sl_strlen(SH_TMPDIR) + 1;
   my_tmp_dir = SH_ALLOC(len);
   sl_strlcpy(my_tmp_dir, SH_TMPDIR, len); 
+#else
+#if defined(SH_WITH_SERVER)
+  len        = sl_strlen(DEFAULT_LOGDIR) + 1;
+  my_tmp_dir = SH_ALLOC(len);
+  sl_strlcpy(my_tmp_dir, DEFAULT_LOGDIR, len); 
 #else
   len        = sl_strlen(sh.effective.home) + 1;
   my_tmp_dir = SH_ALLOC(len);
-  sl_strlcpy(my_tmp_dir, sh.effective.home, len); 
+  sl_strlcpy(my_tmp_dir, sh.effective.home, len);
+#endif 
 #endif
 
   if (0 !=  tf_trust_check (my_tmp_dir, SL_YESPRIV))
@@ -1253,11 +1542,9 @@ SL_TICKET open_tmp ()
 
     /* check whether it already exists (paranoia)
      */
-    sl_set_suid();
     errno  = 0;
     status = retry_lstat(FIL__, __LINE__, file, &buf);
     error  = errno;
-    sl_unset_suid(); 
 
     if ( (status < 0) && (error == ENOENT) ) /* file does not exist        */
       status = GOOD;
@@ -1291,6 +1578,12 @@ SL_TICKET open_tmp ()
   } while (status == BAD);
 
   fd = sl_open_rdwr (file, SL_YESPRIV);
+  if (SL_ISERROR(fd))
+    {
+      TPT(( 0, FIL__, __LINE__, _("msg=<Error %d temporary file %s>\n"), 
+	    fd, file));
+    }
+  
 
   SH_FREE (file);
   SH_FREE(my_tmp_dir);
