@@ -85,7 +85,7 @@ extern  unsigned long mask_ALLIGNORE;
 extern  unsigned long mask_ATTRIBUTES;
 #define MASK_LOGFILES_   (MASK_ATTRIBUTES_|MODI_HLN|MODI_LNK|MODI_INO)
 extern  unsigned long mask_LOGFILES;
-#define MASK_LOGGROW_    MASK_LOGFILES_ 
+#define MASK_LOGGROW_    (MASK_LOGFILES_|MODI_SIZ) 
 extern  unsigned long mask_LOGGROW;
 #define MASK_READONLY_   (MASK_LOGFILES_|MODI_CHK|MODI_SIZ|MODI_MTM|MODI_CTM)
 extern  unsigned long mask_READONLY;
@@ -98,7 +98,7 @@ extern  unsigned long mask_USER1;
 typedef struct file_struct {
   unsigned long    check_mask;
   int              reported;
-  char             fullpath[PATH_MAX+2];
+  char             fullpath[PATH_MAX];
   ShFileType       type;
   dev_t            dev;
   ino_t            ino;
@@ -121,19 +121,30 @@ typedef struct file_struct {
   time_t           mtime;
   time_t           ctime;
 
-  char             linkpath[PATH_MAX+2];
+  char             linkpath[PATH_MAX];
   mode_t           linkmode;
   char             link_c_mode[11];
   int              linkisok;
 } file_type;
 
+/* mlock utilities
+ */
+int sh_unix_mlock(void * addr, size_t len);
+int sh_unix_munlock(void * addr, size_t len);
+int sh_unix_count_mlock();
+
+/* chroot directory
+ */
+int sh_unix_set_chroot(char * str);
+
 /* whether to use localtime for file timesatams in logs
  */
 int sh_unix_uselocaltime (char * c);
 
 /* set I/O limit
  */
-int sh_unix_set_io_limit (char * c);
+int  sh_unix_set_io_limit (char * c);
+void sh_unix_io_pause ();
 
 /* get file type
  */
@@ -155,9 +166,23 @@ unsigned long sh_unix_longtime (void);
  */
 void sh_unix_closeall (int fd);
 
-/* check lock on log file
+
+/* write lock for filename
  */
-int sh_unix_testlock (char * filename);
+int sh_unix_write_lock_file(char * filename);
+
+/* rm lock(s) for log file(s)
+ */
+int sh_unix_rm_lock_file(char * filename);
+
+/* write the PID file
+ */
+int sh_unix_write_pid_file();
+
+/* rm the PID file
+ */
+int sh_unix_rm_pid_file();
+
 
 /* checksum of own binary
  */
@@ -175,6 +200,11 @@ int tf_add_trusted_user(char *);
  */
 int tf_trust_check (char * file, int mode);
 
+/* initialize group vector
+ */
+int  sh_unix_initgroups  (const char * in_user, gid_t in_gid);
+int  sh_unix_initgroups2 (uid_t         in_pid, gid_t in_gid);
+
 /* set the timeserver address
  */
 int sh_unix_settimeserver (char * address);
@@ -182,13 +212,17 @@ void reset_count_dev_time(void);
 
 /* lock the key
  */
-void sh_unix_mlock(void);
+void sh_unix_memlock(void);
 
 /* deamon mode 
  */
 int sh_unix_setdeamon(char * dummy);
 
-/* test whether file exists
+/* Test whether file exists
+ */
+int sh_unix_file_stat(char * path);
+
+/* test whether file exists with proper attributes
  */
 int sh_unix_file_exists(int fd);
 
@@ -207,16 +241,8 @@ int sh_unix_test_proc(void);
 
 /* obtain file info
  */
-int sh_unix_getinfo (int level,
-		     char * filename, file_type * theFile, char * fileHash);
-
-/* unlock the default errlog file
- */
-int sh_unix_unlock (void);
-
-/* lock the log file
- */
-int sh_unix_lock (void);
+int sh_unix_getinfo (int level, char * filename, file_type * theFile, 
+		     char * fileHash, int flagrel);
 
 /* read file, return length read
  */
@@ -242,6 +268,7 @@ int  sh_unix_getUser (void);
  */
 char *  sh_unix_getUIDdir (int level, uid_t uid);
 
+
 #ifdef HAVE_GETTIMEOFDAY
 unsigned long sh_unix_notime (void);
 #endif
@@ -255,7 +282,73 @@ int  sh_unix_getline_stealth  (SL_TICKET fd, char * str, int len);
 void sh_unix_xor_code (char * str, int len);
 #endif
 
-int sh_derr(void);
+#if defined(SCREW_IT_UP)
+/* for raise() 
+ */
+#include <signal.h>
+#include <errno.h>
 
+void   sh_sigtrap_handler (int signum);
+extern volatile int sh_not_traced;
+
+#ifdef HAVE_GETTIMEOFDAY
+#if TIME_WITH_SYS_TIME
+#include <sys/time.h>
+#include <time.h>
+#else
+#if HAVE_SYS_TIME_H
+#include <sys/time.h>
+#else
+#include <time.h>
+#endif
+#endif
+extern struct timeval  save_tv;
+extern struct timezone save_tz;
+#endif
+
+static inline
+int  sh_sigtrap_prepare()
+{
+  struct sigaction act_trap;
+  int              val_retry;
+  act_trap.sa_handler   = &sh_sigtrap_handler;   /* signal action     */
+  act_trap.sa_flags     = 0;                     /* init sa_flags     */
+  sigemptyset ( &act_trap.sa_mask );             /* set an empty mask */
+  do {
+    val_retry = sigaction(SIGTRAP, &act_trap, NULL);
+  } while (val_retry < 0 && errno == EINTR);
+  return 0;
+}
+
+static inline 
+int sh_derr(void)
+{
+  sh_not_traced = 0;
+
+#ifdef HAVE_GETTIMEOFDAY
+  gettimeofday(&save_tv, &save_tz);
+#endif
+
+#if defined(__linux__) && defined(__GNUC__)
+  __asm__ __volatile__ ("int $0x03");
+#else
+  raise(SIGTRAP);
+#endif
+  
+  if (sh_not_traced == 0)
+    _exit(5);
+  sh_not_traced = 0;
+  return (0);
+}
+
+#else
+
+static inline 
+int sh_derr(void)
+{
+  return 0;
+}
+/* #if defined(SCREW_IT_UP) */
+#endif
 
 #endif
