@@ -24,6 +24,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
+#include <unistd.h>
 
 #if TIME_WITH_SYS_TIME
 #include <sys/time.h>
@@ -69,6 +70,114 @@ int sh_util_flagval(char * c, int * fval)
   SL_RETURN( (-1), _("sh_util_flagval"));
 }
 
+int sh_util_timeout_check (SH_TIMEOUT * sh_timer)
+{
+  UINT64 now = time(NULL);
+  UINT64 dif;
+  
+  if (sh_timer->flag_ok == S_FALSE)
+    {
+      /* first time
+       */
+      if (sh_timer->time_last == 0)
+	{
+	  sh_timer->time_last = now;
+	  return S_TRUE;
+	}
+      /* later on
+       */
+      dif = now - sh_timer->time_last;
+      if (dif < sh_timer->time_dist)
+	{
+	  return S_FALSE;
+	}
+      sh_timer->time_last = now;
+      return S_TRUE;
+    }
+  sh_timer->time_last = now;
+  return S_FALSE;
+}
+
+static int sh_ask_update = S_FALSE;
+
+int sh_util_set_interactive(char * str)
+{
+  if (str == NULL)
+    sh_ask_update = S_TRUE;
+  else
+    sh_ask_update = S_TRUE;
+
+  return 0;
+}
+
+#if !defined(STDIN_FILENO)
+#define STDIN_FILENO 0
+#endif
+#if !defined(STDERR_FILENO)
+#define STDERR_FILENO 0
+#endif
+
+int sh_util_ask_update(char * path)
+{
+  int    inchar, c;
+  int    i = S_TRUE;
+  char * tmp = NULL;
+
+  SL_ENTER(_("sh_util_ask_update"));
+
+  if (sh_ask_update != S_TRUE)
+    {
+      SL_RETURN(i, _("sh_util_ask_update"));
+    }
+
+#ifdef HAVE_TTYNAME
+  if (!ttyname(STDIN_FILENO))
+    {
+      if (NULL != ttyname(STDERR_FILENO))
+        {
+          if (NULL == freopen(ttyname(STDERR_FILENO), "r", stdin))
+            {
+              fprintf(stderr, _("Cannot continue: stdin is not a terminal\n"));
+              exit(EXIT_FAILURE);
+	    }
+        }
+      else
+        {
+          fprintf(stderr, _("Cannot continue: stdin is not a terminal\n"));
+          exit(EXIT_FAILURE);
+        }
+    }
+#endif
+
+  if (sh_ask_update == S_TRUE)
+    {
+      tmp = sh_util_safe_name (path);
+      fprintf (stderr, _("Update %s [Y/n] ? "), tmp);
+      SH_FREE(tmp);
+      while (1 == 1)
+	{
+	  c = fgetc(stdin); inchar = c;
+	  while (c != '\n' && c != EOF)
+	    c = fgetc(stdin);
+	  /* fprintf(stderr, "CHAR (1): %c\n", inchar); */
+	  if (inchar == 'Y' || inchar == 'y' || inchar == '\n')
+	    {
+	      break;
+	    }
+	  else if (inchar == 'n' || inchar == 'N')
+	    {
+	      i = S_FALSE;
+	    }
+	  else
+	    {
+	      fprintf(stderr, _("Please answer y(es) or n(o)\n"));
+	    }
+	}
+    }
+
+  SL_RETURN(i, _("sh_util_ask_update"));
+}
+
 int sh_util_hidesetup(char * c)
 {
   int i;
@@ -94,6 +203,36 @@ char * sh_util_strdup (const char * str)
   SL_RETURN( p, _("sh_util_strdup"));
 }
 
+/* by the eircom.net computer incident
+ * response team
+ */
+char * sh_util_strsep (char **str, const char *delim) 
+{
+  char *ret, *c, *d;
+
+  SL_ENTER(_("sh_util_strsep"));
+  ret = *str;
+
+  if (ret == NULL) {
+    SL_RETURN(ret, _("sh_util_strsep"));
+  }
+
+  for (c = *str; *c != '\0'; c++) {
+    for (d = (char *) delim; *d != '\0'; d++) {
+      if (*c == *d) {
+        *c = '\0';
+        *str = c + 1;
+        SL_RETURN(ret, _("sh_util_strsep"));
+      }
+    }
+  }
+
+  /* If we get to here, there's no delimiters in the string */
+  *str = NULL;
+  SL_RETURN(ret, _("sh_util_strsep"));
+}
+
+
 /* returned string must be free'd by caller
  */
 char * sh_util_formatted (const char * formatt, st_format * ftab)
@@ -506,6 +645,9 @@ long sh_util_compress (char * dest, char * src, unsigned long dest_size)
   if (dest_size == 0)
     SL_RETURN((0), _("sh_util_compress"));
   
+  if ((dest == NULL) || (src == NULL))
+    SL_RETURN((0), _("sh_util_compress"));
+  
   dest_end = sl_strlen(dest);
   add      = &dest[dest_end];
   get      = src;
@@ -519,7 +661,8 @@ long sh_util_compress (char * dest, char * src, unsigned long dest_size)
 	  ++count;
 	}
       ++get; 
-      if (*get == '\0') /* end of src reached */
+      if (*get == '\0' && (count < (dest_size-dest_end))) 
+	/* end of src reached */
 	{
 	  *add = *get;  /* copy the '\0'      */
 	  break;        /* and stop copying   */
@@ -527,7 +670,7 @@ long sh_util_compress (char * dest, char * src, unsigned long dest_size)
     }
 
   dest[dest_size-1] = '\0'; /* paranoia       */
-  SL_RETURN((count), _("sh_util_compress")); /* no of chars copyed */    
+  SL_RETURN((count), _("sh_util_compress")); /* no of chars copied */    
 }
 
 
@@ -639,7 +782,7 @@ UINT32 taus_get (void *state1, void *state2, void *state3)
  */
 double taus_get_double (void *vstate)
 {
-  return taus_get_long (vstate) / (4294967296.0 - 1.0) ;
+  return taus_get_long (vstate) / (4294967296.0 + 1.0) ;
 }
 
 #define LCG(n) ((69069 * n) & 0xffffffffUL)
@@ -708,6 +851,10 @@ int taus_seed ()
       memcpy (&skey->rng2[0], &bufx[4*sizeof(UINT32)],   2*sizeof(UINT32));
       memset (bufx, '\0', 9 * sizeof(UINT32) + 1);
 
+      skey->rng0[2] = 0;
+      skey->rng1[2] = 0;
+      skey->rng2[2] = 0;
+
       taus_set_from_state( &(skey->rng0[0]), &(skey->rng0[0]));
       taus_set_from_state( &(skey->rng1[0]), &(skey->rng1[0]));
       taus_set_from_state( &(skey->rng2[0]), &(skey->rng2[0]));
@@ -786,7 +933,7 @@ int sh_util_set_newkey (char * new)
 
   if (NULL == (outpath = malloc(strlen(path) + 1 + 4)))
     goto bail_mem;
-  sprintf (outpath, _("%s.out"), path); 
+  sprintf (outpath, _("%s.out"), path);               /* known to fit  */ 
 
   fp = sl_open_read(path, SL_NOPRIV);
   if (SL_ISERROR(fp))
@@ -899,7 +1046,7 @@ void sh_util_encode (char * data, char * salt, int mode, char fill)
 	     KEY_LEN+1);
 
   sl_strlcpy(skey->vernam, 
-	     sh_util_hmac_tiger (skey->vernam, new_key, 8),
+	     sh_util_hmac_tiger (skey->vernam, (char*) new_key, 8),
 	     KEY_LEN+1);
 
   /* The following routine adds/subtracts  data[j] and vernam[j] mod 16.
@@ -1057,14 +1204,17 @@ int sh_util_keyinit (char * buf, long size)
 		     6*sizeof(UINT32));
   p[size-1] = '\0';
 
-  i = sl_strlcpy(buf, 
+  /*
+  i = sl_strlcpy(buf,
 		 sh_tiger_hash ((char *) bufy, TIGER_DATA, 
 				6*sizeof(UINT32)), 
 		 size);
+  */
+  i = sl_strlcpy(buf, p, size);
 
   memset (bufy, '\0', 6*sizeof(UINT32));
 
-  if (status == 0 && (!SL_ISERROR(i) || (i == SL_ETRUNC)) )
+  if ((status == 0) && (!SL_ISERROR(i)) )
     SL_RETURN((0),_("sh_util_keyinit"));
 
   if (SL_ISERROR(i))
@@ -1075,10 +1225,49 @@ int sh_util_keyinit (char * buf, long size)
 }
 
 #if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
+
+static unsigned char sh_obscure_index[256];
+
+int sh_util_obscure_ok (char * str)
+{
+  unsigned long   i;
+  char * endptr = str;
+
+  SL_ENTER(_("sh_util_obscure_ex"));
+
+  if (0 == sl_strncmp("all", str, 3))
+    {
+      for (i = 0; i < 255; ++i)
+	{
+	  sh_obscure_index[i] = 1;
+	}
+      SL_RETURN(0, _("sh_util_obscure_ex"));
+    }
+
+  for (i = 0; i < 255; ++i)
+    {
+      sh_obscure_index[i] = 0;
+    }
+
+  while (*endptr != '\0')
+    {
+      i = strtoul (endptr, &endptr, 0);
+      if (i > 255)
+	{
+	  SL_RETURN(-1, _("sh_util_obscure_ex"));
+	}
+      sh_obscure_index[i] = 1;
+      if (*endptr == ',')
+	++endptr;
+    }
+  SL_RETURN(0, _("sh_util_obscure_ex"));
+}
+
 int sh_util_obscurename (ShErrLevel level, char * name_orig, int flag)
 {
   char * name = name_orig;
   char * safe;
+  unsigned int i;
 
   SL_ENTER(_("sh_util_obscurename"));
 
@@ -1094,17 +1283,22 @@ int sh_util_obscurename (ShErrLevel level, char * name_orig, int flag)
 	   (*name) == '\v' || iscntrl((int) *name) || 
 	   ((*name) != ' ' && !isgraph ((int) *name)) ) 
 	{
-	  if (flag == S_TRUE)
+	  i = (unsigned char) *name;
+	  if (sh_obscure_index[i] != 1)
 	    {
-	      safe = sh_util_safe_name (name_orig);  
-	      sh_error_handle (level, FIL__, __LINE__, 0, MSG_FI_OBSC, 
-			       safe);
-	      SH_FREE(safe);
+	      if (flag == S_TRUE)
+		{
+		  safe = sh_util_safe_name (name_orig);  
+		  sh_error_handle (level, FIL__, __LINE__, 0, MSG_FI_OBSC, 
+				   safe);
+		  SH_FREE(safe);
+		}
+	      SL_RETURN((-1),_("sh_util_obscurename"));
 	    }
-	  SL_RETURN((-1),_("sh_util_obscurename"));
 	}
       name++;
     }
+
   SL_RETURN((0),_("sh_util_obscurename"));
 }
 #endif
@@ -1127,7 +1321,7 @@ char * sh_util_basename(char * fullpath)
     if (fullpath[i] == '/') break;
   }
 
-  /* -- Not a vaild path. --
+  /* -- Not a valid path. --
    */
   if ((fullpath[i] != '/') && (i == 0) ) 
     SL_RETURN(NULL, _("sh_util_basename"));
@@ -1173,6 +1367,7 @@ char * sh_util_safe_name (const char * name)
   const char  * p;
   char        * retval;
   char          oct[32];
+  char          format[16];
 
   SL_ENTER(_("sh_util_safe_name"));
 
@@ -1195,10 +1390,18 @@ char * sh_util_safe_name (const char * name)
   retval = SH_ALLOC(4 * sl_strlen(name) + 2);
 #endif 
 
+  sl_strncpy(format, _("%c%03o"), 16);
+
   p = name;
 
   while (*p) {
-    if ( (*p) == '\\') {           /* backslash        */
+    /* Most frequent cases first
+     */
+    if ( ((*p) >= 'a' && (*p) <= 'z')  || ((*p) == '/') || ((*p) == '.') ||
+	 ((*p) >= '0' && (*p) <= '9')  || 
+	 ((*p) >= 'A' && (*p) <= 'Z')) {
+      retval[i] = *p; 
+    } else if ( (*p) == '\\') {           /* backslash        */
       retval[i] = '\\'; ++i; 
       retval[i] = '\\';
     } else if ( (*p) == '\n') {    /* newline          */
@@ -1257,7 +1460,7 @@ char * sh_util_safe_name (const char * name)
       retval[i] = '\"';
 #endif
     } else if (!isgraph ((int) *p)) {    /* not printable    */
-      sprintf(oct, _("%c%03o"), '\\',                 /* known to fit  */
+      sprintf(oct, format, '\\',                 /* known to fit  */
 	      (unsigned char) *p);
       retval[i] = oct[0]; ++i;
       retval[i] = oct[1]; ++i;
@@ -1328,6 +1531,7 @@ char * sh_util_strconcat (const char * arg1, ...)
   SL_RETURN(strnew, _("sh_util_strconcat"));
 }
 
+
 #ifdef HAVE_REGEX_H
 
 #include <regex.h>
@@ -1344,13 +1548,13 @@ int sh_util_regcmp (char * regex_str, char * in_str)
 
   SL_ENTER(_("sh_util_regcmp"));
 
-  MBLK( status = regcomp(&preg, regex_str, REG_NOSUB|REG_EXTENDED); )
+  status = regcomp(&preg, regex_str, REG_NOSUB|REG_EXTENDED);
 
   if (status == 0)
     {
       if ((status = regexec(&preg, in_str, 0, NULL, 0)) == 0) 
 	{
-	  MBLK( regfree (&preg); )
+	  regfree (&preg);
 	  SL_RETURN((0), _("sh_util_regcmp"));
 	}
     }
@@ -1365,7 +1569,7 @@ int sh_util_regcmp (char * regex_str, char * in_str)
       SH_FREE(errbuf);
     }
 	
-  MBLK( regfree (&preg); )
+  regfree (&preg);
   SL_RETURN((-1), _("sh_util_regcmp"));
 }
 
