@@ -168,7 +168,7 @@ typedef struct log_user {
   char                name[UT_NAMESIZE+1];
   char                ut_host[UT_HOSTSIZE+1];
   char                ut_ship[16]; /* IP address */
-  time_t              time; 
+  time_t              time;
   struct log_user   * next;
 } blah_utmp;
 
@@ -183,21 +183,21 @@ static void sh_utmp_login_morechecks(struct SH_UTMP_S * ut);
 static void sh_utmp_addlogin (struct SH_UTMP_S * ut);
 static void sh_utmp_check_internal(int mode);
 
-int    ShUtmpLoginSolo    = 7;
-int    ShUtmpLoginMulti   = 7;
-int    ShUtmpLogout       = 7;
-int    ShUtmpActive       = BAD;
+int    ShUtmpLoginSolo    = SH_ERR_INFO;
+int    ShUtmpLoginMulti   = SH_ERR_WARN;
+int    ShUtmpLogout       = SH_ERR_INFO;
+int    ShUtmpActive       = S_TRUE;
 time_t ShUtmpInterval     = 300;
 
 sh_rconf sh_utmp_table[] = {
-  {
-    N_("severitylogin"),
-    sh_utmp_set_login_solo
-  },
   {
     N_("severityloginmulti"),
     sh_utmp_set_login_multi
   },
+  {
+    N_("severitylogin"),
+    sh_utmp_set_login_solo
+  },
   {
     N_("severitylogout"),
     sh_utmp_set_logout_good
@@ -216,6 +216,11 @@ sh_rconf sh_utmp_table[] = {
   },
 };
 
+int sh_utmp_null()
+{
+  return 0;
+}
+
 
 
 
@@ -253,6 +258,10 @@ sh_rconf sh_utmp_table[] = {
 
 static FILE * sh_utmpfile = NULL;
 static char   sh_utmppath[80] = _PATH_UTMP;
+
+/* sh_utmp_feed_forward is for optimizing
+ * (fseek instead of getutent loop)
+ */
 static long   sh_utmp_feed_forward = 0;
 
 static void sh_utmp_utmpname(const char * str)
@@ -260,7 +269,7 @@ static void sh_utmp_utmpname(const char * str)
   SL_ENTER(_("sh_utmp_utmpname"));
   if (sh_utmpfile != NULL)
     {
-      MBLK( fclose (sh_utmpfile); )
+      fclose (sh_utmpfile);
       sh_utmpfile = NULL;
     }
 
@@ -285,7 +294,7 @@ static void sh_utmp_setutent(void)
       fd = aud_open (FIL__, __LINE__, SL_NOPRIV, sh_utmppath, O_RDONLY, 0);
       if (fd >= 0)
 	{
-	  MBLK( sh_utmpfile = fdopen(fd, "r"); )
+	  sh_utmpfile = fdopen(fd, "r");
 	}
 
       /* -- If (sh_utmpfile == NULL) then either the open() or the fdopen()
@@ -300,7 +309,11 @@ static void sh_utmp_setutent(void)
 	}
     }
   (void) fseek (sh_utmpfile, 0L, SEEK_SET);
-  (void) fseek (sh_utmpfile, sh_utmp_feed_forward, SEEK_CUR); 
+  if (-1 == fseek (sh_utmpfile, sh_utmp_feed_forward, SEEK_CUR))
+    {
+      sh_utmp_feed_forward = 0; /* modified Apr 4, 2004 */
+      (void) fseek (sh_utmpfile, 0L, SEEK_SET);
+    }
   clearerr (sh_utmpfile);
   SL_RET0(_("sh_utmp_setutent"));
 }
@@ -308,7 +321,7 @@ static void sh_utmp_setutent(void)
 static void sh_utmp_endutent(void)
 {
   SL_ENTER(_("sh_utmp_endutent"));
-  MBLK( fclose(sh_utmpfile); )
+  fclose(sh_utmpfile);
   sh_utmpfile = NULL;
   SL_RET0(_("sh_utmp_endutent"));
 }
@@ -439,15 +452,23 @@ static time_t  lastcheck;
  *************/
 int sh_utmp_init ()
 {
+  static int done = 0;
+
   SL_ENTER(_("sh_utmp_init"));
   if (ShUtmpActive == BAD)
     SL_RETURN( (-1), _("sh_utmp_init"));
 
+  /* do not re-initialize after a re-configuration
+   */
+  if (done == 1) {
+    SL_RETURN( (0), _("sh_utmp_init"));
+  }
   lastcheck  = time (NULL);
   userlist   = NULL;
   memset (&save_utmp, 0, sizeof(struct SH_UTMP_S));
   sh_utmp_check_internal (2); /* current logins */
   sh_utmp_check_internal (0);
+  done = 1;
   SL_RETURN( (0), _("sh_utmp_init"));
 }
 
@@ -456,6 +477,10 @@ int sh_utmp_init ()
  * module cleanup
  *
  *************/
+#ifdef HAVE_UTTYPE
+static int sh_utmp_login_clean();
+#endif
+
 int sh_utmp_end ()
 {
   struct log_user * user    = userlist;
@@ -468,6 +493,10 @@ int sh_utmp_end ()
       user    = user->next;
       SH_FREE(userold);
     }
+  userlist = NULL;
+#ifdef HAVE_UTTYPE
+  sh_utmp_login_clean();
+#endif
   SL_RETURN( (0), _("sh_utmp_end"));
 }
 
@@ -564,6 +593,87 @@ int sh_utmp_set_login_activate (char * c)
   SL_RETURN(i, _("sh_utmp_set_login_activate"));
 }
 
+#ifdef HAVE_UTTYPE
+struct login_ct {
+  char name[UT_NAMESIZE+1];
+  int  nlogin;
+  struct login_ct * next;
+};
+
+static struct login_ct * login_ct_list = NULL;
+
+static int sh_utmp_login_clean()
+{
+  struct login_ct * list = login_ct_list;
+  struct login_ct * old;
+
+  login_ct_list = NULL;
+
+  while (list)
+    {
+      old  = list;
+      list = list->next;
+      SH_FREE(old);
+    }
+  return 0;
+}
+
+/* add a username to the list of logged-in users
+ */
+static int sh_utmp_login_a(char * str)
+{
+  struct login_ct * list = login_ct_list;
+
+  while (list)
+    {
+      if (0 == sl_strcmp(list->name, str))
+	{
+	  ++(list->nlogin);
+	  return list->nlogin;
+	}
+      list = list->next;
+    }
+  list = SH_ALLOC(sizeof(struct login_ct));
+  sl_strlcpy(list->name, str, UT_NAMESIZE+1);
+  list->nlogin  = 1;
+  list->next    = login_ct_list;
+  login_ct_list = list;
+  return 1;
+}
+
+static int sh_utmp_login_r(char * str)
+{
+  struct login_ct * list = login_ct_list;
+  struct login_ct * old  = login_ct_list;
+
+  while (list)
+    {
+      if (0 == sl_strcmp(list->name, str))
+	{
+	  list->nlogin -= 1;
+	  if (list->nlogin > 0)
+	    {
+	      return list->nlogin;
+	    }
+	  if (login_ct_list == list) /* modified Apr 4, 2004 */
+	    {
+	      login_ct_list = list->next;
+	      SH_FREE(list);
+	    }
+	  else
+	    {
+	      old->next = list->next;
+	      SH_FREE(list);
+	    }
+	  return 0;
+	}
+      old  = list;
+      list = list->next;
+    }
+  return 0;
+}
+
+#endif
 
 
 /* for each login:
@@ -571,14 +681,40 @@ int sh_utmp_set_login_activate (char * c)
  *    - link device.ut_record -> log_record
  *    - link user.ut_record   -> log_record
  */
+#ifdef HAVE_UTTYPE  
+static int sh_utmp_is_virtual (char * in_utline, char * in_uthost)
+{
+  if (in_utline != NULL)
+    {
+      if (0 == sl_strncmp(in_utline, _("ttyp"), 4))
+	{ return 0; }
+      else if (0 == sl_strncmp(in_utline, _("ttyq"), 4))
+	{ return 0; }
+    }
+  if (in_uthost != NULL   && 
+      in_uthost[0] == ':' && 
+      in_uthost[1] == '0' && 
+      in_uthost[2] == '\0')
+    {
+      /* return 0; */
+      return 1;  /* actually does not seem portable, not even among linuxes */
+    }
+  return 1;
+}
+#endif
 
 static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 {
   struct log_user   * user     = userlist;
-  struct log_user   * username = userlist;
   struct log_user   * userold  = userlist;
+#ifdef HAVE_UTTYPE  
+  struct log_user   * username = userlist;
+#endif
 
   char   ttt[TIM_MAX];
+#ifdef HAVE_UTTYPE
+  int    status;
+#endif
 
   SL_ENTER(_("sh_utmp_addlogin"));
 
@@ -594,7 +730,7 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
       SL_RET0(_("sh_utmp_addlogin"));
     }
   memcpy (&save_utmp, ut, sizeof(struct SH_UTMP_S));
-      
+
 
   /* ------- find user -------- 
    */
@@ -606,13 +742,14 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
       user = user->next;
     }
 
+#ifdef HAVE_UTTYPE  
   while (username != NULL) 
     {
       if (0 == sl_strncmp(username->name, ut->ut_name, UT_NAMESIZE) ) 
 	break;
       username = username->next;
     }
-
+#endif
   
 #ifdef HAVE_UTTYPE  
   /* ---------- LOGIN -------------- */
@@ -628,6 +765,8 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
       sl_strlcpy(user->name,    ut->ut_name, UT_NAMESIZE+1);
 #ifdef HAVE_UTHOST
       sl_strlcpy(user->ut_host, ut->ut_host, UT_HOSTSIZE+1);
+#else
+      user->ut_host[0] = '\0';
 #endif
 #ifdef HAVE_UTADDR
       sl_strlcpy(user->ut_ship,my_inet_ntoa(*(struct in_addr*)&(ut->ut_addr)),16);
@@ -635,9 +774,10 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
       user->time = ut->ut_time;
 
       if (username == NULL                              /* not yet logged in */
-	  || 0 == sl_strncmp(ut->ut_line, _("ttyp"), 4) /* in virt. console  */
-	  || 0 == sl_strncmp(ut->ut_line, _("ttyq"), 4) /* in virt. console  */
+          || 0 == sl_strncmp(ut->ut_line, _("ttyp"), 4) /* in virt. console  */
+          || 0 == sl_strncmp(ut->ut_line, _("ttyq"), 4) /* in virt. console  */
 	  ) {
+	status = sh_utmp_login_a(user->name);
 	sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX);
 	sh_error_handle( ShUtmpLoginSolo, FIL__, __LINE__, 0,
 #if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
@@ -655,12 +795,13 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 #elif defined(HAVE_UTHOST)
 			 user->ut_host,
 #endif
-			 ttt
+			 ttt,
+			 status
 			 );
       } else
-	if (0 != sl_strncmp(ut->ut_line, _("ttyp"), 4) &&
-	    0 != sl_strncmp(ut->ut_line, _("ttyq"), 4))
+	if (0 != sh_utmp_is_virtual(ut->ut_line, user->ut_host))
 	{       
+	status = sh_utmp_login_a(user->name);
 	sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX);
 	sh_error_handle( ShUtmpLoginMulti, FIL__, __LINE__, 0,
 #if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
@@ -678,7 +819,8 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 #elif defined(HAVE_UTHOST)
 			 user->ut_host,
 #endif
-			 ttt
+			 ttt,
+			 status
 			 );
       }
       
@@ -694,7 +836,9 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
     {
       if (user != NULL)
 	{
-	  sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX);
+	  status = sh_utmp_login_r(user->name);
+	  /* sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX); */
+	  sl_strlcpy(ttt, sh_unix_time (ut->ut_time), TIM_MAX);
 	  sh_error_handle( ShUtmpLogout, FIL__, __LINE__, 0,
 #if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
 			   MSG_UT_LG3X,
@@ -711,7 +855,8 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 #elif defined(HAVE_UTHOST)
 			   user->ut_host,
 #endif
-			   ttt
+			   ttt,
+			   status
 			   );
 	  userold->next = user->next;
 	  if (user == userlist)
@@ -727,7 +872,7 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 	  sh_error_handle( ShUtmpLogout, FIL__, __LINE__, 0,
 			   MSG_UT_LG3C,
 			   terminated_line,
-			   ttt
+			   ttt, 0
 			   );
 	}
       SL_RET0(_("sh_utmp_addlogin"));
@@ -773,13 +918,15 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 #elif defined(HAVE_UTHOST)
 		       user->ut_host,
 #endif
-		       ttt
+		       ttt,
+		       1
 		       );
       sh_utmp_login_morechecks(ut);
     }
   else  /* probably a logout */
     {
-      sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX);
+      sl_strlcpy(ttt, sh_unix_time (ut->ut_time), TIM_MAX);
+      /* sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX); */
 
       sh_error_handle( ShUtmpLogout, FIL__, __LINE__, 0,
 #if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
@@ -797,10 +944,13 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 #elif defined(HAVE_UTHOST)
 		       user->ut_host,
 #endif
-		       ttt
+		       ttt,
+		       1
 		       );
       sh_utmp_logout_morechecks(user);
       userold->next = user->next;
+      if (user == userlist)       /* inserted Apr 4, 2004 */
+	userlist = user->next;
       SH_FREE(user);
       user = NULL;
     }
@@ -814,7 +964,7 @@ static off_t         lastsize = 0;
 static unsigned long lastread = 0;
 
 char * mode_path[] = { _PATH_WTMP, _PATH_WTMP, _PATH_UTMP };
-    
+
 static void sh_utmp_check_internal (int mode)
 {
   struct stat   buf;
