@@ -100,8 +100,8 @@ typedef long int SL_TICKET;             /* Unique ID for opened files.    */
 #define SL_EBADGID   -1051	/* Group writeable and not trustworthy.*/
 #define SL_EBADOTH   -1052	/* World writeable.                    */
 
-
-
+#define SL_TOOMANY   -1053      /* Too many open files                 */
+#define SL_TIMEOUT   -1054      /* Timeout in read                     */
 /*
  * All int functions return SL_NONE on success.
  */
@@ -114,9 +114,13 @@ extern "C" {
 
   char * sl_get_errmsg();
 
-  void   my_free   (void * ptr);
+  /* ---------------------------------------------------------------- 
+   *
+   *    Heap consistency routines
+   *
+   * ---------------------------------------------------------------- */
 
-  void * sl_malloc (size_t size);
+  int sl_test_heap();
 
   /* ---------------------------------------------------------------- 
    *
@@ -124,6 +128,13 @@ extern "C" {
    *
    * ---------------------------------------------------------------- */
 
+  /*
+   * A memset that does not get optimized away
+   */
+  void *sl_memset(void *s, int c, size_t n);
+#undef  memset
+#define memset sl_memset
+
   /* 
    * Copy src to dst. siz is the length of dst.
    */
@@ -150,7 +161,7 @@ extern "C" {
   /*
    * A robust drop-in replacement of strncpy. strlcpy is preferable.
    */
-  char * sl_strncpy(char *dst, const char *src, size_t size);
+  char * sl_strncpy(/*@out@*/char *dst, const char *src, size_t size);
 
   /*
    * Robust strncat.
@@ -165,21 +176,17 @@ extern "C" {
   /*
    * robust strncmp replacement
    */
-  int sl_strncmp(char * a, char * b, size_t n);
+  int sl_strncmp(const char * a, const char * b, size_t n);
 
   /*
    * robust strcmp replacement
    */
-  int sl_strcmp(char * a, char * b);
+  int sl_strcmp(const char * a, const char * b);
 
   /*
    * robust strlen replacement
    */
-#if 0
-  size_t sl_strlen(const char *s);
-#else
 #define sl_strlen(arg) ((arg == NULL) ? 0 : (strlen(arg)))
-#endif
 
   /* ---------------------------------------------------------------- 
    *
@@ -325,6 +332,10 @@ extern "C" {
    */
   int sl_read (SL_TICKET ticket, void * buf, size_t count);
 
+  int sl_read_timeout (SL_TICKET ticket, void * buf, size_t count, int timeout);
+
+  int sl_read_fast (SL_TICKET ticket, void * buf_in, size_t count);
+
   /* Write file.
    */
   int sl_write (SL_TICKET ticket, void * msg, long nbytes);
@@ -339,15 +350,19 @@ extern "C" {
 
   /* Check whether file is trustworthy.
    */
-  int sl_trustfile(char * path, long * ok, long * bad);
+  int sl_trustfile(char * path, uid_t * ok, uid_t * bad);
 
   /* Check whether file is trustworthy.
    */
-  int sl_trustfile_euid(char * filename, long euid);
+  int sl_trustfile_euid(char * filename, uid_t euid);
+
+  /* purge list of trusted users
+   */
+  int  sl_trust_purge_user ();
 
   /* Add a trusted user.
    */
-  int  sl_trust_add_user (long pwid);
+  int  sl_trust_add_user (uid_t pwid);
 
   /* Get error string.
    */
@@ -376,31 +391,18 @@ extern "C" {
 #define SL_ISERROR(x) ((long)(x) < 0)
 
 #if defined(WITH_TPT) 
-
 #define TPT(arg) dlog arg ;
-
 #else
-
 #define TPT(arg)
-
 #endif
 
-extern int myblock_flag;
-#define MBLK(arg)                                      \
-do {                                                   \
-  if (myblock_flag == 0){                              \
-    myblock_flag = 1;                                  \
-    arg ;                                              \
-    myblock_flag = 0;                                  \
-  }                                                    \
-} while (0);
 
 /*
  * The 'require' macro.
  */
 #define SL_REQUIRE(assertion, astext)                  \
 do {                                                   \
-    if (assertion) ;                                   \
+    /*@i@*/ if (assertion) ;                           \
     else {                                             \
         dlog(0, FIL__, __LINE__, SDG_AFAIL,            \
                  FIL__, __LINE__, astext);             \
@@ -412,10 +414,13 @@ do {                                                   \
 /*
  * The enter macro. Prints the trace if TRACE is on.
  */
+extern int slib_do_trace;
+extern int slib_trace_fd;
+
 #if defined(SL_DEBUG)
-#define SL_ENTER(s)  sl_stack_push(s);
+#define SL_ENTER(s)  sl_stack_push(s, FIL__, __LINE__);
 #else
-#define SL_ENTER(s)  ;
+#define SL_ENTER(s)  /*@i@*/if (slib_do_trace) sl_trace_in(s, FIL__, __LINE__);
 #endif
 
 /*
@@ -424,7 +429,7 @@ do {                                                   \
 #if defined(SL_DEBUG)
 #define SL_RETURN(x, s)   \
 do {                      \
-   sl_stack_pop();        \
+   sl_stack_pop(s, FIL__, __LINE__);       \
    if (1) return(x);      \
 } while(0)
 
@@ -432,6 +437,8 @@ do {                      \
 
 #define SL_RETURN(x, s)   \
 do {                      \
+   if (slib_do_trace)     \
+     sl_trace_out(s, FIL__, __LINE__);     \
    if (1) return(x);      \
 } while(0)
 
@@ -440,22 +447,28 @@ do {                      \
 #if defined(SL_DEBUG)
 #define SL_RET0(s)      \
 do {                    \
-      sl_stack_pop();   \
+      sl_stack_pop(s, FIL__, __LINE__);  \
       if (1) return;    \
 } while(0)
 
 #else
 #define SL_RET0(s)      \
 do {                    \
-      if (1) return;    \
+   if (slib_do_trace)   \
+     sl_trace_out(s, FIL__, __LINE__);   \
+   if (1) return;       \
 } while(0)
 #endif  /* SL_RETURN macro */
 
 #if defined(SL_DEBUG)
-void sl_stack_push(char * c);
-void sl_stack_pop();
+void sl_stack_push(char * c, char * file, int line);
+void sl_stack_pop(char * c, char * file, int line);
 void sl_stack_print();
 #endif
+void sl_trace_in   (char * str, char * file, int line);
+void sl_trace_out  (char * str, char * file, int line);
+int  sl_trace_file (char * str);
+int  sl_trace_use  (char * str);
 
 
 
@@ -468,16 +481,18 @@ void sl_stack_print();
 #define SL_IRETURN(x, s)                                            \
 do {                                                                \
    if((long)(x) < 0) {                                              \
-      TPT((0,    FIL__, __LINE__, SDG_ERROR, (long)(x)))     \
+      TPT((0,    FIL__, __LINE__, SDG_ERROR, (long)(x)))            \
       sl_errno=(x);                                                 \
     }                                                               \
-   sl_stack_pop();                                                  \
+   sl_stack_pop(s, FIL__, __LINE__);                              \
    if (1) return(x);                                                \
 } while(0)
 #else
 #define SL_IRETURN(x, s)             \
 do {                                 \
    if ((long)(x) < 0) sl_errno=(x);  \
+   if (slib_do_trace)                \
+     sl_trace_out(s, FIL__, __LINE__);   \
    if (1) return(x);                 \
 } while(0)
 
