@@ -19,7 +19,7 @@ if test x$UID != x; then
 #  TRUST="--with-trusted=0,2"
   TRUST="--with-trusted=0,2,$UID"
 else
-  TRUST="--with-trusted=0,2"
+  TRUST="--with-trusted=0,2,500"
 fi
 
 echo; echo "${S}__ STARTING TEST RUN STANDALONE __${E}"; echo;
@@ -60,7 +60,21 @@ testrun1 ()
 		$MAKE distclean
 	fi
 	#
-	./configure --quiet $TRUST --enable-debug --enable-xml-log --prefix=$PW_DIR --localstatedir=$PW_DIR --with-config-file=$PW_DIR/testrc_1 --with-log-file=$PW_DIR/.samhain_log --with-pid-file=$PW_DIR/.samhain_lock --with-data-file=$PW_DIR/.samhain_file 
+	# --enable-login-watch --enable-xml-log 
+	# --enable-debug --enable-suidcheck --with-prelude
+
+	./configure --quiet $TRUST --enable-static --prefix=$PW_DIR --localstatedir=$PW_DIR --with-config-file=$PW_DIR/testrc_1 --with-log-file=$PW_DIR/.samhain_log --with-pid-file=$PW_DIR/.samhain_lock --with-data-file=$PW_DIR/.samhain_file
+
+# --with-kcheck=/boot/System.map 
+
+# --with-prelude --enable-xml-log
+
+# --with-kcheck=/boot/System.map-2.4.20-24.9
+# --with-kcheck=/boot/System.map-deb-2.4.18-686 
+
+# --enable-stealth=192
+# --with-console=$PW_DIR/dlog_file
+
 	#
 	if test x$? = x0; then
 		echo "configure completed, exit status 0"; echo;
