@@ -83,7 +83,8 @@ testrun1 ()
 	#
 	echo GPG_HASH is ${GPG_HASH}
 	#
-	./configure --quiet $TRUST --enable-debug --with-gpg=${GPG_PATH} --enable-micro-stealth=137 --enable-login-watch --prefix=$PW_DIR --localstatedir=$PW_DIR --with-config-file=$PW_DIR/testrc_1.signed  --with-checksum="${GPG_HASH}"  --with-log-file=$PW_DIR/.samhain_log --with-pid-file=$PW_DIR/.samhain_lock --with-data-file=$PW_DIR/.samhain_file
+	./configure --quiet $TRUST --enable-debug --with-gpg=${GPG_PATH} --enable-login-watch --prefix=$PW_DIR --localstatedir=$PW_DIR --with-config-file=$PW_DIR/testrc_1.signed  --with-checksum="${GPG_HASH}"  --with-log-file=$PW_DIR/.samhain_log --with-pid-file=$PW_DIR/.samhain_lock --with-data-file=$PW_DIR/.samhain_file --enable-micro-stealth=137
+	# --with-fp="EF6C EF54 701A 0AFD B86A  F4C3 1AAD 26C8 0F57 1F6C"
 	#
 	if test x$? = x0; then
 		echo "configure completed, exit status 0"; echo;
@@ -117,7 +118,7 @@ testrun1 ()
 		echo
 		echo "--- You need to sign the database .samhain_file now ---"
 		echo
-		gpg -a --clearsign $PW_DIR/.samhain_file
+		gpg -a --clearsign --not-dash-escaped $PW_DIR/.samhain_file
 		mv $PW_DIR/.samhain_file.asc $PW_DIR/.samhain_file
 
 		./samhain -t check
