@@ -58,7 +58,9 @@ testrun2a ()
 		$MAKE distclean
 	fi
 	#
-	./configure --quiet  $TRUST --enable-debug --enable-network=client --enable-srp --prefix=$PW_DIR --localstatedir=$PW_DIR --with-config-file=REQ_FROM_SERVER$PW_DIR/testrc_2 --with-data-file=REQ_FROM_SERVER$PW_DIR/.samhain_file --with-logserver=${SH_LOCALHOST}  --with-log-file=$PW_DIR/.samhain_log --with-pid-file=$PW_DIR/.samhain_lock
+	./configure --quiet  $TRUST --enable-static --enable-debug --enable-network=client --enable-srp --prefix=$PW_DIR --localstatedir=$PW_DIR --with-config-file=REQ_FROM_SERVER$PW_DIR/testrc_2 --with-data-file=REQ_FROM_SERVER$PW_DIR/.samhain_file --with-logserver=localhost  --with-log-file=$PW_DIR/.samhain_log --with-pid-file=$PW_DIR/.samhain_lock --enable-encrypt=2
+	#
+	# --disable-encrypt
 	#
 	if test x$? = x0; then
 		echo "configure completed, exit status 0"; echo;
@@ -81,7 +83,9 @@ testrun2a ()
 	make clean
 
 	#
-	./configure --quiet  $TRUST --enable-debug --enable-network=server --enable-srp --prefix=$PW_DIR --localstatedir=$PW_DIR --with-config-file=REQ_FROM_SERVER$PW_DIR/testrc_2 --with-data-file=REQ_FROM_SERVER$PW_DIR/.samhain_file --with-logserver=${SH_LOCALHOST}  --with-log-file=$PW_DIR/.samhain_log --with-pid-file=$PW_DIR/.samhain_lock
+	./configure --quiet  $TRUST  --enable-network=server --prefix=$PW_DIR --localstatedir=$PW_DIR --with-config-file=REQ_FROM_SERVER$PW_DIR/testrc_2 --with-data-file=REQ_FROM_SERVER$PW_DIR/.samhain_file --with-logserver=${SH_LOCALHOST}  --with-log-file=$PW_DIR/.samhain_log --with-pid-file=$PW_DIR/.samhain_lock --enable-encrypt=2
+	# --enable-identity=rainer
+	# --with-libwrap
 	#
 	if test x$? = x0; then
 		echo "configure completed, exit status 0"; echo;
@@ -140,7 +144,7 @@ testrun2a ()
 
 	if test x$SHINI = x0; then
 		echo; echo "${S}Start Server${E}: ./yule  &"; echo;
-		./yule &
+		./yule -q &
 		PROC_Y=$!
 		sleep 5
 
@@ -173,7 +177,7 @@ testrun2a ()
 	exit 8
     fi
 
-    export SH_LOCALHOST=$1
+    SH_LOCALHOST=$1; export SH_LOCALHOST
 
     testrun2a
 
